
Require Import Wf_nat.
Require Import Compare_dec.

Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Singleton.
Require Import AlgorithmEq.
Require Import SubjectReduction.
Require Import Algorithm.
Require Import Principal.
Require Import SoundEq.
Require Import CompleteEq.
Require Import Simp.


Lemma aeqc_complete :
  forall G c1 c2 k,
    wf G
    -> eqc G c1 c2 k
    -> aeqc G c1 c2 k.
Proof.
intros G c1 c2 k Hwf Heqc.
unfold aeqc.
apply aeq_complete; auto.
Qed.


(* Equivalent to ssubk, but with no refl or trans rules. *)
Inductive ssubk : context -> (* kd *) term -> (* kd *) term -> Prop :=
| ssubk_type {G}
    : ssubk G kd_type kd_type

| ssubk_sing {G c c'}
    : eqc G c c' kd_type
      -> ssubk G (kd_sing c) (kd_sing c')

| ssubk_sing_t {G c}
    : ofc G c kd_type
      -> ssubk G (kd_sing c) kd_type

| ssubk_pi {G k1 k1' k2 k2'}
    : ssubk G k1' k1
      -> ssubk (G; cl_cn k1') k2 k2'
      -> ofk (G; cl_cn k1) k2
      -> ssubk G (kd_pi k1 k2) (kd_pi k1' k2')

| ssubk_sigma {G k1 k1' k2 k2'}
    : ssubk G k1 k1'
      -> ssubk (G; cl_cn k1) k2 k2'
      -> ofk (G; cl_cn k1') k2'
      -> ssubk G (kd_sigma k1 k2) (kd_sigma k1' k2')

| ssubk_one {G}
    : ssubk G kd_one kd_one
.


Lemma ssubk_impl_subk :
  forall G k k',
    ssubk G k k'
    -> subk G k k'.
Proof.
intros G k k' H.
induct H; eauto with static.
Qed.


Lemma subst_ssubk :
  forall G1 G2 s k k',
    subof sof G1 s G2
    -> ssubk G2 k k'
    -> ssubk G1 (subst s k) (subst s k').
Proof.
intros G1 G2 s k k' Hs H.
revert G1 s Hs.
induct H;
intros; simp_sub; eauto 7 using ssubk, subst_eqc_type, subst_ofc_type, subof_bind_sof_cn, subst_ofk.
Qed.


Lemma ssubk_refl :
  forall G k,
    ofk G k
    -> ssubk G k k.
Proof.
intros G k H.
induct H; eauto using ssubk.
(* sing *)
intros G c Hofc.
apply ssubk_sing.
apply eqc_refl; auto.
Qed.


Lemma ssubk_trans :
  forall G k1 k2 k3,
    wf G
    -> ssubk G k1 k2
    -> ssubk G k2 k3
    -> ssubk G k1 k3.
Proof.
cut (forall sk G k1 k2 k3, 
       simp k1 = sk
       -> wf G
       -> ssubk G k1 k2
       -> ssubk G k2 k3
       -> ssubk G k1 k3).
  {
  intro Hind.
  intros G k1 k2 k3 Hwf H12 H23.
  eapply Hind; eauto.
  }
intro sk.
induct sk.

(* type *)
{
intros G k1 k2 k3 Hsk Hwf H12 H23.
invert H12; try (intros; subst; discriminate Hsk).
  {
  intros <- <-.
  invert H23.
  intros <-.
  apply ssubk_type.
  }

  {
  intros c1 c2 Heqc12 <- <-.
  invert H23.
    {
    intros c3 Heqc23 <-.
    apply ssubk_sing.
    eapply eqc_trans; eauto.
    }

    {
    intros _ <-.
    apply ssubk_sing_t.
    exact (eqc_reg _#4 Hwf Heqc12 andel).
    }
  }

  {
  intros c1 Hofc1 <- <-.
  invert H23.
  intros <-.
  apply ssubk_sing_t; auto.
  }
}


(* pi *)
{
intros sk1 IH1 sk2 IH2 G kl1 kl2 kl3 Hsk Hwf H12 H23.
invert H12; try (intros; subst; discriminate Hsk).
intros k1 k2 l1 l2 Hk21 Hl12 _ <- <-.
invert H23.
intros k3 l3 Hk32 Hl23 _ <-.
so (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hk32)) as (Hofk3 & _).
invert (subk_reg _#3 Hwf (ssubk_impl_subk _#3 H12) andel).
intros _ Hofl1.
injection Hsk.
intros <- <-.
apply ssubk_pi; eauto.
  {
  eapply IH1; eauto.
  transitivity (simp k2); eauto using ssubk_impl_subk, subk_simp.
  }
eapply IH2; eauto using wf.
so (subst_ssubk _#5 (subof_subsume_0 _#3 (ssubk_impl_subk _#3 Hk32)) Hl12) as H.
simp_sub in H.
exact H.
}

(* sigma *)
{
intros sk1 IH1 sk2 IH2 G kl1 kl2 kl3 Hsk Hwf H12 H23.
invert H12; try (intros; subst; discriminate Hsk).
intros k1 k2 l1 l2 Hk12 Hl12 _ <- <-.
invert H23.
intros k3 l3 Hk23 Hl23 _ <-.
so (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hk12)) as (Hofk1 & _).
invert (subk_reg _#3 Hwf (ssubk_impl_subk _#3 H23) ander).
intros _ Hofl3.
injection Hsk.
intros <- <-.
apply ssubk_sigma; eauto.
eapply IH2; eauto using wf.
so (subst_ssubk _#5 (subof_subsume_0 _#3 (ssubk_impl_subk _#3 Hk12)) Hl23) as H.
simp_sub in H.
exact H.
}

(* one *)
{
intros G k1 k2 k3 Hsk _ H12 H23.
invert H12; try (intros; subst; discriminate Hsk).
intros <- <-.
invert H23.
intros <-.
apply ssubk_one.
}
Qed.  


Lemma eqk_impl_ssubk :
  forall G k k',
    wf G
    -> eqk G k k'
    -> ssubk G k k'.
Proof.
intros G k k' Hwf Heqk.
cut (ssubk G k k' /\ ssubk G k' k).
  {
  intros (? & _); auto.
  }
revert Hwf.
induct Heqk.

(* refl *)
{
intros; auto using ssubk_refl.
}

(* symm *)
{
intros G k k' _ IH Hwf.
destruct (IH Hwf).
auto.
}

(* trans *)
{
intros G k1 k2 k3 _ IH1 _ IH2 Hwf.
destruct (IH1 Hwf).
destruct (IH2 Hwf).
eauto using ssubk_trans.
}

(* sing *)
{
intros G c c' Heqc Hwf.
split; apply ssubk_sing; auto with static.
}

(* pi *)
{
intros G k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 Hwf.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
destruct (IH1 Hwf) as (Hsubk1 & Hsubk1').
destruct (IH2 (wf_cn Hwf Hofk1)) as (Hsubk2 & Hsubk2').
split.
  {
  apply ssubk_pi; auto.
    {
    so (subst_ssubk _#5 (subof_conv_0 _#3 (eqk_symm Heqk1)) Hsubk2) as H.
    simp_sub in H.
    exact H.
    }

    {
    exact (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 andel).
    }
  }

  {
  apply ssubk_pi; auto.
  so (subst_ofk _#4 (subof_conv_0 _#3 (eqk_symm Heqk1)) (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 ander)).
  simp_sub in H.
  exact H.
  }
}

(* sigma *)
{
intros G k1 k1' k2 k2' Heqk1 IH1 Heqk2 IH2 Hwf.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
destruct (IH1 Hwf) as (Hsubk1 & Hsubk1').
destruct (IH2 (wf_cn Hwf Hofk1)) as (Hsubk2 & Hsubk2').
split.
  {
  apply ssubk_sigma; auto.
  so (subst_ofk _#4 (subof_conv_0 _#3 (eqk_symm Heqk1)) (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 ander)).
  simp_sub in H.
  exact H.
  }

  {
  apply ssubk_sigma; auto.
    {
    so (subst_ssubk _#5 (subof_conv_0 _#3 (eqk_symm Heqk1)) Hsubk2') as H.
    simp_sub in H.
    exact H.
    }

    {
    exact (eqk_reg _#3 (wf_cn Hwf Hofk1) Heqk2 andel).
    }
  }
}
Qed.


Lemma subk_impl_ssubk :
  forall G k k',
    wf G
    -> subk G k k'
    -> ssubk G k k'.
Proof.
intros G k k' Hwf Hsub.
revert Hwf.
induct Hsub; eauto using ssubk, eqk_impl_ssubk, ssubk_trans.

(* pi *)
{
intros G k1 k1' k2 k2' _ IH1 _ IH2 Hofk2 Hwf.
so (IH1 Hwf) as Hsub1.
so (IH2 (wf_cn Hwf (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hsub1) andel))) as Hsub2.
apply ssubk_pi; auto.
}

(* sigma *)
{
intros G k1 k1' k2 k2' _ IH1 _ IH2 Hofk2' Hwf.
so (IH1 Hwf) as Hsub1.
so (IH2 (wf_cn Hwf (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hsub1) andel))) as Hsub2.
apply ssubk_sigma; auto.
}
Qed.


Lemma asubk_complete :
  forall G k k',
    wf G
    -> subk G k k'
    -> asubk G k k'.
Proof.
intros G k k' Hwf Hsubk.
so (subk_impl_ssubk _#3 Hwf Hsubk) as H.
clear Hsubk.
revert Hwf.
induct H; eauto using asubk.

(* sing *)
{
intros G c c' Heqc Hwf.
apply asubk_sing; auto.
apply aeq_complete; auto.
}

(* pi *)
{
intros G k1 k1' k2 k2' Hsubk1 IH1 _ IH2 Hofk2 Hwf.
apply asubk_pi; auto.
apply IH2.
apply wf_cn; auto.
exact (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hsubk1) andel).
}

(* sigma *)
{
intros G k1 k1' k2 k2' Hsubk1 IH1 _ IH2 Hofk2 Hwf.
apply asubk_sigma; auto.
apply IH2.
apply wf_cn; auto.
exact (subk_reg _#3 Hwf (ssubk_impl_subk _#3 Hsubk1) andel).
}
Qed.


Lemma aofc_complete_help :
  forall G c k,
    (forall G k, wf G -> prink G c k -> aprink G c k)
    -> wf G
    -> ofc G c k
    -> aofc G c k.
Proof.
intros G c k IH Hwf Hofc.
so (prink_complete _#3 Hwf Hofc) as (k' & Hprink & Hsub).
so (IH _ _ Hwf Hprink) as Haprink.
eapply aofc_i; eauto.
apply asubk_complete; auto.
Qed.


Lemma aofk_prink_complete :
  (forall G k,
     wf G
     -> ofk G k
     -> aofk G k)
  /\
  (forall G c k,
     wf G
     -> prink G c k
     -> aprink G c k).
Proof.
cut (forall m,
       (forall G,
          wf G -> ofk G m -> aofk G m)
       /\
       (forall G k,
          wf G -> prink G m k -> aprink G m k)).
  {
  intro Hind.
  split.
    {
    intros G k.
    so (Hind k andel).
    eauto.
    }

    {
    intros G c.
    so (Hind c ander).
    eauto.
    }
  }
intro m.
induct m;
intros; split; intros;
try (match goal with
     | H : ofk _ _ |- _ => invert H; done
     | H : prink _ _ _ |- _ => invert H; done
     end);
repeat
  (match goal with
   | H : _ /\ _ |- _ => destruct H
   end);
eauto using aofk;
try (match goal with
     | H : prink _ _ _ |- _ => invert H; intros; subst

     | H : ofk _ _ |- _ => invert H; intros; subst
     end;
     eauto using aprink, aofk, wf, aofc_complete_help;
     done).
Qed.


Lemma aofk_complete :
  forall G k,
    wf G
    -> ofk G k
    -> aofk G k.
Proof.
exact (aofk_prink_complete andel).
Qed.


Lemma aprink_complete_pre :
  forall G c k,
    wf G
    -> prink G c k
    -> aprink G c k.
Proof.
exact (aofk_prink_complete ander).
Qed.


Lemma aprink_complete :
  forall G c k,
    wf G
    -> ofc G c k
    -> exists k',
         aprink G c k'
         /\ subk G k' k.
Proof.
intros G c k Hwf Hof.
so (prink_complete _#3 Hwf Hof) as (k' & Hprin & Hsub).
exists k'.
split; auto.
eapply aprink_complete_pre; eauto.
Qed.


Lemma aofc_complete :
  forall G c k,
    wf G
    -> ofc G c k
    -> aofc G c k.
Proof.
intros G c k Hwf Hof.
apply aofc_complete_help; auto.
intros; apply aprink_complete_pre; auto.
Qed.


Lemma aofx_complete :
  (forall G T,
     wf G
     -> ofs G T
     -> aofs G T)
  /\
  (forall G P,
     wf G
     -> ofps G P
     -> aofps G P).
Proof.
exploit
  (ofsig_ind
     (fun G T => wf G -> aofs G T)
     (fun G P => wf G -> aofps G P)) as Hind;
eauto 6 using aofs, aofps, aofk_complete, aofc_complete, wf, ofs_fst.
destruct Hind.
split; intros; eauto.
Qed.


Lemma aofs_complete :
  forall G T,
    wf G
    -> ofs G T
    -> aofs G T.
Proof.
exact (aofx_complete andel).
Qed.


Lemma aofps_complete :
  forall G P,
    wf G
    -> ofps G P
    -> aofps G P.
Proof.
exact (aofx_complete ander).
Qed.


(* Builds in a size measure needed for transitivity. *)
Inductive ssubs : context -> (* sg *) term -> (* sg *) term -> nat -> Prop :=
| ssubs_one {G}
    : ssubs G sg_one sg_one 0

| ssubs_stat {G k k'}
    : ssubk G k k'
      -> ssubs G (sg_stat k) (sg_stat k') 0

| ssubs_dyn {G t t'}
    : eqc G t t' kd_type
      -> ssubs G (sg_dyn t) (sg_dyn t') 0

| ssubs_pi {G T1 T1' k1 k1' T2 T2' m n}
    : ssubs G T1' T1 m
      -> fsts T1' k1'
      -> ssubs (G; cl_cn k1') T2 T2' n
      -> fsts T1 k1
      -> ofs (G; cl_cn k1) T2
      -> ssubs G (sg_pi T1 T2) (sg_pi T1' T2') (S (m + n))

| ssubs_sigma {G T1 T1' k1 k1' T2 T2' m n}
    : ssubs G T1 T1' m
      -> fsts T1 k1
      -> ssubs (G; cl_cn k1) T2 T2' n
      -> fsts T1' k1'
      -> ofs (G; cl_cn k1') T2'
      -> ssubs G (sg_sigma T1 T2) (sg_sigma T1' T2') (S (m + n))

| ssubs_circ {G P P' n}
    : ssubps G P P' n
      -> ssubs G (sg_circ P) (sg_circ P') (S n)


with ssubps : context -> (* ps *) term -> (* ps *) term -> nat -> Prop :=
| ssubps_down {G T T' n}
    : ssubs G T T' n
      -> ssubps G (ps_down T) (ps_down T') (S n)

| ssubps_ex_left {G k P P' n}
    : ofk G k
      -> ssubps (G; cl_cn k) P (shift1 P') n
      -> ofps G P'
      -> ssubps G (ps_ex k P) P' (S n)

| ssubps_ex_right {G P k P' c n}
    : ofc G c k
      -> ssubps G P (subst1 c P') n
      -> ofps (G; cl_cn k) P'
      -> ssubps G P (ps_ex k P') (S n)
.


Scheme ssubs_mut_ind := Minimality for ssubs Sort Prop
with   ssubps_mut_ind := Minimality for ssubps Sort Prop.
Combined Scheme ssubsig_ind from ssubs_mut_ind, ssubps_mut_ind.


Lemma ssubx_impl_subx :
  (forall G T T' n, ssubs G T T' n -> subs G T T')
  /\
  (forall G P P' n, ssubps G P P' n -> subps G P P').
Proof.
exploit
  (ssubsig_ind
     (fun G T T' _ => subs G T T')
     (fun G P P' _ => subps G P P')) as Hind;
eauto using subs, subps;
eauto using ssubk_impl_subk with static.
Qed.


Lemma ssubs_impl_subs :
  forall G T T' n, ssubs G T T' n -> subs G T T'.
Proof.
exact (ssubx_impl_subx andel).
Qed.


Lemma ssubps_impl_subps :
  forall G T T' n, ssubps G T T' n -> subps G T T'.
Proof.
exact (ssubx_impl_subx ander).
Qed.


Lemma subst_ssubs_ssubps :
  (forall G1 G2 s T T' n,
     subof sof G1 s G2
     -> ssubs G2 T T' n
     -> ssubs G1 (subst s T) (subst s T') n)
  /\
  (forall G1 G2 s T T' n,
     subof sof G1 s G2
     -> ssubps G2 T T' n
     -> ssubps G1 (subst s T) (subst s T') n).
Proof.
exploit
  (ssubsig_ind
     (fun G2 T T' n =>
        forall G1 s,
          subof sof G1 s G2
          -> ssubs G1 (subst s T) (subst s T') n)
     (fun G2 T T' n =>
        forall G1 s,
          subof sof G1 s G2
          -> ssubps G1 (subst s T) (subst s T') n)) as H;
try (intros; simp_sub; eauto using ssubs, ssubps, subst_ssubk, subst_eqc_type; done).

(* pi *)
{
intros G T1 T1' k1 k1' T2 T2' m n _ IH1 Hfst' _ IH2 Hfst Hofs G1 s Hsubof.
simp_sub.
eapply ssubs_pi; eauto using subst_fsts, subof_bind_sof_cn, subst_ofs; done.
}

(* sigma *)
{
intros G T1 T1' k1 k1' T2 T2' m n _ IH1 Hfst' _ IH2 Hfst Hofs G1 s Hsubof.
simp_sub.
eapply ssubs_sigma; eauto using subst_fsts, subof_bind_sof_cn, subst_ofs; done.
}

(* ex_left *)
{
intros G k P P' n H H0 IH H2 G1 s H3.
simp_sub.
eapply ssubps_ex_left; eauto using subst_ofk, subst_ofps.
simp_sub.
replace (subst (compose s sh1) P') with (subst (dot (var 0) (compose s sh1)) (shift1 P')) by (simp_sub; reflexivity).
apply IH.
apply subof_bind_sof_cn; auto.
}

(* ex_right *)
{
intros G P k P' c n H H0 IH H2 G1 s H3.
simp_sub.
eapply ssubps_ex_right; eauto using subst_ofc.
  {
  simp_sub.
  replace (subst (dot (subst s c) s) P') with (subst s (subst1 c P')) by (simp_sub; reflexivity).
  apply IH; auto.
  }

  {
  eapply subst_ofps; eauto.
  apply subof_bind_sof_cn; auto.
  }
}

(* wrapup *)
{
destruct H.
split; intros; eauto.
}
Qed.


Lemma subst_ssubs :
  forall G1 G2 s T T' n,
    subof sof G1 s G2
    -> ssubs G2 T T' n
    -> ssubs G1 (subst s T) (subst s T') n.
Proof.
exact (subst_ssubs_ssubps andel).
Qed.


Lemma subst_ssubps :
  forall G1 G2 s T T' n,
    subof sof G1 s G2
    -> ssubps G2 T T' n
    -> ssubps G1 (subst s T) (subst s T') n.
Proof.
exact (subst_ssubs_ssubps ander).
Qed.


Lemma ssubx_refl :
  (forall G T, ofs G T -> exists n, ssubs G T T n)
  /\
  (forall G P, ofps G P -> exists n, ssubps G P P n).
Proof.
exploit
  (ofsig_ind
     (fun G T => exists n, ssubs G T T n)
     (fun G T => exists n, ssubps G T T n)) as Hind;
try (intros;
     repeat
       (match goal with
        | H : exists _, _ |- _ => destruct H
        end);
     eauto using ssubs, ssubps, ssubk_refl, eqc_refl;
     done).
(* ex *)
{
intros G k P Hofk HofP (n & HsubP).
exists (2 + n).
apply ssubps_ex_left; eauto with static.
simp_sub.
apply (@ssubps_ex_right _#4 (var 0)).
  {
  apply ofc_var; auto with sequence.
  }

  {
  simp_sub.
  auto.
  }

  {
  eapply subst_ofps; eauto.
  apply subof_dot; eauto using subof_sh with sequence.
  simp_sub.
  apply sof_cn.
  eapply rel_quest.
    {
    eapply ofc_var; eauto with sequence.
    }
  simp_sub; reflexivity.
  }
}
Qed.


Lemma ssubx_trans :
  (forall G T1 T2 T3 m n,
     wf G
     -> ssubs G T1 T2 m
     -> ssubs G T2 T3 n
     -> exists p, ssubs G T1 T3 p)
  /\
  (forall G T1 T2 T3 m n,
     wf G
     -> ssubps G T1 T2 m
     -> ssubps G T2 T3 n
     -> exists p, ssubps G T1 T3 p).
Proof.
cut (forall q m n,
       m + n = q
       -> (forall G T1 T2 T3,
             wf G
             -> ssubs G T1 T2 m
             -> ssubs G T2 T3 n
             -> exists p, ssubs G T1 T3 p)
          /\
          (forall G T1 T2 T3,
             wf G
             -> ssubps G T1 T2 m
             -> ssubps G T2 T3 n
             -> exists p, ssubps G T1 T3 p)).
  {
  intros Hind.
  split.
    {
    intros G T1 T2 T3 m n Hwf H12 H23.
    eapply (Hind _ m n (eq_refl _) andel); eauto.
    }

    {
    intros G T1 T2 T3 m n Hwf H12 H23.
    eapply (Hind _ m n (eq_refl _) ander); eauto.
    }
  }
intro q.
wfinduct q using lt_wf.
intros q IH m n <-.
split.
  {
  intros G T1 T2 T3 Hwf H12 H23.
  revert n T3 IH Hwf H23.
  case H12; clear codependent H12.
  
    (* one *)
    {
    intros G n T3 _ Hwf H23.
    invert H23.
    intros <- <-.
    eauto.
    }
  
    (* stat *)
    {
    intros G k1 k2 Hk12 n T3 _ Hwf HT23.
    invert HT23.
    intros k3 Hk23 <- <-.
    exists 0.
    apply ssubs_stat.
    eapply ssubk_trans; eauto.
    }
  
    (* dyn *)
    {
    intros G t1 t2 Ht12 n T3 _ Hwf HT23.
    invert HT23.
    intros t3 Ht23 <- <-.
    exists 0.
    apply ssubs_dyn.
    eapply eqc_trans; eauto.
    }
  
    (* pi *)
    {
    intros G T1 T2 k1 k2 U1 U2 m1 m2 HT21 Hfst2 HU12 Hfst1 HofU1 n TU3 IH Hwf HTU23.
    invert HTU23.
    intros T3 k2a k3 U3 n1 n2 HT32 Hfst3 HU23 Hfst2a HofU2 <- <-.
    so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a.
    so (subs_fst _#5 Hwf (ssubs_impl_subs _#4 HT32) Hfst3 Hfst2) as Hsubk32.
    so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 HT32)) as (HofT3 & _).
    so (ofs_fst _#3 HofT3 Hfst3) as Hofk3.
    exploit (IH (n1 + m1)) as IH1; [omega |].
    so (IH1 _ _ (eq_refl _) andel _#4 Hwf HT32 HT21) as (p & HsubT31).
    so (subst_ssubs _#6 (subof_subsume_0 _#3 Hsubk32) HU12) as HU12w.
    simp_sub in HU12w.
    exploit (IH (m2 + n2)) as IH2; [omega |].
    so (IH2 _ _ (eq_refl _) andel _#4 (wf_cn Hwf Hofk3) HU12w HU23) as (q & HsubU13).
    eexists.
    eapply ssubs_pi; eauto.
    }
  
    (* sigma *)
    {
    intros G T1 T2 k1 k2 U1 U2 m1 m2 HT12 Hfst1 HU12 Hfst2 HofU2 n TU3 IH Hwf HTU23.
    invert HTU23.
    intros T3 k2a k3 U3 n1 n2 HT23 Hfst2a HU23 Hfst3 HofU3 <- <-.
    so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a.
    so (subs_fst _#5 Hwf (ssubs_impl_subs _#4 HT12) Hfst1 Hfst2) as Hsubk12.
    so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 HT12)) as (HofT1 & _).
    so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
    exploit (IH (m1 + n1)) as IH1; [omega |].
    so (IH1 _ _ (eq_refl _) andel _#4 Hwf HT12 HT23) as (p & HsubT13).
    so (subst_ssubs _#6 (subof_subsume_0 _#3 Hsubk12) HU23) as HU23w.
    simp_sub in HU23w.
    exploit (IH (m2 + n2)) as IH2; [omega |].
    so (IH2 _ _ (eq_refl _) andel _#4 (wf_cn Hwf Hofk1) HU12 HU23w) as (q & HsubU13).
    eexists.
    eapply ssubs_sigma; eauto.
    }
  
    (* circ *)
    {
    intros G P1 P2 m H12 n T3 IH Hwf H23.
    invert H23.
    intros P3 n' H23' <- <-.
    exploit (IH (m + n')) as IH'; [omega |].
    so (IH' _ _ (eq_refl _) ander _#4 Hwf H12 H23') as (p & H13).
    exists (S p).
    apply ssubs_circ; auto.
    }
  }

  {
  assert (forall m' n',
            m' + n' < m + n
            -> forall G P1 P2 P3 k c,
                 wf G
                 -> ofc G c k
                 -> ssubps G P1 P2 m'
                 -> ssubps G P2 (subst1 c P3) n'
                 -> ofps (G; cl_cn k) P3
                 -> exists p, ssubps G P1 (ps_ex k P3) p) as Hright.
    {
    intros m' n' Hlt G P1 P2 P3 k c Hwf Hofc H12 H23 HofP3.
    exploit (IH (m' + n')) as IH'; [omega |].
    so (IH' _ _ (eq_refl _) ander _#4 Hwf H12 H23) as (p & H13).
    exists (S p).
    eapply ssubps_ex_right; eauto.
    }
  intros G P1 P2 P3 Hwf H12 H23.
  revert n P3 IH Hwf Hright H23.
  case H12; clear codependent H12.

    (* down *)
    {
    intros G T1 T2 m H12 n P3 IH Hwf Hright H23.
    invert H23.
      (* down *)
      {
      intros T3 n' H23' <- <-.
      exploit (IH (m + n')) as IH'; [omega |].
      so (IH' _ _ (eq_refl _) andel _#4 Hwf H12 H23') as (p & H13).
      exists (S p).
      apply ssubps_down; auto.
      }
  
      (* ex_right *)
      {
      intros k P3' c n' Hofc H23' HofP3' <- <-.
      exploit (Hright (S m) n') as Hright'; [omega |].
      eapply Hright'; eauto.
      apply ssubps_down; auto.
      }
    }
  
    (* ex_left *)
    {
    intros G k P1 P2 m Hofk H12 HofP2 n P3 IH Hwf _ H23.
    exploit (IH (m + n)) as IH'; [omega |].
    so (subst_ssubps _#6 (weaken1 _ _ (cl_cn k)) H23) as H23w.
    so (IH' _ _ (eq_refl _) ander _#4 (wf_cn Hwf Hofk) H12 H23w) as (p & H13).
    exists (S p).
    apply ssubps_ex_left; auto.
    exact (subps_reg _#3 Hwf (ssubps_impl_subps _#4 H23) ander).
    }
  
    (* ex_right *)
    {
    intros G P1 k P2 c m Hofc H12 HofP2 n P3 IH Hwf Hright Hsubex.
    invert Hsubex.
      (* ex_left *)
      {
      intros n' Hofk H23 HofP3 <-.
      eassert Prop as H23x; [refine (subst_ssubps G _ (dot c id) _#3 _ H23) |].
        {
        apply subof_dot; auto using subof_id.
        simp_sub.
        apply sof_cn; auto.
        }
      exploit (IH (m + n')) as IH'; [omega |].
      so (IH' _ _ (eq_refl _) ander _#4 Hwf H12 H23x) as (p & H13).        
      simp_sub in H13.
      eauto.
      }
  
      (* ex_right *)
      {
      intros l P3' d n' Hofd H23 HofP3' <- <-.
      exploit (Hright (S m) n') as Hright'; [omega |].
      eapply Hright'; eauto.
      eapply ssubps_ex_right; eauto.
      }
    }
  }
Qed.


Lemma eqx_impl_ssubx :
  (forall G T T',
     wf G
     -> eqs G T T'
     -> exists n, ssubs G T T' n)
  /\
  (forall G T T',
     wf G
     -> eqps G T T'
     -> exists n, ssubps G T T' n).
Proof.
exploit
  (eqsig_ind
     (fun G T T' => wf G -> (exists n, ssubs G T T' n) /\ (exists n, ssubs G T' T n))
     (fun G T T' => wf G -> (exists n, ssubps G T T' n) /\ (exists n, ssubps G T' T n))) as Hind.

(* refl *)
{
intros G T HofT Hwf.
so (ssubx_refl andel _ _ HofT) as (n & Hsub).
eauto.
}

(* symm *)
{
intros G T T' _ IH Hwf.
so (IH Hwf) as (n & H & H').
eauto.
}

(* trans *)
{
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as ((m & H12) & (m' & H21)).
so (IH2 Hwf) as ((n & H23) & (n' & H32)).
so (ssubx_trans andel _#6 Hwf H12 H23) as (p & H13).
so (ssubx_trans andel _#6 Hwf H32 H21) as (p' & H31).
eauto.
}

(* stat *)
{
intros G k k' Heqk Hwf.
split; eauto using ssubs_stat, eqk_impl_ssubk, eqk_symm.
}

(* dyn *)
{
intros G t t' Heqc Hwf.
split; eauto using ssubs_dyn, eqc_symm.
}

(* pi *)
{
intros G T1 T1' k1 T2 T2' Heq1 IH1 Hfst _ IH2 Hwf.
so (eqs_reg _#3 Hwf Heq1) as (HofT1 & HofT1').
so (ofs_fst _#3 HofT1 Hfst) as Hofk1.
so (can_fsts _ _ HofT1') as (k1' & Hfst').
so (ofs_fst _#3 HofT1' Hfst') as Hofk1'.
so (IH1 Hwf) as ((m & Hsub1) & (m' & Hsub1')).
so (IH2 (wf_cn Hwf Hofk1)) as ((n & Hsub2), (n' & Hsub2')).
so (subs_fst _#5 Hwf (ssubs_impl_subs _#4 Hsub1') Hfst' Hfst) as Hsubk.
split.
  {
  exists (S (m' + n)).
  apply (@ssubs_pi _ _ _ k1 k1'); auto.
    {
    so (subst_ssubs _#6 (subof_subsume_0 _#3 Hsubk) Hsub2) as H.
    simp_sub in H.
    exact H.
    }

    {
    exact (subs_reg _#3 (wf_cn Hwf Hofk1) (ssubs_impl_subs _#4 Hsub2) andel).
    }
  }

  {
  exists (S (m + n')).
  apply (@ssubs_pi _ _ _ k1' k1); auto.
  so (subst_ofs _#4 (subof_subsume_0 _#3 Hsubk) (subs_reg _#3 (wf_cn Hwf Hofk1) (ssubs_impl_subs _#4 Hsub2) ander)) as H.
  simp_sub in H.
  exact H.
  }
}

(* sigma *)
{
intros G T1 T1' k1 T2 T2' Heq1 IH1 Hfst _ IH2 Hwf.
so (eqs_reg _#3 Hwf Heq1) as (HofT1 & HofT1').
so (ofs_fst _#3 HofT1 Hfst) as Hofk1.
so (can_fsts _ _ HofT1') as (k1' & Hfst').
so (ofs_fst _#3 HofT1' Hfst') as Hofk1'.
so (IH1 Hwf) as ((m & Hsub1) & (m' & Hsub1')).
so (IH2 (wf_cn Hwf Hofk1)) as ((n & Hsub2), (n' & Hsub2')).
so (subs_fst _#5 Hwf (ssubs_impl_subs _#4 Hsub1') Hfst' Hfst) as Hsubk.
split.
  {
  exists (S (m + n)).
  apply (@ssubs_sigma _ _ _ k1 k1'); auto.
  so (subst_ofs _#4 (subof_subsume_0 _#3 Hsubk) (subs_reg _#3 (wf_cn Hwf Hofk1) (ssubs_impl_subs _#4 Hsub2) ander)) as H.
  simp_sub in H.
  exact H.
  }

  {
  exists (S (m' + n')).
  apply (@ssubs_sigma _ _ _ k1' k1); auto.
    {
    so (subst_ssubs _#6 (subof_subsume_0 _#3 Hsubk) Hsub2') as H.
    simp_sub in H.
    exact H.
    }

    {
    exact (subs_reg _#3 (wf_cn Hwf Hofk1) (ssubs_impl_subs _#4 Hsub2) andel).
    }
  }
}

(* circ *)
{
intros G P P' _ IH Hwf.
so (IH Hwf) as ((n, Hsub), (n', Hsub')).
split; eauto using ssubs_circ.
}

(* down *)
{
intros G T T' _ IH Hwf.
so (IH Hwf) as ((n, Hsub), (n', Hsub')).
split; eauto using ssubps_down.
}

(* ex *)
{
intros G k k' P P' Heqk _ IH Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (IH (wf_cn Hwf Hofk)) as ((n, Hsub), (n', Hsub')).
so (subps_reg _#3 (wf_cn Hwf Hofk) (ssubps_impl_subps _#4 Hsub)) as (HofP & HofP').
split.
  {
  exists (2 + n).
  apply ssubps_ex_left; auto.
    {
    simp_sub.
    apply (@ssubps_ex_right _#4 (var 0)).
      {
      eapply ofc_subsume; eauto using ofc_var with sequence.
      apply weaken1_subk.
      apply subk_refl; auto.
      }

      {
      simp_sub.
      auto.
      }

      {
      eapply subst_ofps; eauto.
      apply subof_dot; eauto using subof_sh with sequence.
      simp_sub.
      apply sof_cn.
      eapply ofc_subsume; eauto using ofc_var with sequence.
      simp_sub.
      apply weaken2_subk.
      apply subk_refl.
      apply eqk_symm; auto.
      }
    }

    {
    apply ofps_ex; auto.
    so (subst_ofps _#4 (subof_conv_0 _#3 (eqk_symm Heqk)) HofP') as H.
    simp_sub in H.
    exact H.
    }
  }

  {
  exists (2 + n').
  apply ssubps_ex_left; auto.
    {
    simp_sub.
    apply (@ssubps_ex_right _#4 (var 0)).
      {
      eapply ofc_subsume; eauto using ofc_var with sequence.
      apply weaken1_subk.
      apply subk_refl.
      apply eqk_symm; auto.
      }

      {
      simp_sub.
      so (subst_ssubps _#6 (subof_conv_0 _#3 (eqk_symm Heqk)) Hsub') as H.
      simp_sub in H.
      exact H.
      }

      {
      eapply subst_ofps; eauto.
      apply subof_dot; eauto using subof_sh with sequence.
      simp_sub.
      apply sof_cn.
      eapply rel_quest; eauto using ofc_var with sequence.
      simp_sub; reflexivity.
      }
    }

    {
    apply ofps_ex; auto.
    }
  }
}

(* epilogue *)
{
destruct Hind as (Hind & Hindp).    
split.
  {
  intros G T T' Hwf Heq.
  so (Hind _#3 Heq Hwf andel) as (n & H).
  eauto.
  }

  {
  intros G T T' Hwf Heq.
  so (Hindp _#3 Heq Hwf andel) as (n & H).
  eauto.
  }
}
Qed.


Lemma subx_impl_ssubx :
  (forall G T T',
     wf G
     -> subs G T T'
     -> exists n, ssubs G T T' n)
  /\
  (forall G T T',
     wf G
     -> subps G T T'
     -> exists n, ssubps G T T' n).
Proof.
exploit
  (subsig_ind
     (fun G T T' =>
        wf G
        -> exists n, ssubs G T T' n)
     (fun G T T' =>
        wf G
        -> exists n, ssubps G T T' n)) as Hind.

(* refl *)
{
intros G T T' Heq Hwf.
exact (eqx_impl_ssubx andel _#3 Hwf Heq).
}

(* trans *)
{
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (m & H12).
so (IH2 Hwf) as (n & H23).
eapply (ssubx_trans andel); eauto.
}

(* stat *)
{
intros G k k' Hsubk Hwf.
exists 0.
apply ssubs_stat.
apply subk_impl_ssubk; auto.
}

(* pi *)
{
intros G T1 T1' k1 k1' T2 T2' _ IH1 Hfst' _ IH2 Hfst HofT2 Hwf.
so (IH1 Hwf) as (m & Hsub1).
so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 Hsub1)) as (HofT1' & _).
so (ofs_fst _#3 HofT1' Hfst') as Hofk1'.
so (IH2 (wf_cn Hwf Hofk1')) as (n & Hsub2).
exists (S (m + n)).
eapply ssubs_pi; eauto.
}

(* sigma *)
{
intros G T1 T1' k1 k1' T2 T2' _ IH1 Hfst _ IH2 Hfst' HofT2 Hwf.
so (IH1 Hwf) as (m & Hsub1).
so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 Hsub1)) as (HofT1 & _).
so (ofs_fst _#3 HofT1 Hfst) as Hofk1.
so (IH2 (wf_cn Hwf Hofk1)) as (n & Hsub2).
exists (S (m + n)).
eapply ssubs_sigma; eauto.
}

(* circ *)
{
intros G P P' _ IH Hwf.
so (IH Hwf) as (n & Hsub).
exists (S n).
apply ssubs_circ; auto.
}

(* refl *)
{
intros G P P' Heq Hwf.
apply (eqx_impl_ssubx ander); auto.
}

(* trans *)
{
intros G P1 P2 P3 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (m & H12).
so (IH2 Hwf) as (n & H23).
eapply (ssubx_trans ander); eauto.
}

(* down *)
{
intros G T T' _ IH Hwf.
so (IH Hwf) as (n & Hsub).
exists (S n).
apply ssubps_down; auto.
}

(* ex_left *)
{
intros G k P P' Hofk _ IH HofP' Hwf.
so (IH (wf_cn Hwf Hofk)) as (n & Hsub).
exists (S n).
apply ssubps_ex_left; eauto.
}

(* ex_right *)
{
intros G P k P' c Hofc _ IH HofP' Hwf.
so (IH Hwf) as (n & Hsub).
exists (S n).
eapply ssubps_ex_right; eauto.
}

(* epilogue *)
{
destruct Hind as (Hind, Hindp).
split; eauto.
}
Qed.


Lemma subs_impl_ssubs :
  forall G T T',
    wf G
    -> subs G T T'
    -> exists n, ssubs G T T' n.
Proof.
exact (subx_impl_ssubx andel).
Qed.


Lemma subps_impl_ssubps :
  forall G T T',
    wf G
    -> subps G T T'
    -> exists n, ssubps G T T' n.
Proof.
exact (subx_impl_ssubx ander).
Qed.


Lemma subst_syn_ana :
  (forall s T, syn T -> syn (subst s T))
  /\
  (forall s P, synp P -> synp (subst s P))
  /\
  (forall s T, ana T -> ana (subst s T))
  /\
  (forall s P, anap P -> anap (subst s P)).
Proof.
exploit
  (syn_ana_ind
     (fun T => forall s, syn (subst s T))
     (fun P => forall s, synp (subst s P))
     (fun T => forall s, ana (subst s T))
     (fun P => forall s, anap (subst s P))) as Hind;
try (intros; simp_sub; eauto using syn, synp, ana, anap).
destruct_all Hind; auto.
Qed.


Lemma subst_syn :
  forall s T, syn T -> syn (subst s T).
Proof.
exact (subst_syn_ana andel).
Qed.


Lemma subst_synp :
  forall s P, synp P -> synp (subst s P).
Proof.
exact (subst_syn_ana anderl).
Qed.


Lemma subst_ana :
  forall s P, ana P -> ana (subst s P).
Proof.
exact (subst_syn_ana anderrl).
Qed.


Lemma subst_anap :
  forall s P, anap P -> anap (subst s P).
Proof.
exact (subst_syn_ana anderrr).
Qed.


Lemma asubx_complete :
  (forall G T T',
     wf G
     -> syn T
     -> ana T'
     -> subs G T T'
     -> asubs G T T')
  /\
  (forall G P P',
     wf G
     -> synp P
     -> anap P'
     -> subps G P P'
     -> asubps G P P').
Proof.
exploit
  (ssubsig_ind
     (fun G T T' n =>
        wf G
        -> syn T
        -> ana T'
        -> asubs G T T')
     (fun G T T' n =>
        wf G
        -> synp T
        -> anap T'
        -> asubps G T T')) as Hind;
eauto using asubs, asubps.

(* stat *)
{
intros G k k' Hsub Hwf _ _.
apply asubs_stat.
apply asubk_complete; auto.
apply ssubk_impl_subk; auto.
}

(* dyn *)
{
intros G t t' Heqt Hwf _ _.
apply asubs_dyn.
apply aeq_complete; auto.
}

(* pi *)
{
intros G T1 T1' k1 k1' T2 T2' m n Hsub1 IH1 Hfst' _ IH2 Hfst HofT2 Hwf Hsyn Hana.
invert Hsyn.
intros Hana1 Hsyn2.
invert Hana.
intros Hsyn1 Hana2.
so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 Hsub1)) as (HofT1' & _).
so (ofs_fst _#3 HofT1' Hfst') as Hofk1'.
eapply asubs_pi; eauto using wf.
}

(* sigma *)
{
intros G T1 T1' k1 k1' T2 T2' m n Hsub1 IH1 Hfst _ IH2 Hfst' HofT2 Hwf Hsyn Hana.
invert Hsyn.
intros Hsyn1 Hsyn2.
invert Hana.
intros Hana1 Hana2.
so (subs_reg _#3 Hwf (ssubs_impl_subs _#4 Hsub1)) as (HofT1 & _).
so (ofs_fst _#3 HofT1 Hfst) as Hofk1.
eapply asubs_sigma; eauto using wf.
}

(* circ *)
{
intros G P P' n _ IH Hwf Hsyn Hana.
invert Hsyn.
intros Hsyn'.
invert Hana.
intros Hana'.
apply asubs_circ; auto.
}

(* down *)
{
intros G P P' n _ IH Hwf Hsyn Hana.
invert Hsyn.
intros Hsyn'.
invert Hana.
intros Hana'.
apply asubps_down; auto.
}

(* ex_left *)
{
intros G k P P' n Hofk _ IH HofP' Hwf Hsyn Hana.
invert Hsyn.
intros Hsyn'.
apply asubps_ex_left; auto.
apply IH; auto using wf.
apply subst_anap; auto.
}

(* ex_right *)
{
intros G P k P' c _ _ _ _ _ _ _ Hana.
invert Hana.
}

(* epilogue *)
{
destruct Hind as (Hind & Hindp).
split.
  {
  intros G T T' Hwf Hsyn Hana Hsub.
  so (subs_impl_ssubs _#3 Hwf Hsub) as (n & Hsub').
  eapply Hind; eauto.
  }

  {
  intros G T T' Hwf Hsyn Hana Hsub.
  so (subps_impl_ssubps _#3 Hwf Hsub) as (n & Hsub').
  eapply Hindp; eauto.
  }
}
Qed.


Lemma asubs_complete :
  forall G T T',
    wf G
    -> syn T
    -> ana T'
    -> subs G T T'
    -> asubs G T T'.
Proof.
exact (asubx_complete andel).
Qed.


Lemma asubps_complete :
  forall G P P',
    wf G
    -> synp P
    -> anap P'
    -> subps G P P'
    -> asubps G P P'.
Proof.
exact (asubx_complete ander).
Qed.


Lemma synctx_append :
  forall G1 G2,
    synctx G1
    -> synctx G2
    -> synctx (G1;; G2).
Proof.
intros G1 G2 H1 H2.
induct H2; simpl; auto using synctx.
Qed.


Lemma subst_noex :
  forall s T, noex T -> noex (subst s T).
Proof.
exploit
  (noexx_ind
     (fun T => forall s, noex (subst s T))
     (fun T => forall s, noexp (subst s T))) as Hind;
try (intros; simp_sub; eauto using noex, noexp; done).
destruct_all Hind; auto.
Qed.


Lemma noexx_syn_ana :
  (forall T, noex T -> syn T /\ ana T)
  /\
  (forall T, noexp T -> synp T /\ anap T).
Proof.
exploit
  (noexx_ind
     (fun T => syn T /\ ana T)
     (fun P => synp P /\ anap P)) as Hind;
intros; 
repeat
  (match goal with
   | H : _ /\ _ |- _ => destruct H
   end);
eauto using syn, ana, synp, anap.
Qed.


Lemma noex_syn :
  forall T, noex T -> syn T.
Proof.
intros T HT.
exact (noexx_syn_ana andel _ HT andel).
Qed.


Lemma noex_ana :
  forall T, noex T -> ana T.
Proof.
intros T HT.
exact (noexx_syn_ana andel _ HT ander).
Qed.


Lemma singles_syn :
  forall c T,
    syn T
    -> syn (singles c T).
Proof.
intros c T H.
revert c.
induct H; simpl; eauto using syn, subst_syn.
Qed.


Lemma linv_syn :
  forall P G T,
    linv P G T
    -> synp P
    -> synctx G /\ syn T.
Proof.
intros P G T Hinv HP.
revert HP.
induct Hinv.

(* down *)
{
intros T HP.
invert HP; auto using synctx.
}

(* ex *)
{
intros k P G T _ IH Hex.
invert Hex.
intros HP.
so (IH HP) as (HG & HT).
split; auto.
apply synctx_append; auto using synctx.
}
Qed.


Lemma close_syn :
  forall G P1 P2,
    close G P1 P2
    -> synp P1
    -> synp P2.
Proof.
intros G P1 P2 Hclose HP1.
revert HP1.
induct Hclose; eauto using synp.
Qed.


Definition allvar s := forall i, exists j, project s i = var j.


Lemma allvar_sh : forall i, allvar (sh i).
Proof.
intros i j.
simp_sub.
eauto.
Qed.


Lemma allvar_dot : forall i s, allvar s -> allvar (dot (var i) s).
Proof.
intros i s Hs j.
destruct j as [| j].
  {
  simp_sub.
  eauto.
  }

  {
  simp_sub.
  so (Hs j) as (k & Heq).
  rewrite -> Heq.
  eauto.
  }
Qed.


Lemma allvar_compose :
  forall s s',
    allvar s
    -> allvar s'
    -> allvar (compose s s').
Proof.
intros s s' Hs Hs'.
intro i.
rewrite -> project_compose.
so (Hs i) as (i' & Heq).
rewrite -> Heq.
simp_sub.
apply Hs'.
Qed.


Lemma allvar_under :
  forall i s, allvar s -> allvar (under i s).
Proof.
intros i s Hs.
intro j.
so (lt_dec j i) as [Hlt | Hnlt].
  {
  rewrite -> project_under_lt; auto.
  eauto.
  }

  {
  rewrite -> project_under_ge; [| omega].
  so (Hs (j - i)) as (k & Heq).
  rewrite -> Heq.
  simp_sub.
  eauto.
  }
Qed.


Lemma allvar_id : allvar id.
Proof.
exact (allvar_sh 0).
Qed.


Lemma allvar_bind :
  forall s, allvar s -> allvar (dot (var 0) (compose s sh1)).
Proof.
intros s Hs.
eauto using allvar_dot, allvar_sh, allvar_compose.
Qed.


Lemma allvar_bind2 :
  forall s, allvar s -> allvar (dot (var 0) (dot (var 1) (compose s (sh 2)))).
Proof.
intros s Hs.
eauto using allvar_dot, allvar_sh, allvar_compose.
Qed.


Lemma subst_user :
  forall s m, allvar s -> user m -> user (subst s m).
Proof.
intros s m Hs H.
revert s Hs.
induct H;
try (intros; simp_sub; eauto using user, allvar_bind, allvar_bind2, subst_noex; done).
(* var *)
{
intros i s Hs.
simp_sub.
so (Hs i) as (j & Heq).
rewrite -> Heq.
apply user_var.
}
Qed.


Lemma aprinx_syn :
  (forall G M T,
     synctx G
     -> user M
     -> aprins G M T
     -> syn T)
  /\
  (forall G M P,
     synctx G
     -> user M
     -> aprinps G M P
     -> synp P).
Proof.
exploit
  (aprin_ind
     (fun G e t => True)
     (fun G M T =>
        synctx G
        -> user M
        -> syn T)
     (fun G M P =>
        synctx G
        -> user M
        -> synp P)) as Hind;
try (intros;
     trivial;
     match goal with
     | H : user _ |- _ => invert H; clear H; intros; subst
     end;
     eauto using syn;
     done).

(* var *)
{
intros G i T c Hindex HG _.
apply singles_syn.
apply subst_syn.
remember (cl_md T c) as A eqn:Heq.
revert HG Heq.
induct Hindex.
  (* 0 *)
  {
  intros A G HG ->.
  invert HG; auto.
  }

  (* S *)
  {
  intros i A G ? _ IH HG ->.
  apply IH; auto.
  invert HG; auto.
  }
}

(* pi *)
{
intros G T1 T2 k1 M _ _ _ IH HG Huser.
invert Huser.
intros HT1 HM.
apply syn_pi; auto using noex_ana.
so (IH (synctx_md (synctx_cn HG) (subst_syn _ _ (noex_syn _ HT1))) HM) as H.
so (subst_syn (dot nonsense id) _ H) as H'.
simp_sub in H'.
exact H'.
}

(* app *)
{
intros G T1 T2 _ M1 M2 c2 _ IH _ _ _ _ HG Huser.
invert Huser.
intros HM1 _.
invert (IH HG HM1).
intros _ HT2.
apply subst_syn; auto.
}

(* pair *)
{
intros G T1 T2 M1 M2 _ IH1 _ IH2 HG Huser.
invert Huser.
intros HM1 HM2.
so (IH1 HG HM1) as HT1.
so (IH2 HG HM2) as HT2.
apply syn_sigma; auto.
apply subst_syn; auto.
}

(* pi1 *)
{
intros G T1 T2 M _ IH HG Huser.
invert Huser.
intro HM.
invert (IH HG HM).
intros HT1 HT2.
auto.
}

(* pi2 *)
{
intros G T1 T2 M c _ IH _ HG Huser.
invert Huser.
intro HM.
invert (IH HG HM).
intros HT1 HT2.
apply subst_syn; auto.
}

(* lett *)
{
intros G t T e M _ _ _ IH HG Huser.
invert Huser.
intros _ HM.
apply subst_syn.
apply IH; auto using synctx_tm.
}

(* down *)
{
intros G M T _ IH HG Huser.
invert Huser.
intros HM.
apply synp_down.
apply IH; auto.
}

(* seal *)
{
intros G M T T' _ _ _ _ HG Huser.
invert Huser.
intros _ HT.
apply synp_down.
apply noex_syn; auto.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P2' _ IH1 Hinv _ _ IH2 Hclose HG Huser.
invert Huser.
intros HM1 HM2.
so (IH1 HG HM1) as Hsyncirc.
invert Hsyncirc.
intros HP1.
so (linv_syn _#3 Hinv HP1) as (HG' & HT).
so (IH2 (synctx_md (synctx_cn (synctx_append _ _ HG HG')) (subst_syn _ _ HT)) (subst_user _ _ (allvar_under _ _ (allvar_sh _)) HM2)) as HP2.
eapply close_syn; eauto.
apply synp_ex.
apply subst_synp; auto.
}

(* unpack *)
{
intros G u k t P e M _ _ _ _ IH HG Huser.
invert Huser.
intros _ HM.
so (IH (synctx_tm (synctx_cn HG)) HM) as HP.
apply synp_ex.
apply subst_synp; auto.
}

(* epilogue *)
{
destruct_all Hind.
split; intros; eauto.
}
Qed.


Lemma aprins_syn :
  forall G M T,
    synctx G
    -> user M
    -> aprins G M T
    -> syn T.
Proof.
exact (aprinx_syn andel).
Qed.


Lemma aprinps_syn :
  forall G M T,
    synctx G
    -> user M
    -> aprinps G M T
    -> synp T.
Proof.
exact (aprinx_syn ander).
Qed.


Lemma close_snoc :
  forall k G P1 P2,
    close G P1 P2
    -> close (nil; cl_cn k;; G) P1 (ps_ex k P2).
Proof.
intros k G P1 P2 Hclose.
induct Hclose.

(* nil *)
{
intros P.
simpl.
auto using close.
}

(* cons *)
{
intros G l P1 P2 _ IH.
autorewrite with canonlist.
apply close_cons; auto.
}
Qed.


Lemma lclose_impl_close :
  forall P1 G T P2 P3,
    linv P1 G T
    -> lclose P1 P2 P3
    -> close G P2 P3.
Proof.
intros P1 G T P2 P3 Hinv Hclose.
revert P2 P3 Hclose.
induct Hinv.

(* down *)
{
intros T P2 P3 Hclose.
invert Hclose.
intros <-.
apply close_nil.
}

(* ex *)
{
intros k P G T _ IH P2 P3 Hclose.
invert Hclose.
intros P3' Hclose' <-.
auto using close_snoc.
}
Qed.


Lemma aprinx_complete :
  (forall G e t,
     wf G
     -> synctx G
     -> user e
     -> print G e t
     -> aprint G e t)
  /\
  (forall G M T,
     wf G
     -> synctx G
     -> user M
     -> prins G M T
     -> aprins G M T)
  /\
  (forall G M P,
     wf G
     -> synctx G
     -> user M
     -> prinps G M P
     -> aprinps G M P).
Proof.
exploit
  (prin_ind
     (fun G e t =>
        wf G 
        -> synctx G
        -> user e
        -> aprint G e t)
     (fun G M T =>
        wf G 
        -> synctx G
        -> user M
        -> aprins G M T)
     (fun G M P => 
        wf G
        -> synctx G
        -> user M
        -> aprinps G M P)) as Hind;
intros;
try match goal with
| H : user _ |- _ => invert H; clear H; intros
end;
revert_all;
eauto 9 using aprint, aprins, aprinps, aofk_complete, aofc_complete, aprink_complete_pre, aofs_complete, wf, synctx, noex_syn, subst_syn.

(* app *)
{
intros G t t1 t2 t1' e1 e2 Haprin1 IH1 Hnorm Haprin2 IH2 Heqc Hwf HG He1 He2.
eapply aprint_app; eauto.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Haprin1)) as Hoft.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Haprin2)) as Hoft1'.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hoft Hnorm) anderl)) as (l & H).
invert H.
intros Hoft1 _ _.
apply aeqc_complete; eauto.
}

(* pack *)
{
intros G k t c e u Hofc _ IH Heq Hoft Hwf HG He.
so (ofc_reg _#3 Hwf Hofc) as Hofk.
so (aofc_complete _#3 (wf_cn Hwf Hofk) Hoft) as Haoft.
eapply aprint_pack; eauto using aofk_complete, aofc_complete, wf, ofc_reg.
eapply aeqc_complete; eauto.
}

(* unpack *)
{
intros G u k t t' e1 e2 Hprin1 IH1 Hnorm Hprin2 IH2 Heq Hwf HG He1 He2.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hofu.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (oft_reg _#3 Hwf' (print_sound _#3 Hwf' Hprin2)) as Hoft'.
eapply aprint_unpack; eauto using synctx.
eapply aeqc_complete; eauto using wf.
}

(* fix *)
{
intros G t e u Hprin IH Heq Hwf HG He.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf Heq anderl)) as (l & H).
invert H.
intros _ Hoft _.
eapply aprint_fix; eauto using aofc_complete.
apply aeqc_complete; eauto with static.
}

(* lett *)
{
intros G t1 t2 e1 e2 Hprin1 IH1 _ IH2 Hwf HG He1 He2.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hoft1.
eapply aprint_lett; eauto using wf, synctx.
}

(* letm *)
{
intros G P G' T k t M e Hprin1 IH1 Hinv Hfst Hprin2 IH2 Heq Hwf HG HM He.
so (oflm_reg _#3 Hwf (prinps_sound _#3 Hwf Hprin1)) as HofP.
so (linv_reg _#4 Hwf HofP Hinv) as (Hwf' & HofT).
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (IH1 Hwf HG HM) as Haprin1.
so (aprinps_syn _#3 HG HM Haprin1) as HP.
so (linv_syn _#3 Hinv HP) as (HG' & HT).
exploit IH2; eauto using wf, synctx, synctx_append, subst_syn.
  {
  apply subst_user; auto.
  apply allvar_under.
  apply allvar_sh.
  }
eapply aprint_letm; eauto.
eapply aeqc_complete; eauto using wf.
}

(* app *)
{
intros G T1 T2 T1' M1 M2 c2 Hprin1 IH1 Hprin2 IH2 Hsub Hfst Hwf HG HM1 HM2.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin1)) as Hofpi.
invert Hofpi.
intros _ HofT1 _ _.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin2)) as HofT1'.
so (IH1 Hwf HG HM1) as Haprin1.
so (IH2 Hwf HG HM2) as Haprin2.
eapply aprins_app; eauto.
apply asubs_complete; eauto using aprins_syn.
so (aprins_syn _#3 HG HM1 Haprin1) as Hsynpi.
invert Hsynpi.
auto.
}

(* lett *)
{
intros G t T e M Hprin1 IH1 _ IH2 Hwf HG He HM.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hoft.
eapply aprins_lett; eauto using wf, synctx.
}

(* seal *)
{
intros G M T2 T1 _ IH1 Hsub Hwf HG HM HT2.
so (subs_reg _#3 Hwf Hsub) as (HofT1 & HofT2).
so (IH1 Hwf HG HM) as Hprin.
so (aprins_syn _#3 HG HM Hprin) as HsynT1.
eapply aprinps_seal; eauto using aofs_complete.
apply asubs_complete; auto using noex_ana.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P2' Hprin1 IH1 Hinv Hfst _ IH2 Hclose Hwf HG HM1 HM2.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin1)) as Hofcirc.
invert Hofcirc.
intros HofP1.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
so (IH1 Hwf HG HM1) as Haprin1.
so (aprins_syn _#3 HG HM1 Haprin1) as Hsyncirc.
invert Hsyncirc.
intros HP1.
so (linv_syn _#3 Hinv HP1) as (HG' & HT).
eapply aprinps_bind; eauto.
apply IH2; eauto using wf, synctx, synctx_append, subst_syn.
apply subst_user; eauto using allvar_under, allvar_sh.
eapply lclose_impl_close; eauto.
}

(* unpack *)
{
intros G u k t P e M Hprin1 IH1 Hnorm _ IH2 Hwf HG He HM.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hofu.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
eapply aprinps_unpack; eauto 7 using wf, synctx.
}

(* epilogue *)
{
intro Hind.
destruct_all Hind.
do2 2 split; intros; eauto.
}
Qed.


Lemma aprint_complete_pre :
  forall G e t,
    wf G
    -> synctx G
    -> user e
    -> print G e t
    -> aprint G e t.
Proof.
exact (aprinx_complete andel).
Qed.


Lemma aprins_complete_pre :
  forall G M T,
    wf G
    -> synctx G
    -> user M
    -> prins G M T
    -> aprins G M T.
Proof.
exact (aprinx_complete anderl).
Qed.


Lemma aprinps_complete_pre :
  forall G M P,
    wf G
    -> synctx G
    -> user M
    -> prinps G M P
    -> aprinps G M P.
Proof.
exact (aprinx_complete anderr).
Qed.


Theorem aprint_complete :
  forall G e t,
    wf G
    -> synctx G
    -> user e
    -> oft G e t
    -> exists t',
         aprint G e t'
         /\ eqc G t' t kd_type.
Proof.
intros G e t Hwf HG He Hofe.
so (print_complete _#3 Hwf Hofe) as (t' & Hprin & Heqc).
exists t'.
split; auto.
apply aprint_complete_pre; auto.
Qed.


Theorem aprins_complete :
  forall G M T,
    wf G
    -> synctx G
    -> user M
    -> ofm G M T
    -> exists T',
         aprins G M T'
         /\ subs G T' T.
Proof.
intros G M T Hwf HG HM Hof.
so (prins_complete _#3 Hwf Hof) as (T' & Hprin & Hsub).
exists T'.
split; auto.
eapply aprins_complete_pre; eauto.
Qed.


Theorem aprinps_complete :
  forall G M P,
    wf G
    -> synctx G
    -> user M
    -> oflm G M P
    -> exists P',
         aprinps G M P'
         /\ subps G P' P.
Proof.
intros G M P Hwf HG HM Hof.
so (prinps_complete _#3 Hwf Hof) as (P' & Hprin & Hsub).
exists P'.
split; auto.
eapply aprinps_complete_pre; eauto.
Qed.
