
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import Regularity.
Require Import Declarative.
Require Import Consistent.
Require Import Preservation.


Inductive tcf : (* tm *) term -> (* cn *) term -> Prop :=
| tcf_unit
    : tcf tm_star cn_unit

| tcf_arrow {t e t1 t2}
    : tcf (tm_lam t e) (cn_arrow t1 t2)

| tcf_prod {e1 e2 t1 t2}
    : tcf (tm_pair e1 e2) (cn_prod t1 t2)

| tcf_all {k e k' t}
    : tcf (tm_plam k e) (cn_all k' t)

| tcf_exists {c e k t k' t'}
    : tcf (tm_pack c e k t) (cn_exists k' t')
  
.


Inductive mcf : (* md *) term -> (* cn *) term -> Prop :=
| mcf_one
    : mcf md_star sg_one

| mcf_stat {c k}
    : mcf (md_stat c) (sg_stat k)

| msf_dyn {e t}
    : mcf (md_dyn e) (sg_dyn t)

| mcf_pi {T M T1 T2}
    : mcf (md_lam T M) (sg_pi T1 T2)

| mcf_sigma {M1 M2 T1 T2}
    : mcf (md_pair M1 M2) (sg_sigma T1 T2)

| mcf_circ {M T}
    : mcf (md_cir M) (sg_circ T)
.


Lemma canon_tm :
  forall e t,
    oft nil e t
    -> similar t t
    -> tvalue nil e
    -> tcf e t.
Proof.
intros e t Hof Hsimt Hval.
so (oft_invert _#3 wf_nil Hof) as (t' & Heqc & Hofr).
assert (similar t' t') as Hsimt'.
  invert Hofr; try (intros; subst; invert Hval; done); try (intros; subst; auto using similar; done); [].
  (* var *)
  intros ? ? Hindex _ _.
  invert Hindex; done.
so (consistent _#3 wf_nil Heqc Hsimt' Hsimt) as Hsim.
invert Hofr;
try (intros; subst; invert Hval; done);
try (intros; subst; invert Hsim; intros; subst; auto using tcf; done).
(* var *)
intros i t'' Hindex _ _.
invert Hindex; done.
Qed.


Lemma mcf_subs :
  forall M T T',
    subs nil T T'
    -> mcf M T
    -> mcf M T'.
Proof.
intros M T T' Hsubs Hcf.
(* We use this contortion, rather than just inverting Hcf, so that
   we don't have to establish the converses for subs_X_invert'.
*)
#6 invert (subs_reg _#3 wf_nil Hsubs ander).

(* one *)
intros <-.
so (subs_one_invert' _ _ Hsubs); subst T.
invert Hcf; intros; subst; auto using mcf; done.

(* stat *)
intros k _ <-.
so (subs_stat_invert' _#3 Hsubs) as (k' & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* dyn *)
intros t _ <-.
so (subs_dyn_invert' _#3 Hsubs) as (t' & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* pi *)
intros T1 T2 _ _ _ _ <-.
so (subs_pi_invert' _#4 wf_nil Hsubs) as (T1' & T2' & _ & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* sigma *)
intros T1 T2 _ _ _ _ <-.
so (subs_sigma_invert' _#4 wf_nil Hsubs) as (T1' & T2' & _ & -> & _).
invert Hcf; intros; subst; auto using mcf; done.

(* circ *)
intros P _ <-.
so (subs_circ_invert' _#3 Hsubs) as (P' & -> & _).
invert Hcf; intros; subst; auto using mcf.
Qed.


Inductive elimcon : ((* md *) term -> (* md *) term) -> Prop :=
| elimcon_null 
    : elimcon (fun M => M)

| elimcon_pi1 {E}
    : elimcon E
      -> elimcon (fun M => md_pi1 (E M))
.


(* Induction principle that adds an extra premise to the extsigma case. *)
Lemma dynamic_ind'
     : forall P P0 P1 : context -> term -> term -> Prop,
       (forall (G : list term) (i : nat) (t : term),
        index i G (cl_tm t) -> P G (var i) (subst (sh (S i)) t)) ->
       (forall G : context, P G tm_star cn_unit) ->
       (forall (G : context) (t1 t2 e : term),
        ofc G t1 kd_type ->
        oft (G; cl_tm t1) e (shift1 t2) ->
        P (G; cl_tm t1) e (shift1 t2) -> P G (tm_lam t1 e) (cn_arrow t1 t2)) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        oft G e1 (cn_arrow t1 t2) ->
        P G e1 (cn_arrow t1 t2) -> oft G e2 t1 -> P G e2 t1 -> P G (tm_app e1 e2) t2) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        oft G e1 t1 ->
        P G e1 t1 -> oft G e2 t2 -> P G e2 t2 -> P G (tm_pair e1 e2) (cn_prod t1 t2)) ->
       (forall (G : context) (t1 t2 e : term),
        oft G e (cn_prod t1 t2) -> P G e (cn_prod t1 t2) -> P G (tm_pi1 e) t1) ->
       (forall (G : context) (t1 t2 e : term),
        oft G e (cn_prod t1 t2) -> P G e (cn_prod t1 t2) -> P G (tm_pi2 e) t2) ->
       (forall (G : context) (k t e : term),
        ofk G k ->
        oft (G; cl_cn k) e t -> P (G; cl_cn k) e t -> svalue e -> P G (tm_plam k e) (cn_all k t)) ->
       (forall (G : context) (k t e c : term),
        oft G e (cn_all k t) -> P G e (cn_all k t) -> ofc G c k -> P G (tm_papp e c) (subst1 c t)) ->
       (forall (G : context) (k t c e : term),
        ofc G c k ->
        oft G e (subst1 c t) ->
        P G e (subst1 c t) -> ofc (G; cl_cn k) t kd_type -> P G (tm_pack c e k t) (cn_exists k t)) ->
       (forall (G : context) (k t t' e1 e2 : term),
        oft G e1 (cn_exists k t) ->
        P G e1 (cn_exists k t) ->
        oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') ->
        P (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') ->
        P G (tm_unpack e1 e2) t') ->
       (forall (G : context) (t e : term),
        oft G e (cn_arrow (cn_arrow cn_unit t) t) ->
        P G e (cn_arrow (cn_arrow cn_unit t) t) -> P G (tm_fix t e) t) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        oft G e1 t1 ->
        P G e1 t1 ->
        oft (G; cl_tm t1) e2 (shift1 t2) ->
        P (G; cl_tm t1) e2 (shift1 t2) -> P G (tm_lett e1 e2) t2) ->
       (forall (G : context) (t t' e : term),
        oft G e t -> P G e t -> eqc G t t' kd_type -> P G e t') ->
       (forall (G : context) (P2 : term) (G' : context) (T k t M e : term),
        oflm G M P2 ->
        P1 G M P2 ->
        linv P2 G' T ->
        fsts T k ->
        oft (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) e) (subst (sh (2 + length G')) t) ->
        P (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) e) (subst (sh (2 + length G')) t) ->
        P G (tm_letm M e) t) ->
       (forall (G : context) (t M : term),
        ofm G M (sg_dyn t) -> P0 G M (sg_dyn t) -> P G (tm_ext M) t) ->
       (forall (G : list term) (i : nat) (T c : term),
        index i G (cl_md T c) -> P0 G (var i) (subst (sh (S i)) T)) ->
       (forall G : context, P0 G md_star sg_one) ->
       (forall (G : context) (c k : term), ofc G c k -> P0 G (md_stat c) (sg_stat k)) ->
       (forall (G : context) (e t : term), oft G e t -> P G e t -> P0 G (md_dyn e) (sg_dyn t)) ->
       (forall (G : context) (T1 T2 k1 M : term),
        ofs G T1 ->
        fsts T1 k1 ->
        ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) ->
        P0 (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) ->
        P0 G (md_lam T1 M) (sg_pi T1 T2)) ->
       (forall (G : context) (T1 T2 M1 M2 c2 : term),
        ofm G M1 (sg_pi T1 T2) ->
        P0 G M1 (sg_pi T1 T2) ->
        ofm G M2 T1 -> P0 G M2 T1 -> fstm G M2 c2 -> P0 G (md_app M1 M2) (subst1 c2 T2)) ->
       (forall (G : context) (T1 T2 M1 M2 : term),
        ofm G M1 T1 ->
        P0 G M1 T1 -> ofm G M2 T2 -> P0 G M2 T2 -> P0 G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))) ->
       (forall (G : context) (T1 T2 M : term),
        ofm G M (sg_sigma T1 T2) -> P0 G M (sg_sigma T1 T2) -> P0 G (md_pi1 M) T1) ->
       (forall (G : context) (T1 T2 M c : term),
        ofm G M (sg_sigma T1 T2) ->
        P0 G M (sg_sigma T1 T2) -> fstm G M c -> P0 G (md_pi2 M) (subst1 (cn_pi1 c) T2)) ->
       (forall (G : context) (t T e M : term),
        oft G e t ->
        P G e t ->
        ofm (G; cl_tm t) M (shift1 T) -> P0 (G; cl_tm t) M (shift1 T) -> P0 G (md_lett e M) T) ->
       (forall (G : context) (M P2 : term),
        oflm G M P2 -> P1 G M P2 -> P0 G (md_cir M) (sg_circ P2)) ->
       (forall (G : context) (k k' M c : term),
        ofm G M (sg_stat k') ->
        P0 G M (sg_stat k') -> fstm G M c -> ofc G c k -> P0 G M (sg_stat k)) ->
       (forall (G : context) (T1 T2 T3 k1 M : term),
        ofm G M (sg_pi T1 T3) ->
        P0 G M (sg_pi T1 T3) ->
        fsts T1 k1 ->
        ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (md_app (subst (sh 2) M) (var 0)) (shift1 T2) ->
        P0 (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (md_app (subst (sh 2) M) (var 0)) (shift1 T2) ->
        P0 G M (sg_pi T1 T2)) ->
    (forall G (T1 T2 T3 T4 M : term),
     ofm G M (sg_sigma T3 T4) ->
     P0 G M (sg_sigma T3 T4) ->
     ofm G (md_pi1 M) T1 ->
     P0 G (md_pi1 M) T1 ->
     ofm G (md_pi2 M) T2 ->
     P0 G (md_pi2 M) T2 -> P0 G M (sg_sigma T1 (shift1 T2))) ->
       (forall (G : context) (T T' M : term), ofm G M T -> P0 G M T -> subs G T T' -> P0 G M T') ->
       (forall (G : context) (M T : term), ofm G M T -> P0 G M T -> P1 G (lm_ret M) (ps_down T)) ->
       (forall (G : context) (M T : term), ofm G M T -> P0 G M T -> P1 G (lm_seal M T) (ps_down T)) ->
       (forall (G : context) (M1 M2 P2 : term) (G' : context) (T k P3 : term),
        ofm G M1 (sg_circ P2) ->
        P0 G M1 (sg_circ P2) ->
        linv P2 G' T ->
        fsts T k ->
        oflm (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) M2) (subst (sh (2 + length G')) P3) ->
        P1 (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) M2) (subst (sh (2 + length G')) P3) ->
        P1 G (lm_bind M1 M2) P3) ->
       (forall (G : context) (k t P2 e M : term),
        oft G e (cn_exists k t) ->
        P G e (cn_exists k t) ->
        oflm (G; cl_cn k; cl_tm t) M (subst (sh 2) P2) ->
        P1 (G; cl_cn k; cl_tm t) M (subst (sh 2) P2) -> P1 G (lm_unpack e M) P2) ->
       (forall (G : context) (P2 P' M : term),
        oflm G M P2 -> P1 G M P2 -> subps G P2 P' -> P1 G M P') ->
       (forall (c : context) (t t0 : term), oft c t t0 -> P c t t0) /\
       (forall (c : context) (t t0 : term), ofm c t t0 -> P0 c t t0) /\
       (forall (c : context) (t t0 : term), oflm c t t0 -> P1 c t t0).

Proof.
intros Pt Pm Plm.
intros.
exploit
  (dynamic_ind
     Pt
     (fun G M T =>
        True
        /\
        Pm G M T
        /\
        (forall E M',
           elimcon E
           -> M = E (md_pi1 M')
           -> exists T3 T4,
                ofm G M' (sg_sigma T3 T4)
                /\ Pm G M' (sg_sigma T3 T4)))
     Plm) as Hind;
  try (intros; 
       repeat (lazymatch goal with
               | H : True /\ _ |- _ => destruct H as (_ & ? & ?)
               end);
       try (split; [| split]);
       try (intros; let HE := findhyp elimcon in invert HE; intros; subst; discriminate);
       eauto; done).

(* pi1 *)
intros G T1 T2 M Hof (_ & IH & IH').
do2 2 split; eauto; [].
intros E M' HE Heq.
#2 invert HE.
  intros <-.
  injection Heq; intros; subst M'.
  eauto; done.

  intros E' HE' <-.
  injection Heq.
  intro.
  apply (IH' E'); auto; done.

(* extsigma *)
intros G T1 T2 M Hof1 (_ & IH1 & IH1') Hof2 (_ & IH2 & _).
so (IH1' _ _ elimcon_null (eq_refl _)) as (T3 & T4 & Hof & Hm).
do2 2 split; eauto; [].
intros E M' HE ->.
apply (IH1' _ _ (elimcon_pi1 HE)); eauto; done.

(* epilogue *)
destruct Hind as (? & Hindm & ?).
do2 2 split; eauto; [].
intros G t M Hof.
so (Hindm _#3 Hof) as (_ & ? & _).
assumption.
Qed.


Lemma canon_md :
  forall M T,
    ofm nil M T
    -> mvalue nil M
    -> mcf M T.
Proof.
exploit
  (dynamic_ind'
     (fun _ _ _ => True)
     (fun G M T =>
        G = nil
        -> mvalue nil M
        -> mcf M T)
     (fun _ _ _ => True)) as Hind;
  trivial;
  try (intros; let Hval := findhyp mvalue in invert Hval; done);
  auto using mcf.

(* var *)
intros G i T c Hindex -> _.
invert Hindex; done.

(* stat *)
intros G k k' M c _ IH _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* extpi *)
intros G T1 T2 T3 k1 M _ IH _ _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* extsigma *)
intros G T1 T2 T3 T4 M _ IH _ _ _ _ -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
invert Hcf; intros; subst; auto using mcf; done.

(* subsume *)
intros G T T' M _ IH Hsubs -> Hval.
so (IH (eq_refl _) Hval) as Hcf.
eapply mcf_subs; eauto; done.

(* epilogue *)
destruct_all Hind.
intros; eauto; done.
Qed.


Lemma oflm_ret_invert_something :
  forall G M P,
    oflm G (lm_ret M) P
    -> exists T, ofm G M T.
Proof.
intros G M P H.
remember (lm_ret M) as N eqn:HeqN.
revert HeqN.
induct H; try (intros; discriminate).

(* ret *)
{
intros G N T H HeqN.
injection HeqN; intros; subst N.
exists T.
auto.
}

(* subsume *)
{
intros G P P' N _ IH _ ->.
exact (IH (eq_refl _)).
}
Qed.


Theorem progress :
  (forall e t,
     oft nil e t
     -> tvalue nil e \/ exists e', tstep nil e e')
  /\
  (forall M T,
     ofm nil M T
     -> mvalue nil M \/ exists M', mstep nil M M')
  /\
  (forall M P,
     oflm nil M P
     -> lmvalue nil M \/ exists M', lmstep nil M M').
Proof.
exploit
  (dynamic_ind
     (fun G e t =>
        G = nil
        -> tvalue nil e \/ exists e', tstep nil e e')
     (fun G M T =>
        G = nil
        -> mvalue nil M \/ exists M', mstep nil M M')
     (fun G M P =>
        G = nil
        -> lmvalue nil M \/ exists M', lmstep nil M M')) as Hind;
try (intros; left; auto with dynamic; done);
eauto with dynamic.

(* var *)
intros G i t Hindex ->.
invert Hindex; done.

(* app *)
intros G t1 t2 e1 e2 Hof1 IH1 _ IH2 ->.
right.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof1 similar_arrow Hval1) as Hcf.
invert Hcf; [].
intros t e <-.
eauto with dynamic; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 ->.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; done.

(* pi1 *)
intros G t1 t2 e Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_prod Hval) as Hcf.
invert Hcf; [].
intros e1 e2 <-.
invert Hval; [].
eauto with dynamic; done.

(* pi2 *)
intros G t1 t2 e Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_prod Hval) as Hcf.
invert Hcf; [].
intros e1 e2 <-.
invert Hval; [].
eauto with dynamic; done.

(* papp *)
intros G k t e c Hof IH _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_all Hval) as Hcf.
invert Hcf; [].
intros; subst; eauto with dynamic; done.

(* pack *)
intros G k t c e _ _ IH _ ->.
so (IH eqrefl) as [| ()]; eauto with dynamic; done.

(* unpack *)
intros G k t t' e1 e2 Hof IH _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_exists Hval) as Hcf.
invert Hcf; [].
intros c e k0 t0 <-.
invert Hval; [].
eauto with dynamic; done.

(* lett *)
intros G t1 t2 e1 e2 _ IH _ _ ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* letm *)
{
intros G P G' T k t M e Hof IH _ _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
invert Hval.
intros V Hval' <-.
so (oflm_ret_invert_something _#3 Hof) as (T' & HofV).
so (can_fstm _#3 wf_nil HofV) as (c & Hfst).
eauto with dynamic; done.
}

(* ext *)
intros G t M Hof IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _#2 Hof Hval) as Hcf.
invert Hcf; [].
intros e <-.
invert Hval; [].
eauto with dynamic; done.

(* var *)
intros G i T c Hindex ->.
invert Hindex; done.

(* dyn *)
intros G e t _ IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* appap *)
intros G T1 T2 M1 M2 c2 Hof IH1 _ IH2 Hfst ->.
right.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | ()]; eauto with dynamic; [].
so (canon_md _ _ Hof Hval1) as Hcf.
invert Hcf; [].
intros T M <-.
eauto with dynamic; done.

(* pair *)
intros G T1 T2 M1 M2 _ IH1 _ IH2 ->.
so (IH1 eqrefl) as [Hval1 | ()]; eauto with dynamic; [].
so (IH2 eqrefl) as [Hval2 | (M2' & Hstep2)]; eauto with dynamic; done.

(* pi1 *)
intros G T1 T2 M Hof IH ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _ _ Hof Hval) as Hcf.
invert Hcf; [].
intros M1 M2 <-.
invert Hval; [].
eauto with dynamic; done.

(* pi2 *)
intros G T1 T2 M c Hof IH _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_md _ _ Hof Hval) as Hcf.
invert Hcf; [].
intros M1 M2 <-.
invert Hval; [].
eauto with dynamic; done.

(* lett *)
intros G t T e M _ IH _ _ ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.

(* cir *)
{
intros G M P _ IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; done.
}

(* extsigma *)
intros G T1 T2 M _ IH _ _ ->.
so (IH eqrefl) as [Hval | (N & Hstep)].
  invert Hval; auto; done.
  
  #2 invert Hstep.
    eauto; done.

    intros; subst.
    eauto with dynamic; done.

(* ret *)
{
intros G M T _ IH ->.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 Hof IH _ _ _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic.
so (canon_md _ _ Hof Hval) as Hcf.
invert Hcf ;[].
intros M' <-.
invert Hval; [].
intros Hval'.
invert Hval'; [].
intros V Hval'' <-.
so (ofm_cir_invert _#3 Hof) as Hof'.
so (oflm_ret_invert_something _#3 Hof') as (T' & Hof'').
so (can_fstm _#3 wf_nil Hof'') as (c & Hfst).
exists (subst (dot V (dot c id)) M2).
apply lmstep_bind2; auto.
}

(* unpack *)
{
intros G k t P e M Hof IH _ _ ->.
right.
so (IH eqrefl) as [Hval | ()]; eauto with dynamic; [].
so (canon_tm _ _ Hof similar_exists Hval) as Hcf.
invert Hcf; [].
intros c e0 k0 t0 <-.
invert Hval; [].
eauto with dynamic; done.
}

(* epilogue *)
destruct_all Hind.
do2 2 split; eauto; done.
Qed.


Lemma progress_tm :    
  forall e t,
    oft nil e t
    -> tvalue nil e \/ exists e', tstep nil e e'.
Proof.
exact (progress andel).
Qed.


Lemma progress_md :
  forall M T,
    ofm nil M T
    -> mvalue nil M \/ exists M', mstep nil M M'.
Proof.
exact (progress anderl).
Qed.


Lemma progress_lm :
  forall M T,
    oflm nil M T
    -> lmvalue nil M \/ exists M', lmstep nil M M'.
Proof.
exact (progress anderr).
Qed.
