
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import Formation.
Require Import TSubst.
Require Import Functionality.


Inductive wfh : context -> context -> Prop :=
| wfh_nil {G}
    : wfh G nil

| wfh_cn {G1 G2 k}
    : wfh G1 G2
      -> ofk (G1;; G2) k
      -> wfh G1 (G2; cl_cn k)

| wfh_tm {G1 G2 t}
    : wfh G1 G2
      -> ofc (G1;; G2) t kd_type
      -> wfh G1 (G2; cl_tm t)

| wfh_md {G1 G2 k T}
    : wfh G1 G2
      -> ofs (G1;; G2) T
      -> fsts T k
      -> wfh G1 (G2; cl_cn k; cl_md (shift1 T) (var 0))
.


Lemma wfh_allcl :
  forall G1 G2, wfh G1 G2 -> Forall iscl G2.
Proof.
intros G1 G2 H.
induct H; eauto using Forall_cons, iscl.
Qed.


Lemma wf_allcl :
  forall G, wf G -> Forall iscl G.
Proof.
intros G H.
induct H;
eauto using Forall_cons, iscl.
Qed.


Lemma wf_wfh_append :
  forall G1 G2,
    wf G1
    -> wfh G1 G2
    -> wf (G1;; G2).
Proof.
intros G1 G2 Hwf1 Hwf2.
revert Hwf1.
induct Hwf2; intros; simpl; auto using wf.
Qed.


Lemma wfh_append :
  forall G1 G2 G3,
    wfh G1 G2
    -> wfh (G1;; G2) G3
    -> wfh G1 (G2;; G3).
Proof.
intros G1 G2 G3 Hwf2 Hwf3.
remember (G1;; G2) as G eqn:Heq.
revert Heq.
induct Hwf3.

(* nil *)
auto; done.

(* cn *)
intros G G3 k _ IH Hof ->.
simpl.
apply wfh_cn; auto; [].
rewrite <- app_assoc; auto; done.

(* tm *)
intros G G3 t _ IH Hof ->.
simpl.
apply wfh_tm; auto; [].
rewrite <- app_assoc; auto; done.

(* md *)
intros G G3 k T _ IH Hof Hfst ->.
simpl.
apply wfh_md; auto; [].
rewrite <- app_assoc; auto; done.
Qed.


Lemma wf_wfh :
  forall G, wf G -> wfh nil G.
Proof.
intros G H.
induct H.

(* nil *)
exact wfh_nil.

(* cn *)
intros G k _ IH Hof.
apply wfh_cn; auto; [].
rewrite -> app_nil_r; [].
assumption.

(* cn *)
intros G t _ IH Hof.
apply wfh_tm; auto; [].
rewrite -> app_nil_r; [].
assumption.

(* md *)
intros G k T _ IH Hof Hfst.
apply wfh_md; auto; [].
rewrite -> app_nil_r; [].
assumption.
Qed.


Lemma subst_wfh :
  forall G1 G2 s G,
    subof sof G1 s G2
    -> wfh G2 G
    -> wfh G1 (subst_ctx s G).
Proof.
intros G1 G2 s G Hs Hwf.
revert G1 s Hs.
induct Hwf.

(* nil *)
intros G G1 s Hs.
simpl.
apply wfh_nil; done.

(* cn *)
intros G2 G k Hwf IH Hof G1 s Hs.
simpl.
simp_sub.
apply wfh_cn; auto; [].
eapply subst_ofk; eauto; [].
apply subof_under_sof; auto; [].
eapply wfh_allcl; eauto; done.

(* tm *)
intros G2 G c Hwf IH Hof G1 s Hs.
simpl.
simp_sub.
apply wfh_tm; auto; [].
eapply subst_ofc_type; eauto; [].
apply subof_under_sof; auto; [].
eapply wfh_allcl; eauto; done.

(* md *)
intros G2 G k T Hwf IH Hof Hfst G1 s Hs.
simpl.
simp_sub.
rewrite -> subst_compose; [].
#2 apply wfh_md; auto.
  eapply subst_ofs; eauto; [].
  apply subof_under_sof; auto; [].
  eapply wfh_allcl; eauto; done.

  apply subst_fsts; auto; done.
Qed.


Fixpoint inh (k : (* kd *) term) {struct k} :=
  (match k with
   | kd_pi k1 k2 =>
       cn_lam k1 (inh k2)

   | kd_sigma k1 k2 =>
       cn_pair (inh k1) (subst1 (inh k1) (inh k2))

   | kd_sing c => c

   | kd_type => cn_unit

   | _ => cn_star
   end).


Lemma inh_ofc :
  forall G k,
    ofk G k
    -> ofc G (inh k) k.
Proof.
intros G k Hof.
induct Hof.

(* type *)
intros G.
simpl.
apply ofc_unit; done.

(* sing *)
intros G c Hof.
simpl.
apply ofc_sing; auto; done.

(* pi *)
intros G k1 k2 Hof1 _ Hof2 IH.
simpl.
apply ofc_lam; auto; done.

(* sigma *)
intros G k1 k2 Hof1 IH1 Hof2 IH2.
simpl.
apply ofc_pair; eauto using subst1_ofc; done.

(* one *)
intros G.
simpl.
apply ofc_star; done.
Qed.


(* Doesn't need any premises. *)
Lemma shift_under_inh :
  forall i j k, subst (under i (sh j)) (inh k) = inh (subst (under i (sh j)) k).
Proof.
intros i j k.
revert i.
#3 induct k; try (intros; simp_sub; simpl; auto; done).

(* var *)
intros n i.
simp_sub.
so (Compare_dec.dec_lt n i) as [Hlt | Hgeq].
  rewrite -> project_under_lt; auto; done.

  rewrite -> project_under_ge; [| omega].
  simp_sub.
  simpl; reflexivity.

(* pi *)
intros k1 _ k2 IH i.
simp_sub.
simpl.
simp_sub.
f_equal; [].
rewrite <- under_succ; [].
auto; done.

(* sigma *)
intros k1 IH1 k2 IH2 i.
simp_sub.
simpl.
simp_sub.
f_equal; auto; [].
rewrite <- under_succ; [].
rewrite <- IH2; [].
simp_sub.
rewrite <- IH1; [].
reflexivity.
Qed.


Inductive ofcr : context -> (* cn *) term -> (* kd *) term -> Prop :=
| ofcr_var {G i k}
    : index i G (cl_cn k)
      -> ofcr G (var i) (subst (sh (S i)) k)

| ofcr_lam {G k1 k2 c}
    : ofk G k1
      -> ofc (G; cl_cn k1) c k2
      -> ofcr G (cn_lam k1 c) (kd_pi k1 k2)

| ofcr_app {G k1 k2 c1 c2}
    : ofc G c1 (kd_pi k1 k2)
      -> ofc G c2 k1
      -> ofcr G (cn_app c1 c2) (subst1 c2 k2)

| ofcr_pair {G k1 k2 c1 c2}
    : ofc G c1 k1
      -> ofc G c2 (subst1 c1 k2)
      -> ofk (G; cl_cn k1) k2
      -> ofcr G (cn_pair c1 c2) (kd_sigma k1 k2)

| ofcr_pi1 {G k1 k2 c}
    : ofc G c (kd_sigma k1 k2)
      -> ofcr G (cn_pi1 c) k1

| ofcr_pi2 {G k1 k2 c}
    : ofc G c (kd_sigma k1 k2)
      -> ofcr G (cn_pi2 c) (subst1 (cn_pi1 c) k2)

| ofcr_star {G}
    : ofcr G cn_star kd_one

| ofcr_unit {G}
    : ofcr G cn_unit kd_type

| ofcr_arrow {G t1 t2}
    : ofc G t1 kd_type
      -> ofc G t2 kd_type
      -> ofcr G (cn_arrow t1 t2) kd_type

| ofcr_prod {G t1 t2}
    : ofc G t1 kd_type
      -> ofc G t2 kd_type
      -> ofcr G (cn_prod t1 t2) kd_type

| ofcr_all {G k t}
    : ofk G k
      -> ofc (G; cl_cn k) t kd_type
      -> ofcr G (cn_all k t) kd_type

| ofcr_exists {G k t}
    : ofk G k
      -> ofc (G; cl_cn k) t kd_type
      -> ofcr G (cn_exists k t) kd_type
.


Inductive elimcon : ((* cn *) term -> (* cn *) term) -> Prop :=
| elimcon_null
    : elimcon (fun c => c)

| elimcon_pi1 {E}
    : elimcon E
      -> elimcon (fun c => cn_pi1 (E c))
.


Lemma ofc_ofcr :
  forall G c k,
    ofc G c k
    -> exists l, ofcr G c l.
Proof.
intros G c k Hof.
assert (exists E, elimcon E /\ ofc G (E c) k) as (E & HE & Hof').
  exists (fun c => c).
  split.
    apply elimcon_null; done.

    assumption.
clear Hof.
remember (E c) as d eqn:Heq.
revert E c Heq HE.
#5 induct Hof';
try (intros;
     invert HE; 
     try (intros; subst; discriminate); [];
     intros; subst; subst;
     eexists; eauto using ofcr; done).

(* pi1 *)
intros G k1 k2 d Hof IH E c Heq HE.
#2 invert HE; try (intros; subst; discriminate).
  intros; subst E c.
  eexists; eapply ofcr_pi1; eauto; done.

  intros E' HE' <-.
  injection Heq.
  intros ->.
  eapply IH; eauto; done.

(* sing *)
intros G c _ IH E c0 Heq HE.
eapply IH; eauto; done.

(* extpi *)
intros G k1 k2 k3 c _ IH _ _ E c0 Heq HE.
eapply IH; eauto; done.

(* extsigma *)
intros G k1 k2 c _ IH _ _ _ E c0 Heq HE.
subst c.
exact (IH _ _ (eq_refl _) (elimcon_pi1 HE)).

(* subsume *)
intros G k k' c _ IH _ E c0 Heq HE.
eapply IH; eauto; done.
Qed.


Lemma ofc_arrow_invert :
  forall G t1 t2 k,
    ofc G (cn_arrow t1 t2) k
    -> ofc G t1 kd_type /\ ofc G t2 kd_type.
Proof.
intros G t1 t2 k Hofc.
so (ofc_ofcr _ _ _ Hofc) as (? & H).
invert H; auto; done.
Qed.


Lemma ofc_prod_invert :
  forall G t1 t2 k,
    ofc G (cn_prod t1 t2) k
    -> ofc G t1 kd_type /\ ofc G t2 kd_type.
Proof.
intros G t1 t2 k Hofc.
so (ofc_ofcr _ _ _ Hofc) as (? & H).
invert H; auto; done.
Qed.


Lemma ofc_all_invert :
  forall G k t l,
    ofc G (cn_all k t) l
    -> ofk G k /\ ofc (G; cl_cn k) t kd_type.
Proof.
intros G k t l Hofc.
so (ofc_ofcr _ _ _ Hofc) as (? & H).
invert H; auto; done.
Qed.


Lemma ofc_exists_invert :
  forall G k t l,
    ofc G (cn_exists k t) l
    -> ofk G k /\ ofc (G; cl_cn k) t kd_type.
Proof.
intros G k t l Hofc.
so (ofc_ofcr _ _ _ Hofc) as (? & H).
invert H; auto; done.
Qed.


Lemma ofk_narrow :
  forall G k k' k1,
    subk G k k'
    -> ofk (G; cl_cn k') k1
    -> ofk (G; cl_cn k) k1.
Proof.
intros G k k' k1 Hsubk Hofk.
so (subst_ofk _#4 (subof_subsume_0 _#3 Hsubk) Hofk) as H.
rewrite -> subst_under_id in H.
assumption.
Qed.


Lemma ofc_narrow :
  forall G k k' c1 k1,
    subk G k k'
    -> ofc (G; cl_cn k') c1 k1
    -> ofc (G; cl_cn k) c1 k1.
Proof.
intros G k k' c1 k1 Hsubk Hofc.
so (subst_ofc _#5 (subof_subsume_0 _#3 Hsubk) Hofc) as H.
rewrite -> !subst_under_id in H.
assumption.
Qed.


Lemma ofs_narrow :
  forall G k k' T,
    subk G k k'
    -> ofs (G; cl_cn k') T
    -> ofs (G; cl_cn k) T.
Proof.
intros G k k' k1 Hsubk Hofk.
so (subst_ofs _#4 (subof_subsume_0 _#3 Hsubk) Hofk) as H.
rewrite -> subst_under_id in H.
assumption.
Qed.


Lemma ofps_narrow :
  forall G k k' T,
    subk G k k'
    -> ofps (G; cl_cn k') T
    -> ofps (G; cl_cn k) T.
Proof.
intros G k k' k1 Hsubk Hofk.
so (subst_ofps _#4 (subof_subsume_0 _#3 Hsubk) Hofk) as H.
rewrite -> subst_under_id in H.
assumption.
Qed.


Lemma eqps_narrow :
  forall G k k' T T',
    subk G k k'
    -> eqps (G; cl_cn k') T T'
    -> eqps (G; cl_cn k) T T'.
Proof.
intros G k k' T T' Hsubk Heq.
so (subst_eqps _#5 (subof_subsume_0 _#3 Hsubk) Heq) as H.
rewrite -> !subst_under_id in H.
assumption.
Qed.


Lemma fsts_fun :
  forall S k k',
    fsts S k
    -> fsts S k'
    -> k = k'.
Proof.
intros S k k' Hfst Hfst'.
revert k' Hfst'.
induct Hfst; try (intros; invert Hfst'; intros; subst; auto; done).

(* piapp *)
intros S1 S2 k1 k2 _ IH1 _ IH2 k' Hfst'.
invert Hfst'.
intros k1' k2' Hfst1 Hfst2 <-.
f_equal; auto; done.

(* sigma *)
intros S1 S2 k1 k2 _ IH1 _ IH2 k' Hfst'.
invert Hfst'.
intros k1' k2' Hfst1 Hfst2 <-.
f_equal; auto; done.
Qed.


Lemma ofs_fst :
  forall G S k,
    ofs G S
    -> fsts S k
    -> ofk G k.
Proof.
intros G S k HS Hfst.
revert k Hfst.
induct HS; try (intros; invert Hfst; intros; subst; auto with static; done).

(* piapp *)
intros G S1 S2 k1 _ IH1 Hfst1 _ IH2 k Hfst.
invert Hfst; [].
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1') as Heqk1.
subst k1'.
auto with static; done.

(* sigma *)
intros G S1 S2 k1 _ IH1 Hfst1 _ IH2 k Hfst.
invert Hfst; [].
intros k1' k2 Hfst1' Hfst2 <-.
so (fsts_fun _#3 Hfst1 Hfst1') as Heqk1.
subst k1'.
auto with static; done.
Qed.


Lemma wf_tail :
  forall G A, wf (G; A) -> wf G.
Proof.
intros G A H.
invert H; try (intros; subst; auto; done); [].
intros G' k S H0 H1 H2 <- H4.
apply wf_cn; eauto using ofs_fst; done.
Qed.


Lemma regularity :
  (forall G k1 k2,
     wf G
     -> eqk G k1 k2
     -> ofk G k1 /\ ofk G k2)
  /\
  (forall G k1 k2,
     wf G
     -> subk G k1 k2
     -> ofk G k1 /\ ofk G k2)
  /\
  (forall G c k,
     wf G
     -> ofc G c k
     -> ofk G k)
  /\
  (forall G c1 c2 k,
     wf G
     -> eqc G c1 c2 k
     -> ofc G c1 k /\ ofc G c2 k /\ ofk G k).
Proof.
#23 exploit
  (static_ind
     (fun _ _ => True)
     (fun G k1 k2 => wf G -> ofk G k1 /\ ofk G k2)
     (fun G k1 k2 => wf G -> ofk G k1 /\ ofk G k2)
     (fun G c k => wf G -> ofk G k)
     (fun G c1 c2 k => wf G -> ofc G c1 k /\ ofc G c2 k /\ ofk G k)) as Hind;
  try (intros;
       repeat (match goal with
               | H:wf ?G -> _ |- _ =>
                   lapply H; [clear H; intro H; destruct_all H | eauto with static; done]
               end);
       eauto with static;
       done).

(* pi *)
intros G k1 k1' k2 k2' H1 IH1 _ IH2 H3.
lapply IH1; [clear IH1 | assumption].
intros (Hofk1 & Hofk1').
lapply IH2; auto with static; [].
intros (Hofk2 & Hofk2').
split; auto with static; [].
apply ofk_pi; auto; [].
eapply ofk_narrow; eauto with static; done.

(* sigma *)
intros G k1 k1' k2 k2' H1 IH1 _ IH2 H3.
lapply IH1; [clear IH1 | assumption].
intros (Hofk1 & Hofk1').
lapply IH2; auto with static; [].
intros (Hofk2 & Hofk2').
split; auto with static; [].
apply ofk_sigma; auto; [].
eapply ofk_narrow; eauto with static; done.

(* var *)
intros G i k Hindex Hwf.
remember (cl_cn k) as A eqn:HeqA.
revert HeqA Hwf.
#2 induct Hindex.
  (* 0 *)
  intros A G -> Hwf.
  invert Hwf.
  auto using weaken1_ofk; done.

  (* S *)
  intros i A G B _ IH HeqA Hwf.
  exploit IH as H; eauto using wf_tail; [].
  so (weaken1_ofk _ A _ H) as H'.
  simp_sub in H'; assumption.

(* app *)
intros G k1 k2 c1 c2 _ IH Hofc2 _ Hwf.
lapply IH; [| assumption].
intro Hofk.
invert Hofk.
intros _ Hofk2.
eapply subst1_ofk; eauto; done.

(* pi1 *)
intros G k1 k2 c _ IH Hwf.
lapply IH; [| assumption].
intro Hofk; invert Hofk; auto; done.

(* pi2 *)
intros G k1 k2 c Hofc IH Hwf.
lapply IH; auto; [].
intro Hofk; invert Hofk; []; intros _ Hofk2.
eapply subst1_ofk; eauto with static; done.

(* extpi *)
intros G k1 k2 k3 c _ IH1 _ H3 Hwf.
lapply IH1; auto; [].
intro Hofk; invert Hofk; []; intros Hofk1 _.
apply ofk_pi; auto with static; done.

(* lam *)
intros G k1 k1' k2 c c' Heqk IH1 Heqc IH2 Hwf.
lapply IH1; auto; [].
clear IH1; intros (Hofk1 & Hofk1').
exploit IH2 as (Hofc & Hofc' & Hofk2); auto with static; [].
repeat2 split; auto with static; [].
#2 apply (@ofc_subsume _ (kd_pi k1' k2)).
  apply ofc_lam; auto; [].
  eapply ofc_narrow; eauto with static; done.

  apply subk_pi; auto with static; [].
  eapply ofk_narrow; eauto with static; done.

(* app *)
intros G k1 k2 c1 c2 c1' c2' _ IH1 Heqc2 IH2 Hwf.
lapply IH1; auto; [].
intros (Hofc1 & Hofc1' & Hofk).
lapply IH2; auto; [].
intros (Hofc2 & Hofc2' & _).
invert Hofk; intros Hofk1 Hofk2.
#3 repeat2 split.
  eapply ofc_app; eauto; done.

  #2 apply (@ofc_subsume _ (subst1 c2' k2)).
    eapply ofc_app; eauto; done.

    apply subk_refl; [].
    apply eqk_symm; [].
    eapply funct_kd_reg; eauto; done.

  eapply subst1_ofk; eauto; done.

(* pair *)
intros G k1 k2 c1 c1' c2 c2' Heqc1 IH1 _ IH2 Hofk2 _ Hwf.
lapply IH1; auto; [].
intros (Hofc1 & Hofc1' & Hofk1).
lapply IH2; auto; [].
intros (Hofc2 & Hofc2' & _).
repeat2 split; auto with static; [].
apply ofc_pair; auto; [].
apply (@ofc_subsume _ (subst1 c1 k2)); auto; [].
apply subk_refl; [].
eapply funct_kd_reg; eauto; done.  

(* pi1 *)
intros G k1 k2 c c' _ IH Hwf.
lapply IH; auto; [].
intros (Hofc & Hofc' & Hofk).
invert Hofk.
intros Hofk1 Hofk2.
repeat2 split; eauto with static; done.

(* pi2 *)
intros G k1 k2 c c' Heqc IH Hwf.
lapply IH; auto; [].
intros (Hofc & Hofc' & Hofk).
invert Hofk.
intros Hofk1 Hofk2.
repeat2 split; eauto using subst1_ofk with static; [].
apply (@ofc_subsume _ (subst1 (cn_pi1 c') k2)); eauto with static; [].
apply subk_refl; [].
apply eqk_symm; [].
eapply funct_kd_reg; eauto with static; done.

(* all *)
intros G k k' t t' Heqk IH1 _ IH2 Hwf.
lapply IH1; auto; [].
intros (Hofk & Hofk').
lapply IH2; eauto with static; [].
intros (Hoft & Hoft' & _).
repeat2 split; eauto with static; [].
apply ofc_all; auto; [].
eapply ofc_narrow; eauto with static; done.

(* exists *)
intros G k k' t t' Heqk IH1 _ IH2 Hwf.
lapply IH1; auto; [].
intros (Hofk & Hofk').
lapply IH2; eauto with static; [].
intros (Hoft & Hoft' & _).
repeat2 split; eauto with static; [].
apply ofc_exists; auto; [].
eapply ofc_narrow; eauto with static; done.

(* sing *)
intros G c c' Heqc IH Hwf.
lapply IH; auto; [].
intros (Hofc & Hofc' & _).
repeat2 split; eauto with static; [].
apply (@ofc_subsume _ (kd_sing c')); eauto with static; done.

(* singelim *)
intros G c c' Hofc IH Hwf.
lapply IH; auto; [].
intro Hofk.
invert Hofk.
intros Hofc'.
repeat2 split; auto with static; [].
apply (@ofc_subsume _ (kd_sing c')); auto; [].
apply subk_sing_t; auto; done.

(* extpi *)
intros G k1 k2 k3 k3' c c' Hofc IH1 Hofc' _ _ IH3 Hwf.
lapply IH1; auto; [].
intro Hofk.
invert Hofk; intros Hofk1 _.
lapply IH3; auto with static; [].
intros (Hofapp & Hofapp' & Hofk2).
repeat2 split; eauto with static; done.

(* extpiw *)
intros G k1 k2 k3 c c' _ IH1 _ IH2 Hwf.
lapply IH1; auto; [].
intros (Hofc & Hofc' & Hofk).
invert Hofk; intros Hofk1 _.
lapply IH2; auto with static; [].
intros (Hofapp & Hofapp' & Hofk2).
repeat2 split; eauto with static; done.

(* extsigma *)
intros G k1 k2 c c' Heqc1 IH1 _ IH2 Hofk2 _ Hwf.
lapply IH1; auto; [].
intros (Hofc1 & Hofc1' & Hofk1).
lapply IH2; auto; [].
intros (Hofc2 & Hofc2' & _).
repeat2 split; eauto with static; [].
apply ofc_extsigma; auto; [].
apply (@ofc_subsume _ (subst1 (cn_pi1 c) k2)); auto; [].
apply subk_refl; [].
eapply funct_kd_reg; eauto; done.

(* beta *)
intros G k1 k2 c1 c2 Hofc2 IH2 Hofc1 IH1 Hwf.
lapply IH1; auto; [].
intro Hofk1.
lapply IH2; auto with static; [].
intro Hofk2.
repeat2 split; eauto using subst1_ofc, subst1_ofk with static; done.

(* beta1 *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hwf.
lapply IH1; auto; [].
intros Hofk1.
lapply IH2; auto; [].
intros Hofk2.
repeat2 split; auto; [].
apply (@ofc_pi1 _ _ (shift1 k2)); [].
#2 apply ofc_pair; auto.
  simp_sub; assumption.

  apply weaken1_ofk; auto; done.

(* beta2 *)
intros G k1 k2 c1 c2 Hofc1 IH1 Hofc2 IH2 Hwf.
lapply IH1; auto; [].
intros Hofk1.
lapply IH2; auto; [].
intros Hofk2.
repeat2 split; auto; [].
repl k2 with (subst1 (cn_pi1 (cn_pair c1 c2)) (shift1 k2)) by (simp_sub; reflexivity).
apply (@ofc_pi2 _ k1); [].
#2 apply ofc_pair; auto.
  simp_sub; assumption.

  apply weaken1_ofk; auto; done.

(* conclusion *)
destruct_all Hind.
do2 3 split; eauto; done.
Qed.


Lemma eqk_reg :
  forall G k1 k2,
    wf G
    -> eqk G k1 k2
    -> ofk G k1 /\ ofk G k2.
Proof.
exact (regularity andel).
Qed.


Lemma subk_reg :
  forall G k1 k2,
    wf G
    -> subk G k1 k2
    -> ofk G k1 /\ ofk G k2.
Proof.
exact (regularity ander andel).
Qed.


Lemma ofc_reg :
  forall G c k,
    wf G
    -> ofc G c k
    -> ofk G k.
Proof.
exact (regularity ander ander andel).
Qed.


Lemma eqc_reg :
  forall G c1 c2 k,
    wf G
    -> eqc G c1 c2 k
    -> ofc G c1 k /\ ofc G c2 k /\ ofk G k.
Proof.
exact (regularity ander ander ander).
Qed.


Lemma funct_kd :
  forall G k c1 c2 k',
    wf G
    -> eqc G c1 c2 k
    -> ofk (G; cl_cn k) k'
    -> eqk G (subst1 c1 k') (subst1 c2 k').
Proof.
intros G k c1 c2 k' Hwf Heqc Hofk.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & _).
eapply funct_kd_reg; eauto.
Qed.


Lemma funct_cn :
  forall G k c1 c2 c' k',
    wf G
    -> eqc G c1 c2 k
    -> ofc (G; cl_cn k) c' k'
    -> eqc G (subst1 c1 c') (subst1 c2 c') (subst1 c1 k').
Proof.
intros G k c1 c2 c' k' Hwf Heqc Hofc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc' & _).
eapply funct_cn_reg; eauto.
Qed.


Lemma funct_cn_type :
  forall G k c1 c2 c',
    wf G
    -> eqc G c1 c2 k
    -> ofc (G; cl_cn k) c' kd_type
    -> eqc G (subst1 c1 c') (subst1 c2 c') kd_type.
Proof.
intros G k c1 c2 c' Hwf Heqc Hof.
replsub kd_type with (subst1 c1 kd_type).
eapply funct_cn; eauto.
Qed.


Lemma funct_sg :
  forall G k c1 c2 T,
    wf G
    -> eqc G c1 c2 k
    -> ofs (G; cl_cn k) T
    -> eqs G (subst1 c1 T) (subst1 c2 T).
Proof.
intros G k c1 c2 T Hwf Heqc Hofs.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & _).
eapply funct_sg_reg; eauto.
Qed.


Lemma funct_ps :
  forall G k c1 c2 T,
    wf G
    -> eqc G c1 c2 k
    -> ofps (G; cl_cn k) T
    -> eqps G (subst1 c1 T) (subst1 c2 T).
Proof.
intros G k c1 c2 T Hwf Heqc Hofs.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & _).
eapply funct_ps_reg; eauto.
Qed.


Lemma funct_eqk :
  forall G k c1 c2 k1 k2,
    wf G
    -> eqc G c1 c2 k
    -> eqk (G; cl_cn k) k1 k2
    -> eqk G (subst1 c1 k1) (subst1 c2 k2).
Proof.
intros G k c1 c2 k1 k2 Hwf Heqc Heqk.
#2 apply (@eqk_trans _ _ (subst1 c2 k1)).
  eapply funct_kd; eauto; [].
  refine (eqk_reg _#3 _ Heqk andel); [].
  apply wf_cn; auto; [].
  eapply eqc_reg; eauto; done.

  eapply subst1_eqk; eauto; [].
  eapply eqc_reg; eauto; done.
Qed.


Lemma funct_eqc :
  forall G k k' c1 c2 d1 d2,
    wf G
    -> eqc G c1 c2 k
    -> eqc (G; cl_cn k) d1 d2 k'
    -> eqc G (subst1 c1 d1) (subst1 c2 d2) (subst1 c1 k').
Proof.
intros G k k' c1 c2 d1 d2 Hwf Heqc Heqd.
so (eqc_reg _#4 Hwf Heqc) as (Hofc1 & _ & Hofk).
#2 apply (@eqc_trans _ _ (subst1 c1 d2)).
  eapply subst1_eqc; eauto; done.

  so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqd) as (_ & Heqd2 & _).
  eapply funct_cn; eauto; done.
Qed.


Lemma funct_eqc_type :
  forall G k c1 c2 d1 d2,
    wf G
    -> eqc G c1 c2 k
    -> eqc (G; cl_cn k) d1 d2 kd_type
    -> eqc G (subst1 c1 d1) (subst1 c2 d2) kd_type.
Proof.
intros G k c1 c2 d1 d2 Hwf Heqc Heqc'.
replsub kd_type with (subst1 c1 kd_type).
eapply funct_eqc; eauto.
Qed.


(* statically equivalent *)
Inductive seq : context -> term -> term -> (* cl *) term -> Prop :=
| seq_cn {G c c' k}
    : eqc G c c' k
      -> seq G c c' (cl_cn k)

| seq_tm {G e e' t}
    : seq G e e' (cl_tm t)

| seq_md {G M M' T c}
    : seq G M M' (cl_md T c)

| seq_ucn {G c c'}
    : seq G c c' cl_ucn

| seq_utm {G t t'}
    : seq G t t' cl_utm

| seq_umd {G M M' c}
    : seq G M M' (cl_umd c)

| seq_ns {G t t'}
    : seq G t t' cl_ns
.


Lemma seq_behaved :
  behavedrel seq.
Proof.
intros G i C HC Hindex.
destruct HC; simp_sub; eauto using seq with static.
Qed.


Lemma seq_wclo :
  wclorel seq.
Proof.
intros G m m' C C' H.
invert H; try (intros; subst; simp_sub; eauto using seq; done).

(* cn *)
intros k Heqc <-.
simp_sub.
apply seq_cn; [].
apply weaken1_eqc; assumption.
Qed.


Lemma sof_seq :
  forall G m C,
    sof G m C
    -> seq G m m C.
Proof.
intros G m C H.
case H; try (intros; eauto using seq; done); [].
clear_all.
intros G c k Hof.
apply seq_cn; [].
apply eqc_refl; auto; done.
Qed.


Lemma subof_subrel_seq_trans :
  forall G1 G2 G3 s s1 s2,
    subof sof G1 s G2
    -> subrel seq G2 s1 s2 G3
    -> subrel seq G1 (compose s1 s) (compose s2 s) G3.
Proof.
intros G1 G2 G3 s s' i Hs Hss.
refine (subof_subrel_trans sof _#4 G2 _#6); eauto using sof_seq; [].
clear_all.
intros G1 G2 s m1 m2 C Hs Hseq.
invert Hseq; try (intros; subst; simp_sub; eauto using seq; done); [].
intros k Heqc <-.
simp_sub.
apply seq_cn; [].
eapply subst_eqc; eauto; done.
Qed.


Lemma subrel_seq_weaken :
  forall G1 G2 G3 s s' i,
    truncate i G1 G2
    -> subrel seq G2 s s' G3
    -> subrel seq G1 (compose s (sh i)) (compose s' (sh i)) G3.
Proof.
intros G1 G2 G3 s s' i Htrunc Hss.
eapply subof_subrel_seq_trans; eauto using subof_sh; done.
Qed.


Lemma seq_seqs :
  forall G m m' C,
    wf G
    -> seq G m m' C
    -> seqs G m m' C.
Proof.
intros G m m' C Hwf Hseq.
invert Hseq; try (intros; subst; eauto using seqs; done); [].
(* cn *)
intros k Heqc <-.
apply seqs_cn; auto; [].
exact (eqc_reg _#4 Hwf Heqc andel).
Qed.


Lemma seq_ofk :
  forall G1 s s' G2 k,
    wf G1
    -> subrel seq G1 s s' G2
    -> ofk G2 k
    -> eqk G1 (subst s k) (subst s' k).
Proof.
intros G1 s s' G2 k Hwf Hs Hof.
eapply seqs_ofk; eauto; [].
refine (subrel_mono' _#7 Hs); [].
eauto using seq_seqs; done.
Qed.


Lemma seq_ofc :
  forall G1 s s' G2 c k,
    wf G1
    -> subrel seq G1 s s' G2
    -> ofc G2 c k
    -> eqc G1 (subst s c) (subst s' c) (subst s k).
Proof.
intros G1 s s' G2 c k Hwf Hs Hof.
eapply seqs_ofc; eauto; [].
refine (subrel_mono' _#7 Hs); [].
eauto using seq_seqs; done.
Qed.


Lemma seq_ofc_type :
  forall G1 s s' G2 c,
    wf G1
    -> subrel seq G1 s s' G2
    -> ofc G2 c kd_type
    -> eqc G1 (subst s c) (subst s' c) kd_type.
Proof.
intros G1 s s' G2 c Hwf Hs Hof.
replsub kd_type with (subst s kd_type).
eapply seq_ofc; eauto.
Qed.


Lemma seq_eqc :
  forall G1 s s' G2 c c' k,
    wf G1
    -> wf G2
    -> subrel seq G1 s s' G2
    -> eqc G2 c c' k
    -> eqc G1 (subst s c) (subst s' c') (subst s k).
Proof.
intros G1 s s' G2 c c' k Hwf1 Hwf2 Hss Heqc.
assert (subrel seqs G1 s s' G2) as Hss'.
  refine (subrel_mono' _#7 Hss); [].
  eauto using seq_seqs; done.
clear Hss.
assert (subof sof G1 s G2) as Hs.
  eapply subrel_subof_mono; eauto using seqs_sof; done.
apply (@eqc_trans _ _ (subst s c')).
  eapply subst_eqc; eauto; done.

  eapply seqs_ofc; eauto; [].
  exact (eqc_reg _#4 Hwf2 Heqc ander andel).
Qed.


Lemma seq_eqc_type :
  forall G1 s s' G2 c c',
    wf G1
    -> wf G2
    -> subrel seq G1 s s' G2
    -> eqc G2 c c' kd_type
    -> eqc G1 (subst s c) (subst s' c') kd_type.
Proof.
intros G1 s s' G2 c c' H H0 H1 H2.
replsub kd_type with (subst s kd_type).
eapply seq_eqc; eauto; done.
Qed.


Lemma seq_ofs :
  forall G1 s s' G2 T,
    wf G1
    -> subrel seq G1 s s' G2
    -> ofs G2 T
    -> eqs G1 (subst s T) (subst s' T).
Proof.
intros G1 s s' G2 T Hwf Hs Hof.
eapply seqs_ofs; eauto; [].
refine (subrel_mono' _#7 Hs); [].
eauto using seq_seqs; done.
Qed.


Lemma seq_ofps :
  forall G1 s s' G2 T,
    wf G1
    -> subrel seq G1 s s' G2
    -> ofps G2 T
    -> eqps G1 (subst s T) (subst s' T).
Proof.
intros G1 s s' G2 T Hwf Hs Hof.
eapply seqs_ofps; eauto; [].
refine (subrel_mono' _#7 Hs); [].
eauto using seq_seqs; done.
Qed.


Lemma hsg_fsts :
  forall T,
    hsg T
    -> exists k, fsts T k.
Proof.
intros T H.
#2 induct H; try (intros; eexists; eauto with static; done).

(* piapp *)
intros T1 T2 _ (k1 & H1) _ (k2 & H2).
eauto with static; done.

(* sigma *)
intros T1 T2 _ (k1 & H1) _ (k2 & H2).
eauto with static; done.
Qed.


(* The typing condition is just to ensure that T is a syntactic signature. *)
Lemma can_fsts :
  forall G T,
    ofs G T
    -> exists k, fsts T k.
Proof.
intros G T H.
apply hsg_fsts; [].
eapply fsg_hsg; [].
apply ofs_fsg; eauto; done.
Qed.


Lemma eqs_fst_o :
  forall G T T',
    wf G
    -> eqs G T T'
    -> exists k k',
         fsts T k /\ fsts T' k' /\ eqk G k k'.
Proof.
intros G T T' Hwf Heqs.
revert Hwf.
induct Heqs.

(* refl *)
intros G T Hof Hwf.
so (can_fsts _ _ Hof) as (k & Hfst).
eexists; eexists; eauto using ofs_fst with static; done.

(* symm *)
intros G T T' _ IH Hwf.
so (IH Hwf) as (k & k' & Hfst & Hfst' & Heqk).
do 2 eexists; eauto with static; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (k1 & k2 & Hfst1 & Hfst2 & Heq12).
so (IH2 Hwf) as (k2' & k3 & Hfst2' & Hfst3 & Heq23).
so (fsts_fun _ _ _ Hfst2 Hfst2'); subst k2'.
do 2 eexists; eauto with static; done.

(* stat *)
intros G k k' Heqk Hwf.
exists k, k'.
auto with static; done.

(* dyn *)
intros G t t' _ _.
exists kd_one, kd_one.
auto with static; done.

(* pi *)
intros G T1 T1' k1 T2 T2' _ IH1 HfstT1 _ IH2 Hwf.
so (IH1 Hwf) as (k1a & k1' & HfstT1a & HfstT1' & Heqk1).
so (fsts_fun _ _ _ HfstT1 HfstT1a); subst k1a.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
sor (refine (IH2 _)) as (k2 & k2' & HfstT2 & HfstT2' & Heqk2); eauto using ofs_fst with static; [].
do 2 eexists; do2 2 split; eauto with static; done.

(* sigma *)
intros G T1 T1' k1 T2 T2' _ IH1 HfstT1 _ IH2 Hwf.
so (IH1 Hwf) as (k1a & k1' & HfstT1a & HfstT1' & Heqk1).
so (fsts_fun _ _ _ HfstT1 HfstT1a); subst k1a.
so (eqk_reg _#3 Hwf Heqk1) as (Hofk1 & _).
sor (refine (IH2 _)) as (k2 & k2' & HfstT2 & HfstT2' & Heqk2); eauto using ofs_fst with static; [].
do 2 eexists; do2 2 split; eauto with static; done.

(* circ *)
{
intros G P P' _ Hwf.
exists kd_one, kd_one.
do2 2 split; eauto with static.
}
Qed.


Lemma subs_fst_o :
  forall G T T',
    wf G
    -> subs G T T'
    -> exists k k',
         fsts T k /\ fsts T' k' /\ subk G k k'.
Proof.
intros G T T' Hwf Heqs.
revert Hwf.
induct Heqs.

(* refl *)
intros G T T' Heqs Hwf.
so (eqs_fst_o _#3 Hwf Heqs) as (k & k' & Hfst & Hfst' & Heqk).
eexists; eexists; eauto using ofs_fst with static; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as (k1 & k2 & Hfst1 & Hfst2 & Heq12).
so (IH2 Hwf) as (k2' & k3 & Hfst2' & Hfst3 & Heq23).
so (fsts_fun _ _ _ Hfst2 Hfst2'); subst k2'.
do 2 eexists; eauto with static; done.

(* stat *)
intros G k k' Hsubk Hwf.
exists k, k'.
auto with static; done.

(* piapp *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1' _ IH2 Hfst1 HofT2 Hwf.
so (IH1 Hwf) as (k1'a & k1a & Hfst1'a & Hfst1a & Hsubk1).
so (fsts_fun _ _ _ Hfst1 Hfst1a); subst k1a.
so (fsts_fun _ _ _ Hfst1' Hfst1'a); subst k1'a.
so (subk_reg _#3 Hwf Hsubk1) as (Hofk1' & _).
sor (refine (IH2 _)) as (k2 & k2' & HfstT2 & HfstT2' & Hsubk2); eauto using ofs_fst with static; [].
do 2 eexists; do2 2 split; eauto using ofs_fst with static; done.

(* sigma *)
intros G T1 T1' k1 k1' T2 T2' Hsubs1 IH1 Hfst1 _ IH2 Hfst1' HofT2' Hwf.
so (IH1 Hwf) as (k1a & k1'a & Hfst1a & Hfst1'a & Hsubk1).
so (fsts_fun _ _ _ Hfst1 Hfst1a); subst k1a.
so (fsts_fun _ _ _ Hfst1' Hfst1'a); subst k1'a.
so (subk_reg _#3 Hwf Hsubk1) as (Hofk1 & _).
sor (refine (IH2 _)) as (k2 & k2' & HfstT2 & HfstT2' & Hsubk2); eauto using ofs_fst with static; [].
do 2 eexists; do2 2 split; eauto using ofs_fst with static; done.

(* circ *)
{
intros G P P' _ Hwf.
exists kd_one, kd_one.
eauto with static.
}
Qed.


Lemma eqs_fst :
  forall G T T' k k',
    wf G
    -> eqs G T T'
    -> fsts T k
    -> fsts T' k'
    -> eqk G k k'.
Proof.
intros G T1 T2 k1 k2 Hwf Heqs HfstT1 HfstT2.
so (eqs_fst_o _#3 Hwf Heqs) as (k1' & k2' & HfstT1' & HfstT2' & Heqk).
so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
so (fsts_fun _ _ _ HfstT2 HfstT2'); subst k2'.
assumption.
Qed.


Lemma subs_fst :
  forall G T T' k k',
    wf G
    -> subs G T T'
    -> fsts T k
    -> fsts T' k'
    -> subk G k k'.
Proof.
intros G T1 T2 k1 k2 Hwf Hsubs HfstT1 HfstT2.
so (subs_fst_o _#3 Hwf Hsubs) as (k1' & k2' & HfstT1' & HfstT2' & Heqk).
so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
so (fsts_fun _ _ _ HfstT2 HfstT2'); subst k2'.
assumption.
Qed.


Lemma eqs_eqps_reg :
  (forall G T T',
     wf G
     -> eqs G T T'
     -> ofs G T /\ ofs G T')
  /\
  (forall G T T',
     wf G
     -> eqps G T T'
     -> ofps G T /\ ofps G T').
Proof.
exploit
  (eqsig_ind
     (fun G T T' =>
        wf G
        -> ofs G T /\ ofs G T')
     (fun G T T' =>
        wf G
        -> ofps G T /\ ofps G T')) as Hind.

(* refl *)
intros; auto; done.

(* symm *)
intros G T T' _ IH Hwf.
so (IH Hwf) as ().
auto; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
split.
  exact (IH1 Hwf andel).

  exact (IH2 Hwf ander).

(* stat *)
intros G k k' Heqk Hwf.
so (eqk_reg _#3 Hwf Heqk) as (H1 & H2).
split; auto with static; done.

(* dyn *)
intros G t t' Heqc Hwf.
so (eqc_reg _#4 Hwf Heqc) as (H1 & H2 & _).
split; auto with static; done.

(* piapp *)
intros G T1 T1' k1 T2 T2' Heqs IH1 HfstT1 _ IH2 Hwf.
so (IH1 Hwf) as (HofT1 & HofT1').
sor (refine (IH2 _)) as (HofT2 & HofT2'); eauto using ofs_fst with static; [].
so (can_fsts _ _ HofT1') as (k1' & HfstT1').
split; eapply ofs_pi; eauto; [].
eapply ofs_narrow; eauto; [].
apply subk_refl; [].
apply eqk_symm; [].
eapply eqs_fst; eauto; done.

(* sigma *)
intros G T1 T1' k1 T2 T2' Heqs IH1 HfstT1 _ IH2 Hwf.
so (IH1 Hwf) as (HofT1 & HofT1').
sor (refine (IH2 _)) as (HofT2 & HofT2'); eauto using ofs_fst with static; [].
so (can_fsts _ _ HofT1') as (k1' & HfstT1').
split; eapply ofs_sigma; eauto; [].
eapply ofs_narrow; eauto; [].
apply subk_refl; [].
apply eqk_symm; [].
eapply eqs_fst; eauto; done.

(* circ *)
{
intros G P P' _ IH Hwf.
so (IH Hwf) as (HofP & HofP').
eauto using ofs_circ.
}

(* down *)
{
intros G T T' _ IH Hwf.
so (IH Hwf) as (HofT & HofT').
eauto using ofps_down.
}

(* ex *)
{
intros G k k' P P' Heqk _ IH Hwf.
so (eqk_reg _#3 Hwf Heqk) as (Hofk & Hofk').
so (IH (wf_cn Hwf Hofk)) as (HofP & HofP').
split; eauto using ofps_ex.
apply ofps_ex; auto.
eapply ofps_narrow; eauto.
apply subk_refl.
apply eqk_symm; auto.
}

(* wrapup *)
{
destruct Hind.
split; intros; eauto.
}
Qed.


Lemma eqs_reg :
  forall G T T',
    wf G
    -> eqs G T T'
    -> ofs G T /\ ofs G T'.
Proof.
exact (eqs_eqps_reg andel).
Qed.


Lemma eqps_reg :
  forall G T T',
    wf G
    -> eqps G T T'
    -> ofps G T /\ ofps G T'.
Proof.
exact (eqs_eqps_reg ander).
Qed.


Lemma funct_eqs :
  forall G k c1 c2 T1 T2,
    wf G
    -> eqc G c1 c2 k
    -> eqs (G; cl_cn k) T1 T2
    -> eqs G (subst1 c1 T1) (subst1 c2 T2).
Proof.
intros G k c1 c2 T1 T2 Hwf Heqc Heqs.
so (eqc_reg _#4 Hwf Heqc) as (Hofc1 & _ & Hofk).
#2 apply (@eqs_trans _ _ (subst1 c1 T2)).
  eapply subst1_eqs; eauto; done.

  so (eqs_reg _#3 (wf_cn Hwf Hofk) Heqs) as (_ & HofT2).
  eapply funct_sg; eauto; done.
Qed.


Lemma eqps_refl :
  forall {G T},
    ofps G T
    -> eqps G T T.
Proof.
intros G T H.
induct H.

(* down *)
{
intros G T H.
apply eqps_down; auto using eqs_refl.
}

(* ex *)
{
intros G k P Hofk _ IH.
apply eqps_ex; auto using eqk_refl.
}
Qed.


Lemma eqps_symm :
  forall {G T T'},
    eqps G T T'
    -> eqps G T' T.
Proof.
intros G T T' H.
induct H.

(* down *)
{
intros G T T' H.
apply eqps_down.
apply eqs_symm; auto.
}

(* ex *)
{
intros G k k' P P' Heqk _ Heqps.
apply eqps_ex; auto using eqk_symm.
so (subst_eqps _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Heqps) as H.
simp_sub in H.
exact H.
}
Qed.


Lemma eqps_trans :
  forall {G T1 T2 T3},
    eqps G T1 T2
    -> eqps G T2 T3
    -> eqps G T1 T3.
Proof.
intros G P1 P2 P3 H12 H23.
revert P3 H23.
induct H12.

(* down *)
{
intros G T1 T2 H12 P3 HP23.
invert HP23; [].
intros T3 H23 <-.
apply eqps_down.
eapply eqs_trans; eauto.
}

(* ex *)
{
intros G k1 k2 P1 P2 Hk12 _ IH P HP23.
invert HP23; [].
intros k3 P3 Hk23 H23 <-.
apply eqps_ex.
  {
  eapply eqk_trans; eauto.
  }
apply IH.
eapply eqps_narrow; eauto.
apply subk_refl; auto.
}
Qed.


Lemma funct_eqps :
  forall G k c1 c2 T1 T2,
    wf G
    -> eqc G c1 c2 k
    -> eqps (G; cl_cn k) T1 T2
    -> eqps G (subst1 c1 T1) (subst1 c2 T2).
Proof.
intros G k c1 c2 T1 T2 Hwf Heqc Heqs.
so (eqc_reg _#4 Hwf Heqc) as (Hofc1 & _ & Hofk).
#2 apply (@eqps_trans _ _ (subst1 c1 T2)).
  eapply subst1_eqps; eauto; done.

  so (eqps_reg _#3 (wf_cn Hwf Hofk) Heqs) as (_ & HofT2).
  eapply funct_ps; eauto; done.
Qed.


Lemma subs_subps_reg :
  (forall G T T',
     wf G
     -> subs G T T'
     -> ofs G T /\ ofs G T')
  /\
  (forall G T T',
     wf G
     -> subps G T T'
     -> ofps G T /\ ofps G T').
Proof.
exploit
   (subsig_ind
     (fun G T T' =>
        wf G
        -> ofs G T /\ ofs G T')
     (fun G T T' =>
        wf G
        -> ofps G T /\ ofps G T')) as Hind.

(* refl *)
intros G T T' Heqs Hwf.
so (eqs_reg _#3 Hwf Heqs) as ().
auto; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
split.
  exact (IH1 Hwf andel).

  exact (IH2 Hwf ander).

(* stat *)
intros G k k' Hsubk Hwf.
so (subk_reg _#3 Hwf Hsubk) as (H1 & H2).
split; auto with static; done.

(* pi *)
intros G T1 T1' k1 k1' T2 T2' Hsubs IH1 HfstT1' _ IH2 HfstT1 HofT2 Hwf.
so (IH1 Hwf) as (HofT1' & HofT1).
sor (refine (IH2 _)) as (_ & HofT2'); eauto using ofs_fst with static; done.

(* sigma *)
intros G T1 T1' k1 k1' T2 T2' Hsubs IH1 HfstT1 _ IH2 HfstT1' HofT2' Hwf.
so (IH1 Hwf) as (HofT1 & HofT1').
sor (refine (IH2 _)) as (HofT2 & _); eauto using ofs_fst with static; done.

(* circ *)
{
intros G P P' _ IH Hwf.
so (IH Hwf) as (HofP & HofP').
eauto using ofs_circ.
}

(* refl *)
intros G T T' Heqs Hwf.
so (eqps_reg _#3 Hwf Heqs) as ().
auto; done.

(* trans *)
intros G T1 T2 T3 _ IH1 _ IH2 Hwf.
split.
  exact (IH1 Hwf andel).

  exact (IH2 Hwf ander).

(* down *)
{
intros G T T' _ IH Hwf.
so (IH Hwf) as (HofT & HofT').
eauto using ofps_down.
}

(* ex_left *)
{
intros G k P P' Hofk _ IH HofP' Hwf.
so (IH (wf_cn Hwf Hofk)) as (HofP & _).
eauto using ofps_ex.
}

(* ex_right *)
{
intros G P k P' c Hofc _ IH HofP' Hwf.
so (IH Hwf) as (HofP & _).
so (ofc_reg _#3 Hwf Hofc) as Hofk.
eauto using ofps_ex.
}

(* wrapup *)
{
destruct Hind.
split; intros; eauto.
}
Qed.


Lemma subs_reg :
  forall G T T',
    wf G
    -> subs G T T'
    -> ofs G T /\ ofs G T'.
Proof.
exact (subs_subps_reg andel).
Qed.


Lemma subps_reg :
  forall G T T',
    wf G
    -> subps G T T'
    -> ofps G T /\ ofps G T'.
Proof.
exact (subs_subps_reg ander).
Qed.


Lemma seq_subs :
  forall G1 s s' G2 T T',
    wf G1
    -> wf G2
    -> subrel seq G1 s s' G2
    -> subs G2 T T'
    -> subs G1 (subst s T) (subst s' T').
Proof.
intros G1 s s' G2 T T' Hwf1 Hwf2 Hss Hsubs.
assert (subrel seqs G1 s s' G2) as Hss'.
  refine (subrel_mono' _#7 Hss); [].
  eauto using seq_seqs; done.
clear Hss.
assert (subof sof G1 s G2) as Hs.
  eapply subrel_subof_mono; eauto using seqs_sof; done.
apply (@subs_trans _ _ (subst s T')).
  eapply subst_subs; eauto; done.

  apply subs_refl; [].
  eapply seqs_ofs; eauto; [].
  exact (subs_reg _#3 Hwf2 Hsubs ander).
Qed.


Lemma seq_subps :
  forall G1 s s' G2 T T',
    wf G1
    -> wf G2
    -> subrel seq G1 s s' G2
    -> subps G2 T T'
    -> subps G1 (subst s T) (subst s' T').
Proof.
intros G1 s s' G2 T T' Hwf1 Hwf2 Hss Hsubs.
assert (subrel seqs G1 s s' G2) as Hss'.
  refine (subrel_mono' _#7 Hss); [].
  eauto using seq_seqs; done.
clear Hss.
assert (subof sof G1 s G2) as Hs.
  eapply subrel_subof_mono; eauto using seqs_sof; done.
apply (@subps_trans _ _ (subst s T')).
  eapply subst_subps; eauto; done.

  apply subps_refl; [].
  eapply seqs_ofps; eauto; [].
  exact (subps_reg _#3 Hwf2 Hsubs ander).
Qed.


Lemma fstm_fun :
  forall G M c c',
    fstm G M c
    -> fstm G M c'
    -> c = c'.
Proof.
intros G M c c' Hfstm Hfstm'.
revert c' Hfstm'.
induct Hfstm.

(* var *)
intros G i T c Hindex d Hfstm'.
#2 invert Hfstm'.
  intros T' c' Hindex' <-.
  so (index_fun _#5 Hindex Hindex') as Heq.
  injection Heq.
  intros; subst T' c'.
  reflexivity.

  intros c' Hindex' <-.
  so (index_fun _#5 Hindex Hindex') as Heq.
  discriminate Heq.

(* var *)
intros G i c Hindex d Hfstm'.
#2 invert Hfstm'.
  intros c' T Hindex' <-.
  so (index_fun _#5 Hindex Hindex') as Heq.
  discriminate Heq.

  intros c' Hindex' <-.
  so (index_fun _#5 Hindex Hindex') as Heq.
  injection Heq.
  intros; subst c'.
  reflexivity.

(* star *)
intros G c' Hfstm'.
invert Hfstm'; [].
intros; subst; reflexivity.

(* stat *)
intros G c d Hfstm'.
invert Hfstm'; [].
auto; done.

(* dyn *)
intros G c d Hfstm'.
invert Hfstm'; [].
auto; done.

(* lam *)
intros G T M k c HfstT _ IH d Hfstm'.
invert Hfstm'; [].
intros k' c' HfstT' Hfstc' <-.
so (fsts_fun _ _ _ HfstT HfstT'); subst k'.
f_equal; auto; [].
so (IH _ Hfstc') as Heq.
f_apply (subst1 nonsense) in Heq as Heq'.
simp_sub in Heq'.
assumption.

(* app *)
intros G M1 M2 c1 c2 _ IH1 _ IH2 d Hfstm'.
invert Hfstm'; [].
intros c1' c2' HfstM1 HfstM2 <-.
f_equal; auto; done.

(* pair *)
intros G M1 M2 c1 c2 _ IH1 _ IH2 d Hfstm'.
invert Hfstm'; [].
intros c1' c2' HfstM1 HfstM2 <-.
f_equal; auto; done.

(* pi1 *)
intros G M c _ IH d Hfstm'.
invert Hfstm'; [].
intros c' Hfst <-.
f_equal; auto; done.

(* pi2 *)
intros G M c _ IH d Hfstm'.
invert Hfstm'; [].
intros c' Hfst <-.
f_equal; auto; done.

(* lett *)
intros G e M c _ IH c' Hfstm'.
invert Hfstm'; [].
intro Hfst.
so (IH _ Hfst) as Heq.
f_apply (subst1 nonsense) in Heq as Heq'.
simp_sub in Heq'.
assumption.

(* cir *)
{
intros G M c' Hfstm'.
invert Hfstm'; [].
auto.
}
Qed.


Lemma linv_reg_gen :
  forall G P G' T,
    wf G
    -> ofps G P
    -> linv P G' T
    -> wfh G G' /\ ofs (G;; G') T.
Proof.
intros G P G' T Hwf HofP Hinv.
revert G Hwf HofP.
induct Hinv.

(* down *)
{
intros T G Hwf HofP.
invert HofP.
intro HofT.
simpl.
auto using wfh.
}

(* ex *)
{
intros k P G' T _ IH G Hwf Hofex.
invert Hofex; [].
intros Hofk HofP.
so (IH (G; cl_cn k) (wf_cn Hwf Hofk) HofP) as (Hwf' & HofT).
rewrite <- app_assoc.
simpl.
split; auto.
apply wfh_append; auto.
apply wfh_cn; auto using wfh.
}
Qed.


Lemma linv_reg :
  forall G P G' T,
    wf G
    -> ofps G P
    -> linv P G' T
    -> wf (G;; G') /\ ofs (G;; G') T.
Proof.
intros G P G' T Hwf Hof Hinv.
so (linv_reg_gen _#4 Hwf Hof Hinv) as (Hwfh & HofT).
split; auto.
apply wf_wfh_append; auto.
Qed.


Lemma linv_inhabitation_sof :
  forall G G' P T,
    ofps G P
    -> linv P G' T
    -> exists s, subof sof G s (G;; G') /\ compose (sh (length G')) s = id.
Proof.
intros G G' P T HofP Hinv.
revert HofP.
revert G.
induct Hinv.

(* down *)
{
intros _ G _.
exists id.
simpl.
split; auto.
apply subof_id.
}

(* exists *)
{
intros k P G' T _ IH G Hofex.
invert Hofex; [].
intros Hofk HofP.
so (IH (G; cl_cn k) HofP) as (s & Hs & Heqs).
exists (compose s (dot (inh k) id)).
split.
  {
  apply (subof_sof_trans _ (G; cl_cn k)).
    {
    apply subof_dot; auto using subof_id.
    simp_sub.
    apply sof_cn.
    apply inh_ofc; auto.
    }

    {
    rewrite <- app_assoc.
    simpl.
    exact Hs.
    }
  }
  
  {
  rewrite -> app_length.
  simpl length.
  replace (length G' + 1) with (1 + length G') by omega.
  rewrite <- compose_sh_sh.
  rewrite -> compose_assoc.
  setoid_rewrite <- compose_assoc at 2.
  rewrite -> Heqs.
  simp_sub.
  reflexivity.
  }
}
Qed.


Lemma dynamic_reg :
  (forall G e t,
     wf G
     -> oft G e t
     -> ofc G t kd_type)
  /\
  (forall G M T,
     wf G
     -> ofm G M T
     -> ofs G T
        /\ (forall c, fstm G M c -> exists k, fsts T k /\ ofc G c k))
  /\
  (forall G M P,
     wf G
     -> oflm G M P
     -> ofps G P).
Proof.
exploit
  (dynamic_ind
     (fun G e t =>
        wf G
        -> ofc G t kd_type)
     (fun G M T =>
        wf G
        -> ofs G T
           /\ (forall c, fstm G M c -> exists k, fsts T k /\ ofc G c k))
     (fun G M P =>
        wf G
        -> ofps G P)) as Hind.

(* var *)
intros G i t Hindex Hwf.
remember (cl_tm t) as A eqn:HeqA.
revert Hwf HeqA.
induct Hindex.
  (* 0 *)
  intros ? G Hwf ->.
  invert Hwf; [].
  intros _ Hofc.
  repl kd_type with (shift1 kd_type) by (simp_sub; reflexivity).
  apply weaken1_ofc; auto; done.

  (* S *)
  intros i B G ? _ IH Hwf ->.
  so (IH (wf_tail _ _ Hwf) (eq_refl _)) as H.
  so (weaken1_ofc _ B _ _ H) as H'.
  simp_sub in H'.
  assumption.

(* star *)
auto with static; done.

(* lam *)
intros G t1 t2 e Hoft1 _ IH Hwf.
apply ofc_arrow; auto; [].
so (IH (wf_tm Hwf Hoft1)) as Hoft2.
so (subst_ofc _#5 (unshift_tm _ _) Hoft2) as H.
simp_sub in H.
assumption.

(* app *)
intros G t1 t2 e1 e2 _ IH1 _ _ Hwf.
so (IH1 Hwf) as Hofc.
eapply ofc_arrow_invert; eauto; done.

(* pair *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
apply ofc_prod; auto; done.

(* pi1 *)
intros G t1 t2 e _ IH Hwf.
refine (ofc_prod_invert _#5 andel); eauto; done.

(* pi2 *)
intros G t1 t2 e _ IH Hwf.
eapply ofc_prod_invert; eauto; done.

(* plam *)
intros G k t e Hofk _ IH _ Hwf.
apply ofc_all; auto with static; done.

(* papp *)
intros G k t e c _ IH Hofc Hwf.
so (ofc_all_invert _#4 (IH Hwf)) as (_ & Hoft).
repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
eapply subst1_ofc; eauto; done.

(* pack *)
intros G k t c e Hofc _ IH Hoft Hwf.
eauto using ofc_reg with static; done.

(* unpack *)
{
intros G k t t' e1 e2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofex.
so (ofc_ofcr _#3 Hofex) as (l & H).
invert H.
intros Hofk Hoft _.
clear l H.
so (IH2 (wf_tm (wf_cn Hwf Hofk) Hoft)) as Hoft'.
eassert Prop as H; [refine (subst_ofc_type G _ (dot nonsense (dot (inh k) id)) _ _ Hoft') |].
  {
  apply subof_dot.
    {
    apply subof_dot; auto using subof_id.
    simp_sub.
    apply sof_cn.
    apply inh_ofc; auto.
    }
  simp_sub.
  apply sof_tm.
  }
simp_sub in H.
exact H.
}

(* fix *)
intros G t e _ IH Hwf.
exact (ofc_arrow_invert _#4 (IH Hwf) ander).

(* lett *)
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
sor (refine (IH2 _)) as Hoft2; eauto with static; [].
so (subst_ofc _#5 (unshift_tm _ _) Hoft2) as Hoft2'.
simp_sub in Hoft2'; assumption.

(* equiv *)
intros G t t' e _ _ Heqc Hwf.
exact (eqc_reg _#4 Hwf Heqc ander andel).

(* letm *)
{
intros G P G' T k t M e _ IH1 Hinv Hfst _ IH2 Hwf.
so (IH1 Hwf) as HofP.
so (linv_reg _#4 Hwf HofP Hinv) as (Hwf' & HofT).
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (IH2 (wf_md Hwf' HofT Hfst)) as Hoft.
so (linv_inhabitation_sof _#4 HofP Hinv) as (s & Hs & Heqs).
assert (subof sof G (dot nonsense (dot (inh (subst s k)) s)) (G;; G'; cl_cn k; cl_md (shift1 T) (var 0))) as Hs'.
  {
  apply subof_dot.
    {
    apply subof_dot; auto.
    simp_sub.
    apply sof_cn.
    apply inh_ofc.
    eapply subst_ofk; eauto.
    }
  simp_sub.
  apply sof_md; auto.
  }
so (subst_ofc_type _#4 Hs' Hoft) as Hoft'.
simp_sub in Hoft'.
replace (2 + length G') with (S (S (length G'))) in Hoft' by omega.
simp_sub in Hoft'.
rewrite -> Heqs in Hoft'.
simp_sub in Hoft'.
exact Hoft'.
}

(* ext *)
intros G t M _ IH Hwf.
so (IH Hwf andel) as Hofs.
invert Hofs; auto; done.

(* var *)
intros G i T c Hindex Hwf.
split.
  remember (cl_md T c) as A eqn:HeqA.
  revert Hwf HeqA.
  induct Hindex.
    (* 0 *)
    intros ? G Hwf ->.
    invert Hwf; [].
    intros G' k T' _ Hofs _ <- <- <-.
    simp_sub.
    eapply subst_ofs; eauto; [].
    apply subof_sh; auto with sequence; done.
  
    (* S *)
    intros i B G ? _ IH Hwf ->.
    so (IH (wf_tail _ _ Hwf) (eq_refl _)) as H.
    so (weaken1_ofs _ B _ H) as H'.
    simp_sub in H'.
    assumption.

  intros c' Hfstm.
  #2 invert Hfstm.
    intros T' c'' Hindex' <-.
    so (index_fun _#5 Hindex Hindex') as Heq.
    injection Heq; clear Heq; intros; subst T' c''.
    clear Hfstm Hindex'.
    remember (cl_md T c) as A eqn:HeqA.
    revert Hwf HeqA.
    induct Hindex.
      (* 0 *)
      intros ? G Hwf ->.
      invert Hwf; [].
      intros G' k' T' _ Hofs Hfsts' <- <- <-.
      exists (subst (sh 2) k').
      simp_sub.
      split.
        apply subst_fsts; auto; done.
        
        apply ofc_var; auto with sequence; done.
    
      (* S *)
      intros i B G ? _ IH Hwf ->.
      so (IH (wf_tail _ _ Hwf) (eq_refl _)) as (k & Hfsts & Hofc).
      exists (shift1 k).
      split.
        so (subst_fsts _ _ sh1 Hfsts) as H.
        simp_sub in H.
        assumption.
    
        so (weaken1_ofc _ B _ _ Hofc) as H.
        simp_sub in H.
        assumption.

    intros c'' Hindex' <-.
    so (index_fun _#5 Hindex Hindex') as Heq.
    discriminate Heq.

(* star *)
intros G Hwf.
split.
  auto with static; done.

  intros c Hfstm.
  invert Hfstm; [].
  intros; subst c.
  eexists; split; eauto with static; done.

(* stat *)
intros G c k Hofc Hwf.
split.
  apply ofs_stat; [].
  eapply ofc_reg; eauto; done.

  intros c' Hfstm.
  invert Hfstm.
  intro; subst c'.
  exists k; auto with static; done.

(* dyn *)
intros G e t _ IH Hwf.
split.
  apply ofs_dyn; auto; done.

  intros c Hfstm.
  invert Hfstm.
  intro; subst c.
  exists kd_one; auto with static; done.

(* lam *)
intros G T1 T2 k1 M HofT1 HfstT1 _ IH Hwf.
split.
  eapply ofs_pi; eauto; [].
  sor (refine (IH _ andel)) as HofT2w; eauto using ofs_fst with static; [].
  so (subst_ofs _#4 (unshift_md _#3) HofT2w) as HofT2; simp_sub in HofT2.
  assumption.

  intros c Hfstm.
  invert Hfstm; [].
  intros k1' c1 HfstT1' Hfstm' <-.
  so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
  sor (refine (IH _ ander _ Hfstm')) as (k2w & HfstT2w & Hof); eauto with static; [].
  so (subst_ofc _#5 (unshift_md _#3) Hof) as Hof'.
  simp_sub in Hof'.
  exists (kd_pi k1 (subst1 nonsense k2w)).
  split.
    apply fsts_pi; auto; [].
    so (subst_fsts _ _ (dot nonsense id) HfstT2w) as HfstT2.
    simp_sub in HfstT2.
    assumption.

    apply ofc_lam; eauto using ofs_fst; done.

(* app *)
intros G T1 T2 M1 M2 c2 _ IH1 _ IH2 HfstM2 Hwf.
split.
  so (IH1 Hwf andel) as Hofs.
  invert Hofs.
  intros k1 HofT1 HfstT1 HofT2.
  eapply subst1_ofs; eauto; [].
  so (IH2 Hwf ander c2 HfstM2) as (k1' & HfstT1' & Hofc).
  so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
  assumption.
  
  intros c Hfstm.
  invert Hfstm; [].
  intros c1 c2' HfstM1 HfstM2' <-.
  so (fstm_fun _#4 HfstM2 HfstM2'); subst c2'.
  so (IH1 Hwf ander _ HfstM1) as (k & Hfsts & Hofc1).
  invert Hfsts.
  intros k1 k2 HfstT1 HfstT2 <-.
  so (IH2 Hwf ander _ HfstM2) as (k1' & HfstT1' & Hofc2).
  so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
  exists (subst1 c2 k2).
  split.
    apply subst_fsts; auto; done.
  
    eapply ofc_app; eauto; done.

(* pair *)
intros G T1 T2 M1 M2 _ IH1 _ IH2 Hwf.
split.
  so (IH1 Hwf andel) as HofT1.
  so (IH2 Hwf andel) as HofT2.
  so (can_fsts _ _ HofT1) as (k1 & HfstT1).
  eapply ofs_sigma; eauto; [].
  apply weaken1_ofs; assumption.

  intros c Hfstm.
  invert Hfstm; [].
  intros c1 c2 HfstM1 HfstM2 <-.
  so (IH1 Hwf ander _ HfstM1) as (k1 & HfstT1 & Hofc1).
  so (IH2 Hwf ander _ HfstM2) as (k2 & HfstT2 & Hofc2).
  exists (kd_sigma k1 (shift1 k2)).
  split; eauto using subst_fsts with static; [].
  #2 apply ofc_pair; auto.
    simp_sub.
    assumption.
  
    apply weaken1_ofk; [].
    eapply ofc_reg; eauto; done.

(* pi1 *)
intros G T1 T2 M _ IH Hwf.
split.
  so (IH Hwf andel) as Hofs.
  invert Hofs; auto; done.

  intros d Hfstm.
  invert Hfstm; [].
  intros c HfstM <-.
  so (IH Hwf ander _ HfstM) as (k & Hfsts & Hofc).
  invert Hfsts.
  intros k1 k2 HfstT1 HfstT2 <-.
  exists k1.
  split; eauto with static; done.

(* pi2 *)
intros G T1 T2 M c _ IH HfstM Hwf.
split.
  so (IH Hwf andel) as Hofs.
  invert Hofs.
  intros k1 HofT1 HfstT1 HofT2.
  eapply subst1_ofs; eauto; [].
  so (IH Hwf ander _ HfstM) as (l & Hfsts & Hofc).
  invert Hfsts.
  intros k1' k2 HfstT1' _ <-.
  so (fsts_fun _#3 HfstT1 HfstT1'); subst k1'.
  eapply ofc_pi1; eauto; done.
  
  intros d Hfstm.
  invert Hfstm; [].
  intros c' HfstM' <-.
  so (fstm_fun _#4 HfstM HfstM'); subst c'.
  so (IH Hwf ander _ HfstM) as (k & Hfsts & Hofc).
  invert Hfsts.
  intros k1 k2 HfstT1 HfstT2 <-.
  exists (subst1 (cn_pi1 c) k2).
  split; eauto using subst_fsts with static; done.

(* lett *)
intros G t T e M _ IH1 _ IH2 Hwf.
split.
  sor (refine (IH2 _ andel)) as Hofs; eauto with static; [].
  so (subst_ofs _#4 (unshift_tm _ _) Hofs) as H; simp_sub in H.
  assumption.

  intros c Hfstm.
  invert Hfstm; [].
  intros HfstM.
  so (IH1 Hwf) as Hoft.
  assert (subof dof (G; cl_tm t) (dot (var 0) sh1) (G; cl_utm)) as Hs.
    apply subof_dot; eauto using dof with dynamic sequence; [].
    apply subof_sh; auto with sequence; done.
  so (subst_fstm _#5 (subof_dof_tof _#3 Hs) HfstM) as HfstM'.
  simp_sub in HfstM'.
  sor (refine (IH2 _ ander _ HfstM')) as (k & Hfsts & Hofc); auto with static; [].
  so (subst_ofc _#5 (unshift_tm _ _) Hofc) as Hofc'; simp_sub in Hofc'.
  exists (subst1 nonsense k).
  split.
    so (subst_fsts _ _ (dot nonsense id) Hfsts) as Hfsts'.
    simp_sub in Hfsts'.
    assumption.

    assumption.

(* cir *)
{
intros G M P _ IH Hwf.
so (IH Hwf) as HofP.
split; eauto using ofs_circ.
intros c Hfstm.
invert Hfstm; [].
intros <-.
exists kd_one.
split; eauto with static.
}

(* extstat *)
intros G k k' M c _ IH Hfstm Hofc Hwf.
split.
  apply ofs_stat; [].
  eapply ofc_reg; eauto; done.

  intros c' Hfstm'.
  so (fstm_fun _#4 Hfstm Hfstm'); subst c'.
  exists k.
  auto with static; done.

(* extpi *)
intros G T1 T2 T3 k1 M _ IH1 HfstT1 _ IH2 Hwf.
split.
  so (IH1 Hwf andel) as Hofs.
  invert Hofs.
  intros _ HofT1 _ _.
  sor (refine (IH2 _ andel)) as HofT2w; eauto using ofs_fst with static; [].
  so (subst_ofs _#4 (unshift_md _#3) HofT2w) as HofT2; simp_sub in HofT2.
  eapply ofs_pi; eauto; done.

  intros c Hfstm.
  invert (IH1 Hwf andel).
  intros _ HofT1 _ _.
  #2 sor (refine (IH2 _ ander (cn_app (subst (sh 2) c) (var 1)) _)) as (k2w & HfstT2w & Hofc); auto with static.
    #2 apply fstm_app.
      eapply subst_fstm; eauto; [].
      apply subof_sh; auto with sequence; done.

      sor (refine (@fstm_var _ _ _ (var 0) index_0)) as H.
      simp_sub in H.
      eassumption.

    so (subst_ofc _#5 (unshift_md _#3) Hofc) as Hofc'; simp_sub in Hofc'.
    exists (kd_pi k1 (subst1 nonsense k2w)).
    split.
      apply fsts_pi; auto; [].
      so (subst_fsts _ _ (dot nonsense id) HfstT2w) as HfstT2.
      simp_sub in HfstT2.
      assumption.

      so (IH1 Hwf ander _ Hfstm) as (l & Hfsts & Hofcalt).
      invert Hfsts; [].
      intros k1' k3 HfstT1' HfstT3 <-.
      so (fsts_fun _ _ _ HfstT1 HfstT1'); subst k1'.
      apply (@ofc_extpi _#3 k3); auto; done.

(* extsigma *)
intros G T1 T2 M _ IH1 _ IH2 Hwf.
split.
  so (IH1 Hwf andel) as HofT1.
  so (IH2 Hwf andel) as HofT2.
  so (can_fsts _ _ HofT1) as (k1 & HfstT1).
  eapply ofs_sigma; eauto; [].
  apply weaken1_ofs; assumption.

  intros c Hfstm.
  so (IH1 Hwf ander _ (fstm_pi1 Hfstm)) as (k1 & HfstT1 & Hofc1).
  so (IH2 Hwf ander _ (fstm_pi2 Hfstm)) as (k2 & HfstT2 & Hofc2).
  exists (kd_sigma k1 (shift1 k2)).
  split; eauto using subst_fsts with static; [].
  #2 apply ofc_extsigma; auto.
    simp_sub; assumption.

    apply weaken1_ofk; [].
    eapply ofc_reg; eauto; done.

(* subsume *)
intros G T T' M _ IH Hsubs Hwf.
split.
  exact (subs_reg _#3 Hwf Hsubs ander).

  intros c Hfstm.
  so (IH Hwf ander c Hfstm) as (k & Hfsts & Hofc).
  so (subs_fst_o _#3 Hwf Hsubs) as (ka & k' & Hfstsa & Hfsts' & Hsubk).
  so (fsts_fun _ _ _ Hfsts Hfstsa); subst ka.
  exists k'; split; eauto with static; done.

(* ret *)
{
intros G M T _ IH Hwf.
so (IH Hwf) as (Hof & _).
apply ofps_down; auto.
}

(* seal *)
{
intros G M T _ IH Hwf.
so (IH Hwf) as (Hof & _).
apply ofps_down; auto.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 _ IH1 Hinv Hfst _ IH2 Hwf.
so (IH1 Hwf) as (Hof1 & _).
invert Hof1; [].
intros HofP1.
clear Hof1.
assert (exists s, subof sof G s (G;; G') /\ compose (sh (length G')) s = id) as (s & Hs & Heqs).
  {
  clear M1 M2 k P2 IH1 Hfst IH2 Hwf.
  revert HofP1.
  generalize G.
  clear G.
  induct Hinv.
    {
    intros _ G _.
    exists id.
    simpl.
    split; auto.
    apply subof_id.
    }

    {
    intros k P G' T _ IH G Hofex.
    invert Hofex; [].
    intros Hofk HofP.
    so (IH (G; cl_cn k) HofP) as (s & Hs & Heqs).
    exists (compose s (dot (inh k) id)).
    split.
      {
      apply (subof_sof_trans _ (G; cl_cn k)).
        {
        apply subof_dot; auto using subof_id.
        simp_sub.
        apply sof_cn.
        apply inh_ofc; auto.
        }

        {
        rewrite <- app_assoc.
        simpl.
        exact Hs.
        }
      }
      
      {
      rewrite -> app_length.
      simpl length.
      replace (length G' + 1) with (1 + length G') by omega.
      rewrite <- compose_sh_sh.
      rewrite -> compose_assoc.
      setoid_rewrite <- compose_assoc at 2.
      rewrite -> Heqs.
      simp_sub.
      reflexivity.
      }
    }
  }
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
so (IH2 (wf_md Hwf' HofT Hfst)) as HofP2w.
eassert Prop as HofP2; [refine (subst_ofps G _ (dot nonsense (dot (inh (subst s k)) s)) _ _ HofP2w) |].
  {
  apply subof_dot.
    {
    apply subof_dot; auto.
    simp_sub.
    apply sof_cn.
    apply inh_ofc.
    so (ofs_fst _#3 HofT Hfst) as Hofk.
    exact (subst_ofk _#4 Hs Hofk).
    }
  simp_sub.
  apply sof_md.
  }
simp_sub in HofP2.
replace (2 + length G') with (S (S (length G'))) in HofP2 by reflexivity.
rewrite -> compose_sh_succ_dot in HofP2.
rewrite -> compose_sh_succ_dot in HofP2.
rewrite -> Heqs in HofP2.
simp_sub in HofP2.
exact HofP2.
}

(* unpack *)
{
intros G k t P e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hofex.
so (ofc_ofcr _#3 Hofex) as (l & H).
invert H; [].
intros Hofk Hoft _.
clear l H.
so (IH2 (wf_tm (wf_cn Hwf Hofk) Hoft)) as HofPw.
assert (subof sof G (dot nonsense (dot (inh k) id)) (G; cl_cn k; cl_tm t)) as Hs.
  {
  apply subof_dot.
    {
    apply subof_dot; auto using subof_id.
    simp_sub.
    apply sof_cn.
    apply inh_ofc; auto.
    }
  simp_sub.
  apply sof_tm.
  }
so (subst_ofps _#4 Hs HofPw) as HofP.
simp_sub in HofP.
exact HofP.
}

(* subsume *)
{
intros G P P' M _ _ Hsub Hwf.
so (subps_reg _#3 Hwf Hsub) as (_ & H).
exact H.
}

(* epilogue *)
destruct_all Hind.
do2 2 split; intros; eauto; done.
Qed.


Lemma oft_reg :
  forall G e t,
    wf G
    -> oft G e t
    -> ofc G t kd_type.
Proof.
exact (dynamic_reg andel).
Qed.


Lemma ofm_reg :
  forall G M T,
    wf G
    -> ofm G M T
    -> ofs G T.
Proof.
intros G M T Hwf Hof.
exact (dynamic_reg anderl G M T Hwf Hof andel).
Qed.


Lemma oflm_reg :
  forall G M T,
    wf G
    -> oflm G M T
    -> ofps G T.
Proof.
intros G M T Hwf Hof.
exact (dynamic_reg anderr G M T Hwf Hof).
Qed.


Lemma ofm_fst_o :
  forall G M T c,
    wf G
    -> ofm G M T
    -> fstm G M c
    -> exists k, fsts T k /\ ofc G c k.
Proof.
intros G M T c Hwf Hof Hfst.
exact (dynamic_reg anderl G M T Hwf Hof ander c Hfst).
Qed.


Lemma ofm_fst :
  forall G M T c k,
    wf G
    -> ofm G M T
    -> fstm G M c
    -> fsts T k
    -> ofc G c k.
Proof.
intros G M T c k Hwf Hofm Hfstm Hfsts.
so (ofm_fst_o _#4 Hwf Hofm Hfstm) as (k' & Hfsts' & Hof).
so (fsts_fun _ _ _ Hfsts Hfsts'); subst k'.
assumption.
Qed.


Lemma static_wf_ind :
  forall (P : context -> term -> Prop)
    (P0 P1 P2 : context -> term -> term -> Prop)
    (P3 : context -> term -> term -> term -> Prop),
  (forall G : context, wf G -> P G kd_type) ->
  (forall (G : context) (c : term),
   wf G -> ofc G c kd_type -> P2 G c kd_type -> P G (kd_sing c)) ->
  (forall (G : context) (k1 k2 : term),
   wf G -> ofk G k1 ->
   P G k1 ->
   ofk (G; cl_cn k1) k2 -> P (G; cl_cn k1) k2 -> P G (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 : term),
   wf G -> ofk G k1 ->
   P G k1 ->
   ofk (G; cl_cn k1) k2 -> P (G; cl_cn k1) k2 -> P G (kd_sigma k1 k2)) ->
  (forall G : context, wf G -> P G kd_one) ->
  (forall (G : context) (k : term), wf G -> ofk G k -> P G k -> P0 G k k) ->
  (forall (G : context) (k k' : term),
   wf G -> eqk G k k' -> P0 G k k' -> P0 G k' k) ->
  (forall (G : context) (k1 k2 k3 : term),
   wf G -> eqk G k1 k2 -> P0 G k1 k2 -> eqk G k2 k3 -> P0 G k2 k3 -> P0 G k1 k3) ->
  (forall (G : context) (c c' : term),
   wf G -> eqc G c c' kd_type ->
   P3 G c c' kd_type -> P0 G (kd_sing c) (kd_sing c')) ->
  (forall (G : context) (k1 k1' k2 k2' : term),
   wf G -> eqk G k1 k1' ->
   P0 G k1 k1' ->
   eqk (G; cl_cn k1) k2 k2' ->
   P0 (G; cl_cn k1) k2 k2' -> P0 G (kd_pi k1 k2) (kd_pi k1' k2')) ->
  (forall (G : context) (k1 k1' k2 k2' : term),
   wf G -> eqk G k1 k1' ->
   P0 G k1 k1' ->
   eqk (G; cl_cn k1) k2 k2' ->
   P0 (G; cl_cn k1) k2 k2' -> P0 G (kd_sigma k1 k2) (kd_sigma k1' k2')) ->
  (forall (G : context) (k k' : term),
   wf G -> eqk G k k' -> P0 G k k' -> P1 G k k') ->
  (forall (G : context) (k1 k2 k3 : term),
   wf G -> subk G k1 k2 ->
   P1 G k1 k2 -> subk G k2 k3 -> P1 G k2 k3 -> P1 G k1 k3) ->
  (forall (G : context) (c : term),
   wf G -> ofc G c kd_type -> P2 G c kd_type -> P1 G (kd_sing c) kd_type) ->
  (forall (G : context) (k1 k1' k2 k2' : term),
   wf G -> subk G k1' k1 ->
   P1 G k1' k1 ->
   subk (G; cl_cn k1') k2 k2' ->
   P1 (G; cl_cn k1') k2 k2' ->
   ofk (G; cl_cn k1) k2 ->
   P (G; cl_cn k1) k2 -> P1 G (kd_pi k1 k2) (kd_pi k1' k2')) ->
  (forall (G : context) (k1 k1' k2 k2' : term),
   wf G -> subk G k1 k1' ->
   P1 G k1 k1' ->
   subk (G; cl_cn k1) k2 k2' ->
   P1 (G; cl_cn k1) k2 k2' ->
   ofk (G; cl_cn k1') k2' ->
   P (G; cl_cn k1') k2' -> P1 G (kd_sigma k1 k2) (kd_sigma k1' k2')) ->
  (forall (G : list term) (i : nat) (k : term),
   wf G -> index i G (cl_cn k) -> P2 G (var i) (subst (sh (S i)) k)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofk G k1 ->
   P G k1 ->
   ofc (G; cl_cn k1) c k2 ->
   P2 (G; cl_cn k1) c k2 -> P2 G (cn_lam k1 c) (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 (kd_pi k1 k2) ->
   P2 G c1 (kd_pi k1 k2) ->
   ofc G c2 k1 -> P2 G c2 k1 -> P2 G (cn_app c1 c2) (subst1 c2 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 k1 ->
   P2 G c1 k1 ->
   ofc G c2 (subst1 c1 k2) ->
   P2 G c2 (subst1 c1 k2) ->
   ofk (G; cl_cn k1) k2 ->
   P (G; cl_cn k1) k2 -> P2 G (cn_pair c1 c2) (kd_sigma k1 k2)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G c (kd_sigma k1 k2) ->
   P2 G c (kd_sigma k1 k2) -> P2 G (cn_pi1 c) k1) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G c (kd_sigma k1 k2) ->
   P2 G c (kd_sigma k1 k2) -> P2 G (cn_pi2 c) (subst1 (cn_pi1 c) k2)) ->
  (forall G : context, wf G -> P2 G cn_star kd_one) ->
  (forall G : context, wf G -> P2 G cn_unit kd_type) ->
  (forall (G : context) (t1 t2 : term),
   wf G -> ofc G t1 kd_type ->
   P2 G t1 kd_type ->
   ofc G t2 kd_type -> P2 G t2 kd_type -> P2 G (cn_arrow t1 t2) kd_type) ->
  (forall (G : context) (t1 t2 : term),
   wf G -> ofc G t1 kd_type ->
   P2 G t1 kd_type ->
   ofc G t2 kd_type -> P2 G t2 kd_type -> P2 G (cn_prod t1 t2) kd_type) ->
  (forall (G : context) (k t : term),
   wf G -> ofk G k ->
   P G k ->
   ofc (G; cl_cn k) t kd_type ->
   P2 (G; cl_cn k) t kd_type -> P2 G (cn_all k t) kd_type) ->
  (forall (G : context) (k t : term),
   wf G -> ofk G k ->
   P G k ->
   ofc (G; cl_cn k) t kd_type ->
   P2 (G; cl_cn k) t kd_type -> P2 G (cn_exists k t) kd_type) ->
  (forall (G : context) (c : term),
   wf G -> ofc G c kd_type -> P2 G c kd_type -> P2 G c (kd_sing c)) ->
  (forall (G : context) (k1 k2 k3 c : term),
   wf G -> ofc G c (kd_pi k1 k3) ->
   P2 G c (kd_pi k1 k3) ->
   ofc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2 ->
   P2 (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2 ->
   P2 G c (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G (cn_pi1 c) k1 ->
   P2 G (cn_pi1 c) k1 ->
   ofc G (cn_pi2 c) (subst1 (cn_pi1 c) k2) ->
   P2 G (cn_pi2 c) (subst1 (cn_pi1 c) k2) ->
   ofk (G; cl_cn k1) k2 -> P (G; cl_cn k1) k2 -> P2 G c (kd_sigma k1 k2)) ->
  (forall (G : context) (k k' c : term),
   wf G -> ofc G c k -> P2 G c k -> subk G k k' -> P1 G k k' -> P2 G c k') ->
  (forall (G : context) (c k : term),
   wf G -> ofc G c k -> P2 G c k -> P3 G c c k) ->
  (forall (G : context) (c c' k : term),
   wf G -> eqc G c c' k -> P3 G c c' k -> P3 G c' c k) ->
  (forall (G : context) (c1 c2 c3 k : term),
   wf G -> eqc G c1 c2 k ->
   P3 G c1 c2 k -> eqc G c2 c3 k -> P3 G c2 c3 k -> P3 G c1 c3 k) ->
  (forall (G : context) (k1 k1' k2 c c' : term),
   wf G -> eqk G k1 k1' ->
   P0 G k1 k1' ->
   eqc (G; cl_cn k1) c c' k2 ->
   P3 (G; cl_cn k1) c c' k2 ->
   P3 G (cn_lam k1 c) (cn_lam k1' c') (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 c1' c2' : term),
   wf G -> eqc G c1 c1' (kd_pi k1 k2) ->
   P3 G c1 c1' (kd_pi k1 k2) ->
   eqc G c2 c2' k1 ->
   P3 G c2 c2' k1 -> P3 G (cn_app c1 c2) (cn_app c1' c2') (subst1 c2 k2)) ->
  (forall (G : context) (k1 k2 c1 c1' c2 c2' : term),
   wf G -> eqc G c1 c1' k1 ->
   P3 G c1 c1' k1 ->
   eqc G c2 c2' (subst1 c1 k2) ->
   P3 G c2 c2' (subst1 c1 k2) ->
   ofk (G; cl_cn k1) k2 ->
   P (G; cl_cn k1) k2 ->
   P3 G (cn_pair c1 c2) (cn_pair c1' c2') (kd_sigma k1 k2)) ->
  (forall (G : context) (k1 k2 c c' : term),
   wf G -> eqc G c c' (kd_sigma k1 k2) ->
   P3 G c c' (kd_sigma k1 k2) -> P3 G (cn_pi1 c) (cn_pi1 c') k1) ->
  (forall (G : context) (k1 k2 c c' : term),
   wf G -> eqc G c c' (kd_sigma k1 k2) ->
   P3 G c c' (kd_sigma k1 k2) ->
   P3 G (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2)) ->
  (forall (G : context) (t1 t1' t2 t2' : term),
   wf G -> eqc G t1 t1' kd_type ->
   P3 G t1 t1' kd_type ->
   eqc G t2 t2' kd_type ->
   P3 G t2 t2' kd_type ->
   P3 G (cn_arrow t1 t2) (cn_arrow t1' t2') kd_type) ->
  (forall (G : context) (t1 t1' t2 t2' : term),
   wf G -> eqc G t1 t1' kd_type ->
   P3 G t1 t1' kd_type ->
   eqc G t2 t2' kd_type ->
   P3 G t2 t2' kd_type -> P3 G (cn_prod t1 t2) (cn_prod t1' t2') kd_type) ->
  (forall (G : context) (k k' t t' : term),
   wf G -> eqk G k k' ->
   P0 G k k' ->
   eqc (G; cl_cn k) t t' kd_type ->
   P3 (G; cl_cn k) t t' kd_type ->
   P3 G (cn_all k t) (cn_all k' t') kd_type) ->
  (forall (G : context) (k k' t t' : term),
   wf G -> eqk G k k' ->
   P0 G k k' ->
   eqc (G; cl_cn k) t t' kd_type ->
   P3 (G; cl_cn k) t t' kd_type ->
   P3 G (cn_exists k t) (cn_exists k' t') kd_type) ->
  (forall (G : context) (c c' : term),
   wf G -> eqc G c c' kd_type -> P3 G c c' kd_type -> P3 G c c' (kd_sing c)) ->
  (forall (G : context) (c c' : term),
   wf G -> ofc G c (kd_sing c') -> P2 G c (kd_sing c') -> P3 G c c' kd_type) ->
  (forall (G : context) (k1 k2 k3 k3' c c' : term),
   wf G -> ofc G c (kd_pi k1 k3) ->
   P2 G c (kd_pi k1 k3) ->
   ofc G c' (kd_pi k1 k3') ->
   P2 G c' (kd_pi k1 k3') ->
   eqc (G; cl_cn k1) (cn_app (shift1 c) (var 0))
     (cn_app (shift1 c') (var 0)) k2 ->
   P3 (G; cl_cn k1) (cn_app (shift1 c) (var 0))
     (cn_app (shift1 c') (var 0)) k2 -> P3 G c c' (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 k3 c c' : term),
   wf G -> eqc G c c' (kd_pi k1 k3) ->
   P3 G c c' (kd_pi k1 k3) ->
   eqc (G; cl_cn k1) (cn_app (shift1 c) (var 0))
     (cn_app (shift1 c') (var 0)) k2 ->
   P3 (G; cl_cn k1) (cn_app (shift1 c) (var 0))
     (cn_app (shift1 c') (var 0)) k2 -> P3 G c c' (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c c' : term),
   wf G -> eqc G (cn_pi1 c) (cn_pi1 c') k1 ->
   P3 G (cn_pi1 c) (cn_pi1 c') k1 ->
   eqc G (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2) ->
   P3 G (cn_pi2 c) (cn_pi2 c') (subst1 (cn_pi1 c) k2) ->
   ofk (G; cl_cn k1) k2 ->
   P (G; cl_cn k1) k2 -> P3 G c c' (kd_sigma k1 k2)) ->
  (forall (G : context) (c c' : term),
   wf G -> ofc G c kd_one ->
   P2 G c kd_one ->
   ofc G c' kd_one -> P2 G c' kd_one -> P3 G c c' kd_one) ->
  (forall (G : context) (c c' k k' : term),
   wf G -> eqc G c c' k ->
   P3 G c c' k -> subk G k k' -> P1 G k k' -> P3 G c c' k') ->
  (forall (G : list term) (k1 k2 c1 c2 : term),
   wf G -> ofc (G; cl_cn k1) c2 k2 ->
   P2 (G; cl_cn k1) c2 k2 ->
   ofc G c1 k1 ->
   P2 G c1 k1 ->
   P3 G (cn_app (cn_lam k1 c2) c1) (subst1 c1 c2) (subst1 c1 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 k1 ->
   P2 G c1 k1 ->
   ofc G c2 k2 -> P2 G c2 k2 -> P3 G (cn_pi1 (cn_pair c1 c2)) c1 k1) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 k1 ->
   P2 G c1 k1 ->
   ofc G c2 k2 -> P2 G c2 k2 -> P3 G (cn_pi2 (cn_pair c1 c2)) c2 k2) ->
  (forall (c : context) (t : term), wf c -> ofk c t -> P c t) /\
  (forall (c : context) (t t0 : term), wf c -> eqk c t t0 -> P0 c t t0) /\
  (forall (c : context) (t t0 : term), wf c -> subk c t t0 -> P1 c t t0) /\
  (forall (c : context) (t t0 : term), wf c -> ofc c t t0 -> P2 c t t0) /\
  (forall (c : context) (t t0 t1 : term), wf c -> eqc c t t0 t1 -> P3 c t t0 t1)
.
Proof.
intros P1 P2 P3 P4 P5.
intros.
#16 exploit
  (static_ind
     (fun G k => wf G -> P1 G k)
     (fun G k k' => wf G -> P2 G k k')
     (fun G k k' => wf G -> P3 G k k')
     (fun G c k => wf G -> P4 G c k)
     (fun G c c' k => wf G -> P5 G c c' k)) as Hind;
try (intros;
     eauto using wf;
     done).

(* pi *)
intros G k1 k1' k2 k2' Heqk ? ? ? Hwf.
so (eqk_reg _#3 Hwf Heqk) as (? & _).
eauto using wf; done.

(* sigma *)
intros G k1 k1' k2 k2' Heqk ? ? ? Hwf.
so (eqk_reg _#3 Hwf Heqk) as (? & _).
eauto using wf; done.

(* pi *)
intros G k1 k1' k2 k2' Hsubk ? ? ? ? ? Hwf.
so (subk_reg _#3 Hwf Hsubk) as (? & ?).
eauto 7 using wf; done.

(* sigma *)
intros G k1 k1' k2 k2' Hsubk ? ? ? ? ? Hwf.
so (subk_reg _#3 Hwf Hsubk) as (? & ?).
eauto 7 using wf; done.

(* pair *)
intros G k1 k2 c1 c2 Hof ? ? ? ? ? Hwf.
so (ofc_reg _#3 Hwf Hof).
eauto 7 using wf; done.

(* extpi *)
intros G k1 k2 k3 c Hof ? ? ? Hwf.
so (ofc_reg _#3 Hwf Hof) as Hofk.
invert Hofk; [].
intros; eauto using wf; done.

(* extsigma *)
intros G k1 k2 c Hof ? ? ? ? ? Hwf.
so (ofc_reg _#3 Hwf Hof).
eauto 7 using wf; done.

(* lam *)
intros G k1 k1' k2 c c' Heqk ? ? ? Hwf.
so (eqk_reg _#3 Hwf Heqk) as (? & _).
eauto using wf; done.

(* pair *)
intros G k1 k2 c1 c1' c2 c2' Heqc ? ? ? ? ? Hwf.
so (eqc_reg _#4 Hwf Heqc) as (_ & _ & ?).
eauto 7 using wf; done.

(* all *)
intros G k k' t t' Heqk ? ? ? Hwf.
so (eqk_reg _#3 Hwf Heqk) as (? & _).
eauto using wf; done.

(* exists *)
intros G k k' t t' Heqk ? ? ? Hwf.
so (eqk_reg _#3 Hwf Heqk) as (? & _).
eauto using wf; done.

(* extpi *)
intros G k1 k2 k3 k3' c c' Hof ? ? ? ? ? Hwf.
so (ofc_reg _#3 Hwf Hof) as Hofk.
invert Hofk; [].
intros; eauto 7 using wf; done.

(* extpiw *)
intros G k1 k2 k3 c c' Heqc ? ? ? Hwf.
so (eqc_reg _#4 Hwf Heqc) as (_ & _ & Hofk).
invert Hofk; [].
intros; eauto using wf; done.

(* extsigma *)
intros G k1 k2 c c' Heqc ? ? ? ? ? Hwf.
so (eqc_reg _#4 Hwf Heqc) as (_ & _ & ?).
eauto 7 using wf; done.

(* beta *)
intros G k1 k2 c1 c2 ? ? Hof ? Hwf.
so (ofc_reg _#3 Hwf Hof).
eauto using wf; done.

(* epilogue *)
destruct_all Hind.
split; auto; done.
Qed.


Lemma dynamic_wf_ind
     : forall P P0 P1 : context -> term -> term -> Prop,
       (forall (G : list term) (i : nat) (t : term),
        wf G -> index i G (cl_tm t) -> P G (var i) (subst (sh (S i)) t)) ->
       (forall G : context, wf G -> P G tm_star cn_unit) ->
       (forall (G : context) (t1 t2 e : term),
        wf G -> ofc G t1 kd_type ->
        oft (G; cl_tm t1) e (shift1 t2) ->
        P (G; cl_tm t1) e (shift1 t2) -> P G (tm_lam t1 e) (cn_arrow t1 t2)) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        wf G -> oft G e1 (cn_arrow t1 t2) ->
        P G e1 (cn_arrow t1 t2) -> oft G e2 t1 -> P G e2 t1 -> P G (tm_app e1 e2) t2) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        wf G -> oft G e1 t1 ->
        P G e1 t1 -> oft G e2 t2 -> P G e2 t2 -> P G (tm_pair e1 e2) (cn_prod t1 t2)) ->
       (forall (G : context) (t1 t2 e : term),
        wf G -> oft G e (cn_prod t1 t2) -> P G e (cn_prod t1 t2) -> P G (tm_pi1 e) t1) ->
       (forall (G : context) (t1 t2 e : term),
        wf G -> oft G e (cn_prod t1 t2) -> P G e (cn_prod t1 t2) -> P G (tm_pi2 e) t2) ->
       (forall (G : context) (k t e : term),
        wf G -> ofk G k ->
        oft (G; cl_cn k) e t -> P (G; cl_cn k) e t -> svalue e -> P G (tm_plam k e) (cn_all k t)) ->
       (forall (G : context) (k t e c : term),
        wf G -> oft G e (cn_all k t) -> P G e (cn_all k t) -> ofc G c k -> P G (tm_papp e c) (subst1 c t)) ->
       (forall (G : context) (k t c e : term),
        wf G -> ofc G c k ->
        oft G e (subst1 c t) ->
        P G e (subst1 c t) -> ofc (G; cl_cn k) t kd_type -> P G (tm_pack c e k t) (cn_exists k t)) ->
       (forall (G : context) (k t t' e1 e2 : term),
        wf G -> oft G e1 (cn_exists k t) ->
        P G e1 (cn_exists k t) ->
        oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') ->
        P (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t') -> P G (tm_unpack e1 e2) t') ->
       (forall (G : context) (t e : term),
        wf G -> oft G e (cn_arrow (cn_arrow cn_unit t) t) ->
        P G e (cn_arrow (cn_arrow cn_unit t) t) -> P G (tm_fix t e) t) ->
       (forall (G : context) (t1 t2 e1 e2 : term),
        wf G -> oft G e1 t1 ->
        P G e1 t1 ->
        oft (G; cl_tm t1) e2 (shift1 t2) ->
        P (G; cl_tm t1) e2 (shift1 t2) -> P G (tm_lett e1 e2) t2) ->
       (forall (G : context) (t t' e : term),
        wf G -> oft G e t -> P G e t -> eqc G t t' kd_type -> P G e t') ->
       (forall (G : context) (P2 : term) (G' : context) (T k t M e : term),
        wf G -> oflm G M P2 ->
        P1 G M P2 ->
        linv P2 G' T ->
        fsts T k ->
        oft (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) e) (subst (sh (2 + length G')) t) ->
        P (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) e) (subst (sh (2 + length G')) t) ->
        P G (tm_letm M e) t) ->
       (forall (G : context) (t M : term),
        wf G -> ofm G M (sg_dyn t) -> P0 G M (sg_dyn t) -> P G (tm_ext M) t) ->
       (forall (G : list term) (i : nat) (T c : term),
        wf G -> index i G (cl_md T c) -> P0 G (var i) (subst (sh (S i)) T)) ->
       (forall G : context, wf G -> P0 G md_star sg_one) ->
       (forall (G : context) (c k : term), wf G -> ofc G c k -> P0 G (md_stat c) (sg_stat k)) ->
       (forall (G : context) (e t : term), wf G -> oft G e t -> P G e t -> P0 G (md_dyn e) (sg_dyn t)) ->
       (forall (G : context) (T1 T2 k1 M : term),
        wf G -> ofs G T1 ->
        fsts T1 k1 ->
        ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) ->
        P0 (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2) ->
        P0 G (md_lam T1 M) (sg_pi T1 T2)) ->
       (forall (G : context) (T1 T2 M1 M2 c2 : term),
        wf G -> ofm G M1 (sg_pi T1 T2) ->
        P0 G M1 (sg_pi T1 T2) ->
        ofm G M2 T1 -> P0 G M2 T1 -> fstm G M2 c2 -> P0 G (md_app M1 M2) (subst1 c2 T2)) ->
       (forall (G : context) (T1 T2 M1 M2 : term),
        wf G -> ofm G M1 T1 ->
        P0 G M1 T1 -> ofm G M2 T2 -> P0 G M2 T2 -> P0 G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))) ->
       (forall (G : context) (T1 T2 M : term),
        wf G -> ofm G M (sg_sigma T1 T2) -> P0 G M (sg_sigma T1 T2) -> P0 G (md_pi1 M) T1) ->
       (forall (G : context) (T1 T2 M c : term),
        wf G -> ofm G M (sg_sigma T1 T2) ->
        P0 G M (sg_sigma T1 T2) -> fstm G M c -> P0 G (md_pi2 M) (subst1 (cn_pi1 c) T2)) ->
       (forall (G : context) (t T e M : term),
        wf G -> oft G e t ->
        P G e t ->
        ofm (G; cl_tm t) M (shift1 T) -> P0 (G; cl_tm t) M (shift1 T) -> P0 G (md_lett e M) T) ->
       (forall (G : context) (M P2 : term),
        wf G -> oflm G M P2 -> P1 G M P2 -> P0 G (md_cir M) (sg_circ P2)) ->
       (forall (G : context) (k k' M c : term),
        wf G -> ofm G M (sg_stat k') ->
        P0 G M (sg_stat k') -> fstm G M c -> ofc G c k -> P0 G M (sg_stat k)) ->
       (forall (G : context) (T1 T2 T3 k1 M : term),
        wf G -> ofm G M (sg_pi T1 T3) ->
        P0 G M (sg_pi T1 T3) ->
        fsts T1 k1 ->
        ofm (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (md_app (subst (sh 2) M) (var 0)) (shift1 T2) ->
        P0 (G; cl_cn k1; cl_md (shift1 T1) (var 0)) (md_app (subst (sh 2) M) (var 0)) (shift1 T2) ->
        P0 G M (sg_pi T1 T2)) ->
       (forall (G : context) (T1 T2 M : term),
        wf G -> ofm G (md_pi1 M) T1 ->
        P0 G (md_pi1 M) T1 ->
        ofm G (md_pi2 M) T2 -> P0 G (md_pi2 M) T2 -> P0 G M (sg_sigma T1 (shift1 T2))) ->
       (forall (G : context) (T T' M : term), wf G -> ofm G M T -> P0 G M T -> subs G T T' -> P0 G M T') ->
       (forall (G : context) (M T : term), wf G -> ofm G M T -> P0 G M T -> P1 G (lm_ret M) (ps_down T)) ->
       (forall (G : context) (M T : term), wf G -> ofm G M T -> P0 G M T -> P1 G (lm_seal M T) (ps_down T)) ->
       (forall (G : context) (M1 M2 P2 : term) (G' : context) (T k P3 : term),
        wf G -> ofm G M1 (sg_circ P2) ->
        P0 G M1 (sg_circ P2) ->
        linv P2 G' T ->
        fsts T k ->
        oflm (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) M2) (subst (sh (2 + length G')) P3) ->
        P1 (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) 
          (subst (under 2 (sh (length G'))) M2) (subst (sh (2 + length G')) P3) ->
        P1 G (lm_bind M1 M2) P3) ->
       (forall (G : context) (k t P2 e M : term),
        wf G -> oft G e (cn_exists k t) ->
        P G e (cn_exists k t) ->
        oflm (G; cl_cn k; cl_tm t) M (subst (sh 2) P2) ->
        P1 (G; cl_cn k; cl_tm t) M (subst (sh 2) P2) -> P1 G (lm_unpack e M) P2) ->
       (forall (G : context) (T T' M : term), wf G -> oflm G M T -> P1 G M T -> subps G T T' -> P1 G M T') ->
       (forall (c : context) (t t0 : term), wf c -> oft c t t0 -> P c t t0) /\
       (forall (c : context) (t t0 : term), wf c -> ofm c t t0 -> P0 c t t0) /\
       (forall (c : context) (t t0 : term), wf c -> oflm c t t0 -> P1 c t t0)
.
Proof.
intros Pt Pm Plm.
intros.
exploit
  (dynamic_ind
     (fun G e t => wf G -> Pt G e t)
     (fun G M T => wf G -> Pm G M T)
     (fun G M T => wf G -> Plm G M T)) as Hind;
try (intros;
     eauto using wf;
     done).

(* unpack *)
{
intros G k t t' e1 e2 Hof ? ? ? Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hof)) as (l & Hofc).
invert Hofc; [].
intros; subst.
eauto 6 using wf; done.
}

(* lett *)
intros G t1 t2 e1 e2 Hof ? ? ? Hwf.
so (oft_reg _#3 Hwf Hof).
eauto using wf; done.

(* letm *)
{
intros G P G' T k t M e HofM ? Hinv Hfst ? ? Hwf.
so (oflm_reg _#3 Hwf HofM) as HofP.
so (linv_reg _#4 Hwf HofP Hinv) as (Hwf' & HofT).
eauto using wf; done.
}

(* lett *)
intros G t T e M Hof ? ? ? Hwf.
so (oft_reg _#3 Hwf Hof).
eauto using wf; done.

(* extpi *)
intros G T1 T2 T3 k1 M Hof ? ? ? ? Hwf.
so (ofm_reg _#3 Hwf Hof) as Hofs.
invert Hofs; [].
intros; subst.
eauto using wf; done.

(* bind *)
{
intros G M1 M2 P1 G' T k P2 HofM1 IH1 Hinv Hfst HofM2 IH2 Hwf.
so (ofm_reg _#3 Hwf HofM1) as Hofcirc.
invert Hofcirc; [].
intro HofP1.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
so (IH2 (wf_md Hwf' HofT Hfst)).
eauto; done.
}

(* unpack *)
{
intros G k t P e M Hofe IH1 HofM IH2 Hwf.
so (oft_reg _#3 Hwf Hofe) as Hofex.
so (ofc_ofcr _#3 Hofex) as (l & Hofex').
invert Hofex'; [].
intros Hofk Hoft.
so (IH2 (wf_tm (wf_cn Hwf Hofk) Hoft)).
eauto.
}

(* epilogue *)
destruct_all Hind.
do2 2 split; auto; done.
Qed.


Lemma ofc_wf_ind :
  forall (P : context -> term -> term -> Prop),
  (forall (G : list term) (i : nat) (k : term),
   wf G -> index i G (cl_cn k) -> P G (var i) (subst (sh (S i)) k)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofk G k1 ->
   ofc (G; cl_cn k1) c k2 ->
   P (G; cl_cn k1) c k2 -> P G (cn_lam k1 c) (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 (kd_pi k1 k2) ->
   P G c1 (kd_pi k1 k2) ->
   ofc G c2 k1 -> P G c2 k1 -> P G (cn_app c1 c2) (subst1 c2 k2)) ->
  (forall (G : context) (k1 k2 c1 c2 : term),
   wf G -> ofc G c1 k1 ->
   P G c1 k1 ->
   ofc G c2 (subst1 c1 k2) ->
   P G c2 (subst1 c1 k2) ->
   ofk (G; cl_cn k1) k2 ->
   P G (cn_pair c1 c2) (kd_sigma k1 k2)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G c (kd_sigma k1 k2) ->
   P G c (kd_sigma k1 k2) -> P G (cn_pi1 c) k1) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G c (kd_sigma k1 k2) ->
   P G c (kd_sigma k1 k2) -> P G (cn_pi2 c) (subst1 (cn_pi1 c) k2)) ->
  (forall G : context, wf G -> P G cn_star kd_one) ->
  (forall G : context, wf G -> P G cn_unit kd_type) ->
  (forall (G : context) (t1 t2 : term),
   wf G -> ofc G t1 kd_type ->
   P G t1 kd_type ->
   ofc G t2 kd_type -> P G t2 kd_type -> P G (cn_arrow t1 t2) kd_type) ->
  (forall (G : context) (t1 t2 : term),
   wf G -> ofc G t1 kd_type ->
   P G t1 kd_type ->
   ofc G t2 kd_type -> P G t2 kd_type -> P G (cn_prod t1 t2) kd_type) ->
  (forall (G : context) (k t : term),
   wf G -> ofk G k ->
   ofc (G; cl_cn k) t kd_type ->
   P (G; cl_cn k) t kd_type -> P G (cn_all k t) kd_type) ->
  (forall (G : context) (k t : term),
   wf G -> ofk G k ->
   ofc (G; cl_cn k) t kd_type ->
   P (G; cl_cn k) t kd_type -> P G (cn_exists k t) kd_type) ->
  (forall (G : context) (c : term),
   wf G -> ofc G c kd_type -> P G c kd_type -> P G c (kd_sing c)) ->
  (forall (G : context) (k1 k2 k3 c : term),
   wf G -> ofc G c (kd_pi k1 k3) ->
   P G c (kd_pi k1 k3) ->
   ofc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2 ->
   P (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2 ->
   P G c (kd_pi k1 k2)) ->
  (forall (G : context) (k1 k2 c : term),
   wf G -> ofc G (cn_pi1 c) k1 ->
   P G (cn_pi1 c) k1 ->
   ofc G (cn_pi2 c) (subst1 (cn_pi1 c) k2) ->
   P G (cn_pi2 c) (subst1 (cn_pi1 c) k2) ->
   ofk (G; cl_cn k1) k2 -> P G c (kd_sigma k1 k2)) ->
  (forall (G : context) (k k' c : term),
   wf G -> ofc G c k -> P G c k -> subk G k k' -> P G c k') ->
  (forall (c : context) (t t0 : term), wf c -> ofc c t t0 -> P c t t0)
.
Proof.
intro P.
intros.
exploit
  (static_wf_ind
     (fun _ _ => True)
     (fun _ _ _ => True)
     (fun _ _ _ => True)
     P
     (fun _ _ _ _ => True)) as Hind;
try (trivial; eauto; done); [].
destruct_all Hind; auto; done.
Qed.


Lemma wf_index_cn :
  forall G i k,
    wf G
    -> index i G (cl_cn k)
    -> ofk G (subst (sh (S i)) k).
Proof.
intros G i k Hwf Hindex.
remember (cl_cn k) as C eqn:HeqC.
revert Hwf HeqC.
induct Hindex.

(* 0 *)
intros C G Hwf ->.
invert Hwf; [].
intros _ Hofk.
eapply subst_ofk; eauto using weaken1; done.

(* S *)
intros i D G C _ IH Hwf ->.
so (IH (wf_tail _ _ Hwf) (eq_refl _)) as Hofk.
so (subst_ofk _#4 (weaken1 _ _ D) Hofk) as H; simp_sub in H.
simp_sub.
assumption.
Qed.


Lemma wf_index_tm :
  forall G i t,
    wf G
    -> index i G (cl_tm t)
    -> ofc G (subst (sh (S i)) t) kd_type.
Proof.
intros G i t Hwf Hindex.
remember (cl_tm t) as C eqn:HeqC.
revert Hwf HeqC.
induct Hindex.

(* 0 *)
intros C G Hwf ->.
invert Hwf; [].
intros _ Hofc.
so (subst_ofc _#5 (weaken1 _ _ (cl_tm t)) Hofc) as H; simp_sub in H.
assumption.

(* S *)
intros i D G C _ IH Hwf ->.
so (IH (wf_tail _ _ Hwf) (eq_refl _)) as Hofc.
so (subst_ofc _#5 (weaken1 _ _ D) Hofc) as H; simp_sub in H.
simp_sub.
assumption.
Qed.


Lemma wf_index_md :
  forall G i T c,
    wf G
    -> index i G (cl_md T c)
    -> ofs G (subst (sh (S i)) T) /\ c = var 0.
Proof.
intros G i T c Hwf Hindex.
remember (cl_md T c) as C eqn:HeqC.
revert Hwf HeqC.
induct Hindex.

(* 0 *)
intros C G Hwf ->.
invert Hwf; [].
intros G' k T' Hwf' Hofs Hfsts <- <- <-.
split; trivial; [].
simp_sub.
eapply subst_ofs; eauto using weaken2; done.

(* S *)
intros i D G C _ IH Hwf ->.
so (IH (wf_tail _ _ Hwf) (eq_refl _)) as (Hofk & Heq).
split; trivial; [].
so (subst_ofs _#4 (weaken1 _ _ D) Hofk) as H; simp_sub in H.
simp_sub.
assumption.
Qed.


Lemma wf_index_iscl :
  forall G i A,
    wf G
    -> index i G A
    -> iscl A.
Proof.
intros G i A Hwf Hindex.
revert Hwf.
induct Hindex.

(* 0 *)
intros A G Hwf.
invert Hwf; intros; subst; auto using iscl; done.

(* S *)
intros i B G A _ IH Hwf.
apply IH; [].
eapply wf_tail; eauto; done.
Qed.


Lemma wfh_inh :
  forall G G',
    wfh G G'
    -> exists s,
         subof sof G s (G;; G')
         /\ compose (sh (length G')) s = id.
Proof.
intros G G' Hwf.
induct Hwf.

(* nil *)
{
intros G.
exists id.
simpl.
auto using subof_id.
}

(* cn *)
{
intros G1 G2 k _ IH Hofk.
destruct IH as (s & Hs & Heq).
exists (dot (subst s (inh k)) s).
split.
  {
  simpl app.
  apply subof_dot; auto.
  simp_sub.
  apply sof_cn.
  eapply subst_ofc; eauto.
  apply inh_ofc; auto.
  }

  {
  simpl length.
  simp_sub.
  auto.
  }
}

(* tm *)
{
intros G1 G2 t _ IH _.
destruct IH as (s & Hs & Heq).
exists (dot nonsense s).
split.
  {
  simpl app.
  apply subof_dot; auto.
  simp_sub.
  apply sof_tm.
  }

  {
  simpl length.
  simp_sub.
  auto.
  }
}

(* md *)
{
intros G1 G2 k T _ IH HofT Hfst.
destruct IH as (s & Hs & Heq).
exists (dot nonsense (dot (subst s (inh k)) s)).
split.
  {
  simpl app.
  apply subof_dot.
    {
    apply subof_dot; auto.
    simp_sub.
    apply sof_cn.
    eapply subst_ofc; eauto.
    apply inh_ofc; auto.
    eapply ofs_fst; eauto.
    }
  simp_sub.
  apply sof_md.
  }

  {
  simpl length.
  simp_sub.
  auto.
  }
}
Qed.
