
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import Subst.
Require Import TSubst.
Require Import Formation.
Require Import Regularity.
Require Import Declarative.
Require Import Singleton.
Require Import SubjectReduction.
Require Import Principal.
Require Import Complete.


Inductive transp : (* kd *) term -> Prop :=
| transp_sing {c}
    : transp (kd_sing c)

| transp_pi {k1 k2}
    : transp k2
      -> transp (kd_pi k1 k2)

| transp_sigma {k1 k2}
    : transp k1
      -> transp k2
      -> transp (kd_sigma k1 k2)

| transp_one
    : transp kd_one.


Inductive stransp : (* sg *) term -> Prop :=
| stransp_one
    : stransp sg_one

| stransp_stat {k}
    : transp k
      -> stransp (sg_stat k)

| stransp_dyn {t}
    : stransp (sg_dyn t)

| stransp_pi {T1 T2}
    : stransp T2
      -> stransp (sg_pi T1 T2)

| stransp_sigma {T1 T2}
    : stransp T1
      -> stransp T2
      -> stransp (sg_sigma T1 T2)

| stransp_circ {P}
    : stransp (sg_circ P)
.


Inductive pstransp : (* ps *) term -> Prop :=
| pstransp_down {T}
    : stransp T
      -> pstransp (ps_down T)

| pstransp_ex {k P}
    : pstransp P
      -> pstransp (ps_ex k P)
.


Lemma subst_transp :
  forall s k,
    transp k
    -> transp (subst s k).
Proof.
intros s k H.
revert s.
induct H; intros; simp_sub; eauto using transp.
Qed.


Lemma subst_stransp :
  forall s T,
    stransp T
    -> stransp (subst s T).
Proof.
intros s k H.
revert s.
induct H; intros; simp_sub; eauto using stransp, subst_transp.
Qed.


Lemma subst_pstransp :
  forall s P,
    pstransp P
    -> pstransp (subst s P).
Proof.
intros s k H.
revert s.
induct H; intros; simp_sub; eauto using pstransp, subst_stransp.
Qed.


Lemma single_transp :
  forall c k,
    transp (single c k).
Proof.
intros c k.
revert c.
induct k; try (simpl; eauto using transp; done).
(* sigma *)
intros k1 IH1 k2 IH2 c.
simpl.
apply transp_sigma; auto using subst_transp.
Qed.


Lemma singles_stransp :
  forall c T,
    stransp (singles c T).
Proof.
intros c T.
revert c.
induct T; try (simpl; eauto using stransp, single_transp; done).
(* sigma *)
intros T1 IH1 T2 IH2 c.
simpl.
apply stransp_sigma; auto using subst_stransp.
Qed.


Lemma linv_pstransp :
  forall P G T,
    linv P G T
    -> pstransp P
    -> stransp T.
Proof.
intros P G T Hinv HP.
revert HP.
induct Hinv.

(* down *)
{
intros T HP.
invert HP; auto.
}

(* ex *)
{
intros k P G T _ IH Hex.
invert Hex; auto.
}
Qed.


(* Principal kinds are transparent. *)
Lemma prink_transp :
  forall G c k,
    prink G c k
    -> transp k.
Proof.
intros G c k H.
induct H; eauto using transp, subst_transp, single_transp.

(* app *)
{
intros G k1 k2 c1 c2 _ IH _.
apply subst_transp.
invert IH; auto.
}

(* pi1 *)
{
intros G k1 k2 c _ IH.
invert IH; auto.
}

(* pi2 *)
{
intros G k1 k2 c _ IH.
apply subst_transp.
invert IH; auto.
}
Qed.


(* Principal signatures of non-lax modules are transparent. *)
Lemma prins_stransp :
  forall G M T,
    prins G M T
    -> stransp T.
Proof.
intros G M T H.
induct H; eauto using stransp, subst_stransp, singles_stransp, prink_transp.

(* lam *)
{
intros G T1 T2 k1 M _ _ _ IH.
apply stransp_pi.
so (subst_stransp (dot nonsense id) _ IH) as H.
simp_sub in H.
exact H.
}

(* app *)
{
intros G T1 T2 T1' M1 M2 c2 _ IH _ _ _ _.
apply subst_stransp.
invert IH; auto.
}

(* pi1 *)
{
intros G T1 T2 M _ IH.
invert IH; auto.
}

(* pi2 *)
{
intros G T1 T2 M c _ IH _.
apply subst_stransp.
invert IH; auto.
}
Qed.


Lemma transp_single_self :
  forall G c k,
    wf G
    -> ofc G c k
    -> transp k
    -> eqk G k (single c k).
Proof.
intros G c k Hwf Hofc Hk.
remember (skel k) as sk eqn:Hsk.
revert k Hsk G c Hwf Hofc Hk.
induct sk.

(* type *)
{
intros k Hsk G c Hwf Hofc Hk.
destruct Hk; try (discriminate Hsk).
}

(* sing *)
{
intros k.
case k; clear k; try (intros; discriminate Hsk).
intros c' Hsk G c Hwf Hofc _.
simpl.
eauto with static.
}

(* pi *)
{
intros sk1 _ sk2 IH k.
case k; clear k; try (intros; discriminate Hsk).
intros k1 k2 Hsk G c Hwf Hofc Hk.
invert Hk.
intros Hk2.
so (ofc_reg _#3 Hwf Hofc) as Hofpi.
invert Hofpi.
intros Hofk1 _.
simpl in Hsk.
injection Hsk.
intros Hsk2 _.
so (wf_cn Hwf Hofk1) as Hwf'.
assert (ofc (G; cl_cn k1) (cn_app (shift1 c) (var 0)) k2) as Hofc'.
  {
  eapply rel_quest.
    {
    eapply ofc_app.
      {
      eapply rel_quest.
        {
        eapply subst_ofc; eauto.
        apply subof_sh; auto with sequence.
        }
      simp_sub.
      reflexivity.
      }
    
      {
      apply ofc_var; eauto with sequence.
      }
    }
  simp_sub.
  reflexivity.
  }
so (IH _ Hsk2 _ _ Hwf' Hofc' Hk2) as Heqk2.
simpl.
apply eqk_pi; eauto with static.
}

(* sigma *)
{
intros sk1 IH1 sk2 IH2 k.
case k; clear k; try (intros; discriminate Hsk).
intros k1 k2 Hsk G c Hwf Hofc Hk.
invert Hk.
intros Hk1 Hk2.
simpl in Hsk.
injection Hsk.
intros Hsk2 Hsk1.
so (IH1 _ Hsk1 _ _ Hwf (ofc_pi1 Hofc) Hk1) as Heqk1.
so (ofc_reg _#3 Hwf Hofc) as Hofsigma.
invert Hofsigma.
intros Hofk1 Hofk2.
so (wf_cn Hwf Hofk1) as Hwf'.
assert (sk2 = skel (subst1 (cn_pi1 c) k2)) as Hsk2'.
  {
  unfold subst1.
  rewrite -> subst_skel; eauto using ofk_hkd.
  }
so (IH2 _ Hsk2' _ _ Hwf (ofc_pi2 Hofc) (subst_transp _ _ Hk2)) as Heqk2.
simpl.
apply eqk_sigma.
  {
  eapply eqk_trans; eauto.
  apply eqk_refl.
  apply single_ofk; eauto with static.
  }

  {
  apply (@eqk_trans _ _ (subst (dot (shift1 (cn_pi1 c)) sh1) k2)).
    {
    rewrite <- (subst_dot0_sh k2) at 1.
    eapply seq_ofk; eauto.
    apply subrel_dot; eauto using subrel_sh with sequence.
    simp_sub.
    apply seq_cn.
    so (subst_eqc _#6 (subof_conv_0 _#3 Heqk1) (single_eqc_weak _#3 Hwf (ofc_pi1 Hofc))) as H.
    simp_sub in H.
    exact H.
    }
  rewrite -> subst_single; eauto using ofk_hkd.
  simp_sub.
  so (subst_eqk _#5 (weaken1 _ _ (cl_cn k1)) Heqk2) as Heqk2'.
  simp_sub in Heqk2'.
  unfold shift1 in Heqk2'.
  rewrite -> subst_single in Heqk2'; eauto using ofk_hkd, subst_hkd.
  simp_sub in Heqk2'.
  exact Heqk2'.
  }
}

(* one *)
{
intros k Hsk G c Hwf Hofc Hk.
destruct Hk; try (discriminate Hsk).
simpl.
eauto with static.
}
Qed.


Lemma stransp_singles_self :
  forall G c T k,
    wf G
    -> ofs G T
    -> fsts T k
    -> ofc G c k
    -> stransp T
    -> eqs G T (singles c T).
Proof.
intros G c T k Hwf Hofs Hfst Hofc HT.
remember (skels T) as sk eqn:Hsk.
revert T Hsk G c k Hwf Hofs Hfst Hofc HT.
induct sk.

(* one *)
{
intros T Hsk G c k Hwf Hofs Hfst Hofc HT.
destruct HT; try (discriminate Hsk).
simpl.
eauto with static.
}

(* stat *)
{
intros T.
case T; clear T; try (intros; discriminate Hsk).
intros k Hsk G c k' Hwf Hof Hfst Hofc HT.
invert Hfst.
intros <-.
invert HT.
intros Hk.
simpl.
apply eqs_stat.
apply transp_single_self; auto.
}

(* dyn *)
{
intros T.
case T; clear T; try (intros; discriminate Hsk).
intros t Hsk G c k Hwf Hofs Hfst Hofc HT.
invert Hofs.
intros Hoft.
simpl.
eauto with static.
}

(* pi *)
{
intros sk1 _ sk2 IH T.
case T; clear T; try (intros; discriminate Hsk).
intros T1 T2 Hsk G c k Hwf Hofs Hfst Hofc HT.
invert Hsk.
intros Hsk2 _.
invert Hfst.
intros k1 k2 Hfst1 Hfst2.
invert Hofs.
intros k1a HofT1 Hfst1a HofT2 <-.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a.
invert HT.
intros HT2.
simpl.
eapply eqs_pi; eauto with static.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
eapply IH; eauto using wf.
eapply rel_quest.
  {
  eapply ofc_app.
    {
    eapply rel_quest.
      {
      eapply subst_ofc; eauto.
      apply subof_sh; auto with sequence.
      }
    simp_sub.
    reflexivity.
    }
  
    {
    apply ofc_var; eauto with sequence.
    }
  }
simp_sub.
reflexivity.
}

(* sigma *)
{
intros sk1 IH1 sk2 IH2 T.
case T; clear T; try (intros; discriminate Hsk).
intros T1 T2 Hsk G c k Hwf Hofs Hfst Hofc HT.
invert Hsk.
intros Hsk2 Hsk1.
invert Hfst.
intros k1 k2 Hfst1 Hfst2 <-.
invert Hofs.
intros k1a HofT1 Hfst1a HofT2.
so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a.
invert HT.
intros HT1 HT2.
so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
so (wf_cn Hwf Hofk1) as Hwf'.
assert (sk2 = skels (subst1 (cn_pi1 c) T2)) as Hsk2'.
  {
  unfold subst1.
  rewrite -> subst_skels; eauto using ofs_hsg.
  }
so (IH1 _ Hsk1 _ _ _ Hwf HofT1 Hfst1 (ofc_pi1 Hofc) HT1) as HeqT1.
so (IH2 _ Hsk2' _ _ _ Hwf (subst1_ofs _#4 (ofc_pi1 Hofc) HofT2) (subst_fsts _#3 Hfst2) (ofc_pi2 Hofc) (subst_stransp _ _ HT2)) as HeqT2.
simpl.
eapply eqs_sigma; eauto.
apply (@eqs_trans _ _ (subst (dot (shift1 (cn_pi1 c)) sh1) T2)).
  {
  rewrite <- (subst_dot0_sh T2) at 1.
  eapply seq_ofs; eauto.
  apply subrel_dot; eauto using subrel_sh with sequence.
  simp_sub.
  apply seq_cn.
  so (singles_fst (cn_pi1 c) _ _ Hfst1) as Hfst1s.
  so (eqs_fst _#5 Hwf HeqT1 Hfst1 Hfst1s) as Heqk1.
  so (subst_eqc _#6 (subof_conv_0 _#3 Heqk1) (single_eqc_weak _#3 Hwf (ofc_pi1 Hofc))) as H.
  simp_sub in H.
  exact H.
  }
so (ofs_hsg _ _ HofT2) as HsT2.
rewrite -> subst_singles; auto.
simp_sub.
so (subst_eqs _#5 (weaken1 _ _ (cl_cn k1)) HeqT2) as HeqT2'.
simp_sub in HeqT2'.
unfold shift1 in HeqT2'.
rewrite -> subst_singles in HeqT2'; eauto using subst_hsg.
simp_sub in HeqT2'.
exact HeqT2'.
}

(* circ *)
{
intros T.
case T; clear T; try (intros; discriminate Hsk).
intros P Hsk G c k Hwf Hofs Hfst Hofc HT.
simpl.
apply eqs_refl; auto.
}
Qed.
