
Require Import Tactics.
Require Import Sequence.
Require Import Static.
Require Import TSubst.
Require Import Regularity.
Require Import Singleton.
Require Import AlgorithmEq.
Require Import SubjectReduction.
Require Import Algorithm.
Require Import SoundEq.
Require Import Principal.


Lemma aeqc_sound :
  forall G c1 c2 k,
    wf G
    -> ofc G c1 k
    -> ofc G c2 k
    -> aeqc G c1 c2 k
    -> eqc G c1 c2 k.
Proof.
intros G c1 c2 k Hwf Hof1 Hof2 Heq.
unfold aeqc in Heq.
eapply aeq_sound; eauto using ctxeq_refl.
apply eqk_refl.
eapply ofc_reg; eauto.
Qed.


Lemma asubk_sound :
  forall G k k',
    wf G
    -> ofk G k
    -> ofk G k'
    -> asubk G k k'
    -> subk G k k'.
Proof.
intros G k k' Hwf Hofk Hofk' Hsub.
revert Hwf Hofk Hofk'.
induct Hsub; eauto with static.

(* sing *)
{
intros G c c' Haeq Hwf Hofk Hofk'.
invert Hofk.
intros Hofc.
invert Hofk'.
intros Hofc'.
apply subk_refl.
apply eqk_sing.
eapply aeqc_sound; eauto.
}

(* sing_t *)
{
intros G c Hwf Hofk Hofk'.
invert Hofk.
intro Hofc.
apply subk_sing_t; auto.
}

(* pi *)
{
intros G k1 k1' k2 k2' _ IH1 _ IH2 Hwf Hofk Hofk'.
invert Hofk.
intros Hofk1 Hofk2.
invert Hofk'.
intros Hofk1' Hofk2'.
so (IH1 Hwf Hofk1' Hofk1) as Hsubk1.
apply subk_pi; auto.
apply IH2; auto using wf.
so (subst_ofk _#4 (subof_subsume_0 _#3 Hsubk1) Hofk2) as H.
simp_sub in H.
exact H.
}

(* sigma *)
{
intros G k1 k1' k2 k2' _ IH1 _ IH2 Hwf Hofk Hofk'.
invert Hofk.
intros Hofk1 Hofk2.
invert Hofk'.
intros Hofk1' Hofk2'.
so (IH1 Hwf Hofk1 Hofk1') as Hsubk1.
apply subk_sigma; auto.
apply IH2; auto using wf.
so (subst_ofk _#4 (subof_subsume_0 _#3 Hsubk1) Hofk2') as H.
simp_sub in H.
exact H.
}
Qed.


Lemma aofk_aofc_sound :
  (forall G k,
     wf G
     -> aofk G k
     -> ofk G k)
  /\
  (forall G c k,
     wf G
     -> aprink G c k
     -> prink G c k)
  /\
  (forall G c k,
     wf G
     -> aofc G c k
     -> ofk G k
     -> ofc G c k).
Proof.
exploit
  (aofk_aofc_ind
     (fun G k => wf G -> ofk G k)
     (fun G c k => wf G -> prink G c k)
     (fun G c k => wf G -> ofk G k -> ofc G c k)) as Hind;
eauto 6 using ofk, prink, ofc, wf.

(* app *)
{
intros G k1 k2 c1 c2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hprin1.
so (ofc_reg _#3 Hwf (prink_sound _#3 Hwf Hprin1)) as Hofpi.
invert Hofpi.
intros Hofk1 _.
so (IH2 Hwf Hofk1) as Hprin2.
eapply prink_app; eauto.
}

(* ofc *)
{
intros G c k k' _ IH Hsub Hwf Hofk.
so (IH Hwf) as Hprink.
so (prink_sound _#3 Hwf Hprink) as Hofc.
so (ofc_reg _#3 Hwf Hofc) as Hofk'.
so (asubk_sound _#3 Hwf Hofk' Hofk Hsub) as Hsub'.
eapply ofc_subsume; eauto.
}

(* epilogue *)
{
destruct_all Hind.
do2 2 split; eauto.
}
Qed.


Lemma aofk_sound :
  forall G k,
    wf G
    -> aofk G k
    -> ofk G k.
Proof.
exact (aofk_aofc_sound andel).
Qed.


Lemma aprink_sound_pre :
  forall G c k,
     wf G
     -> aprink G c k
     -> prink G c k.
Proof.
exact (aofk_aofc_sound anderl).
Qed.


Lemma aofc_sound :
  forall G c k,
    wf G
    -> aofc G c k
    -> ofk G k
    -> ofc G c k.
Proof.
exact (aofk_aofc_sound anderr).
Qed.


Lemma aprink_sound :
  forall G c k,
    wf G
    -> aprink G c k
    -> ofc G c k.
Proof.
intros G c k Hwf Hprin.
eapply prink_sound; eauto.
apply aprink_sound_pre; auto.
Qed.


Lemma aofx_sound :
  (forall G T,
     wf G
     -> aofs G T
     -> ofs G T)
  /\
  (forall G T,
     wf G
     -> aofps G T
     -> ofps G T).
Proof.
exploit
  (aofsig_ind
     (fun G T => wf G -> ofs G T)
     (fun G P => wf G -> ofps G P)) as Hind;
eauto 8 using ofs, ofps, aofk_sound, aofc_sound, wf, ofs_fst with static.
destruct Hind; split; eauto.
Qed.


Lemma aofs_sound :
  forall G T,
    wf G
    -> aofs G T
    -> ofs G T.
Proof.
exact (aofx_sound andel).
Qed.


Lemma aofps_sound :
  forall G T,
    wf G
    -> aofps G T
    -> ofps G T.
Proof.
exact (aofx_sound ander).
Qed.


Lemma asubx_sound :
  (forall G T T',
     wf G
     -> ofs G T
     -> ofs G T'
     -> asubs G T T'
     -> subs G T T')
  /\
  (forall G T T',
     wf G
     -> ofps G T
     -> ofps G T'
     -> asubps G T T'
     -> subps G T T').
Proof.
exploit
  (asubsig_ind
     (fun G T T' =>
        wf G
        -> ofs G T
        -> ofs G T'
        -> subs G T T')
     (fun G T T' =>
        wf G
        -> ofps G T
        -> ofps G T'
        -> subps G T T')) as Hind.

(* one *)
{
eauto with static.
}

(* stat *)
{
intros G k k' Hasubk Hwf Hofs Hofs'.
invert Hofs.
intros Hofk.
invert Hofs'.
intros Hofk'.
apply subs_stat.
eapply asubk_sound; eauto.
}

(* dyn *)
{
intros G t t' Haeq Hwf Hofs Hofs'.
invert Hofs.
intros Hoft.
invert Hofs'.
intros Hoft'.
apply subs_refl.
apply eqs_dyn.
eapply aeq_sound; eauto using ctxeq_refl with static.
}

(* pi *)
{
intros G T1 T1' k1' T2 T2' _ IH1 Hfst' _ IH2 Hwf Hofs Hofs'.
invert Hofs.
intros k1 HofT1 Hfst HofT2.
invert Hofs'.
intros k1'a HofT1' Hfst'a HofT2'.
so (fsts_fun _#3 Hfst' Hfst'a); subst k1'a.
so (IH1 Hwf HofT1' HofT1) as HsubT1.
eapply subs_pi; eauto.
so (ofs_fst _#3 HofT1' Hfst') as Hofk1'.
so (subs_fst _#5 Hwf HsubT1 Hfst' Hfst) as Hsubk1.
so (subst_ofs _#4 (subof_subsume_0 _#3 Hsubk1) HofT2) as HofT2w.
simp_sub in HofT2w.
so (IH2 (wf_cn Hwf Hofk1') HofT2w HofT2') as HsubT2.
exact HsubT2.
}

(* sigma *)
{
intros G T1 T1' k1 T2 T2' _ IH1 Hfst _ IH2 Hwf Hofs Hofs'.
invert Hofs.
intros k1a HofT1 Hfsta HofT2.
invert Hofs'.
intros k1' HofT1' Hfst' HofT2'.
so (fsts_fun _#3 Hfst Hfsta); subst k1a.
so (IH1 Hwf HofT1 HofT1') as HsubT1.
eapply subs_sigma; eauto.
so (ofs_fst _#3 HofT1 Hfst) as Hofk1.
so (subs_fst _#5 Hwf HsubT1 Hfst Hfst') as Hsubk1.
so (subst_ofs _#4 (subof_subsume_0 _#3 Hsubk1) HofT2') as HofT2'w.
simp_sub in HofT2'w.
so (IH2 (wf_cn Hwf Hofk1) HofT2 HofT2'w) as HsubT2.
exact HsubT2.
}

(* circ *)
{
intros G P P' _ IH Hwf HofT HofT'.
invert HofT.
intro HofP.
invert HofT'.
intro HofP'.
eauto with static.
}

(* down *)
{
intros G P P' _ IH Hwf HofT HofT'.
invert HofT.
intro HofP.
invert HofT'.
intro HofP'.
eauto with static.
}

(* ex *)
{
intros G k P P' _ IH Hwf Hofex HofP'.
invert Hofex.
intros Hofk HofP.
apply subps_ex_left; auto.
apply IH; auto using wf.
exact (subst_ofps _#4 (weaken1 _#3) HofP').
}

(* epilogue *)
{
destruct Hind; eauto.
}
Qed.


Lemma asubs_sound :
  forall G T T',
    wf G
    -> ofs G T
    -> ofs G T'
    -> asubs G T T'
    -> subs G T T'.
Proof.
exact (asubx_sound andel).
Qed.


Lemma asubps_sound :
  forall G T T',
    wf G
    -> ofps G T
    -> ofps G T'
    -> asubps G T T'
    -> subps G T T'.
Proof.
exact (asubx_sound ander).
Qed.


Lemma close_snoc_invert :
  forall k G P1 P2,
    close (nil; cl_cn k;; G) P1 P2
    -> exists P2',
         close G P1 P2'
         /\ P2 = ps_ex k P2'.
Proof.
intros k G P1 P2 Hclose.
remember (nil; cl_cn k;; G) as G' eqn:Heq.
revert G Heq.
induct Hclose.

(* nil *)
{
intros P G Heq.
exfalso.
refine (app_cons_not_nil _#4 _); eauto.
}

(* cons *)
{
intros GG l P P' Hclose IH G Heq.
so (app_eq_cons _#5 Heq) as [(G' & -> & ->) | (-> & Heq')].
  {
  rewrite <- app_comm_cons in Heq.
  so (IH _ (eq_refl _)) as (P2' & Hclose' & ->).
  exists P2'.
  split; auto.
  apply close_cons; auto.
  }

  {
  injection Heq'.
  intros <- <-.
  invert Hclose; [].
  intros <-.
  eauto using close.
  }
}
Qed.


Lemma close_impl_lclose :
  forall P1 G T P2 P3,
    linv P1 G T
    -> close G P2 P3
    -> lclose P1 P2 P3.
Proof.
intros P1 G T P2 P3 Hinv Hclose.
revert P2 P3 Hclose.
induct Hinv.

(* down *)
{
intros T P2 P3 Hclose.
invert Hclose.
intros <-.
apply lclose_down; auto.
}

(* ex *)
{
intros k P G T _ IH P2 P3 Hclose.
so (close_snoc_invert _#4 Hclose) as (P3' & Hclose' & ->).
apply lclose_ex; auto.
}
Qed.


Lemma aprinx_sound :
  (forall G e t,
     wf G
     -> aprint G e t
     -> print G e t)
  /\
  (forall G M T,
     wf G
     -> aprins G M T
     -> prins G M T)
  /\ 
  (forall G M P,
     wf G
     -> aprinps G M P
     -> prinps G M P).
Proof.
exploit
  (aprin_ind
     (fun G e t => wf G -> print G e t)
     (fun G M T => wf G -> prins G M T)
     (fun G M T => wf G -> prinps G M T)) as Hind;
eauto 7 using print, prins, prinps, aofk_sound, aofc_sound, aofs_sound, aprink_sound_pre, wf with static.

(* app *)
{
intros G t t1 t2 t1' e1 e2 _ IH1 Hnorm _ IH2 Haeqc Hwf.
so (IH1 Hwf) as Hprin1.
so (IH2 Hwf) as Hprin2.
eapply print_app; eauto.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hoft.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin2)) as Hoft1'.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hoft Hnorm) anderl)) as (l & H).
invert H.
intros Hoft1 _ _.
apply aeqc_sound; eauto.
}

(* papp *)
{
intros G u k t e c _ IH Hnorm Hofc Hwf.
so (IH Hwf) as Hprin.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin)) as Hofu.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk _ _.
eapply print_papp; eauto using aofc_sound.
}

(* pack *)
{
intros G k t c e u Haofk Haofc _ IH Heq Haoft Hwf.
so (aofk_sound _ _ Hwf Haofk) as Hofk.
so (aofc_sound _#3 (wf_cn Hwf Hofk) Haoft ofk_type) as Hoft.
eapply print_pack; eauto using aofc_sound, wf, ofc_reg, aeqc_sound, oft_reg, subst1_ofc_type.
so (aofc_sound _#3 Hwf Haofc Hofk) as Hofc.
eapply aeqc_sound; eauto.
  {
  exact (oft_reg _#3 Hwf (print_sound _#3 Hwf (IH Hwf))).
  }
eapply subst1_ofc_type; eauto using aofc_sound, wf.
}

(* unpack *)
{
intros G u k t t' e1 e2 _ IH1 Hnorm _ IH2 Heq Hwf.
so (IH1 Hwf) as Hprin1.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hofu.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
so (IH2 Hwf') as Hprin2.
so (oft_reg _#3 Hwf' (print_sound _#3 Hwf' Hprin2)) as Hoft'.
eapply print_unpack; eauto.
eapply aeqc_sound; eauto using wf.
  {
  eapply subst_ofc_type; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_tm.
  }

  {
  eapply subst_ofc_type; eauto.
  apply subof_dot.
    {
    apply subof_dot; eauto using subof_sh with sequence.
    simp_sub.
    apply sof_cn.
    apply weaken1_ofc.
    apply inh_ofc; auto.
    }
  simp_sub.
  apply sof_tm.
  }
}

(* fix *)
{
intros G t e u Hoft _ IH Heq Hwf.
so (aofc_sound _#3 Hwf Hoft) as Hoft'.
so (IH Hwf) as Hprin.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin)) as Hofu.
eapply print_fix; eauto.
eapply aeqc_sound; eauto with static.
}

(* lett *)
{
intros G t1 t2 e1 e2 _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hprin1.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hoft1.
eapply print_lett; eauto using wf.
}

(* letm *)
{
intros G P G' T k t M e _ IH1 Hinv Hfst _ IH2 Heq Hwf.
so (IH1 Hwf) as Hprin1.
so (oflm_reg _#3 Hwf (prinps_sound _#3 Hwf Hprin1)) as HofP.
so (linv_reg_gen _#4 Hwf HofP Hinv) as (Hwfi' & HofT).
so (wf_wfh_append _ _ Hwf Hwfi') as Hwf'.
so (ofs_fst _#3 HofT Hfst) as Hofk.
so (wf_md Hwf' HofT Hfst) as Hwf''.
so (IH2 Hwf'') as Hprin2.
so (oft_reg _#3 Hwf'' (print_sound _#3 Hwf'' Hprin2)) as Hoft.
eapply print_letm; eauto.
eapply aeqc_sound; eauto using wf.
  {
  eapply subst_ofc_type; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_md.
  }

  {
  eapply subst_ofc_type; eauto.
  assert (subof sof (G;; G'; cl_cn k) (compose (cinh G') (sh (S (length G')))) (G;; G')) as Hs.
    {
    eapply (subof_sof_trans _ G); eauto using cinh_subof.
    rewrite -> app_comm_cons.
    replace (S (length G')) with (length (G'; cl_cn k)) by reflexivity.
    apply subof_sh.
    apply app_truncate.
    }
  apply subof_dot.
    {
    apply subof_dot; auto.
    simp_sub.
    apply sof_cn.
    eapply (subst_ofc _ (G;; G')); eauto using inh_ofc.
    }
  simp_sub.
  apply sof_md.
  }
}

(* app *)
{
intros G T1 T2 T1' M1 M2 c2 _ IH1 _ IH2 Hsub Hfst Hwf.
so (IH1 Hwf) as Hprin1.
so (IH2 Hwf) as Hprin2.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin1)) as Hofpi.
invert Hofpi.
intros _ HofT1 _ _.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin2)) as HofT1'.
eapply prins_app; eauto.
apply asubs_sound; auto.
}

(* lett *)
{
intros G t T e M _ IH1 _ IH2 Hwf.
so (IH1 Hwf) as Hprin1.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hoft.
eapply prins_lett; eauto using wf.
}

(* seal *)
{
intros G M T2 T1 _ IH1 HofT2 Hsub Hwf.
so (IH1 Hwf) as Hprin.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin)) as HofT1.
so (aofs_sound _ _ Hwf HofT2) as HofT2'.
eapply prinps_seal; eauto using asubs_sound.
}

(* bind *)
{
intros G M1 M2 P1 G' T k P2 P2' _ IH1 Hinv Hfst _ IH2 Hclose Hwf.
so (IH1 Hwf) as Hprin1.
so (ofm_reg _#3 Hwf (prins_sound _#3 Hwf Hprin1)) as Hofcirc.
invert Hofcirc.
intros HofP1.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
eapply prinps_bind; eauto using wf.
eapply close_impl_lclose; eauto.
}

(* unpack *)
{
intros G u k t P e M _ IH1 Hnorm _ IH2 Hwf.
so (IH1 Hwf) as Hprin1.
so (oft_reg _#3 Hwf (print_sound _#3 Hwf Hprin1)) as Hofu.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
eapply prinps_unpack; eauto using wf.
}

(* epilogue *)
{
destruct_all Hind.
do2 2 split; intros; eauto.
}
Qed.


Lemma aprint_sound_pre :
  forall G e t,
    wf G
    -> aprint G e t
    -> print G e t.
Proof.
exact (aprinx_sound andel).
Qed.


Lemma aprins_sound_pre :
  forall G M T,
    wf G
    -> aprins G M T
    -> prins G M T.
Proof.
exact (aprinx_sound anderl).
Qed.


Lemma aprinps_sound_pre :
  forall G M P,
    wf G
    -> aprinps G M P
    -> prinps G M P.
Proof.
exact (aprinx_sound anderr).
Qed.


Theorem aprint_sound :
  forall G e t,
    wf G
    -> aprint G e t
    -> oft G e t.
Proof.
intros G e t Hwf H.
apply print_sound; auto.
apply aprint_sound_pre; auto.
Qed.


Theorem aprins_sound :
  forall G M T,
    wf G
    -> aprins G M T
    -> ofm G M T.
Proof.
intros G M T Hwf H.
apply prins_sound; auto.
apply aprins_sound_pre; auto.
Qed.


Theorem aprinps_sound :
  forall G M P,
    wf G
    -> aprinps G M P
    -> oflm G M P.
Proof.
intros G M P Hwf H.
apply prinps_sound; auto.
apply aprinps_sound_pre; auto.
Qed.
